/*----------------------------------------------------------------------
  File    : bcdb.c
  Contents: generate a database from a Bayes classifier
  Author  : Christian Borgelt
  History : 2003.04.26 file created from file bcx.c
            2003.08.16 slight changes in error message output
            2007.02.13 adapted to modified module attset
            2010.10.12 adapted to modified module tabread
            2011.02.01 adapted to modified scanner module and as_parse()
            2011.07.29 adapted to modified attset and utility modules
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "params.h"
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef NBC_PARSE
#define NBC_PARSE
#endif
#include "nbayes.h"
#ifndef FBC_PARSE
#define FBC_PARSE
#endif
#include "fbayes.h"
#include "random.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "bcdb"
#define DESCRIPTION "generate a database from a Bayes classifier"
#define VERSION     "version 2.5 (2016.11.11)         " \
                    "(c) 2003-2016   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARSE     (-9)        /* parse error on classifier file */
#define E_LCORR    (-10)        /* invalid Laplace correction */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_LCORR   -10 */  "invalid Laplace correction %g",
  /*           -11 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner (for domain file) */
static ATTSET   *attset = NULL; /* attribute set */
static NBC      *nbc    = NULL; /* naive Bayes classifier */
static FBC      *fbc    = NULL; /* full  Bayes classifier */
static TABWRITE *twrite = NULL; /* table writer */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (nbc)    nbc_delete(nbc, 0);    \
  if (fbc)    fbc_delete(fbc, 0);    \
  if (attset) as_delete (attset);    \
  if (scan)   scn_delete(scan,   1); \
  if (twrite) twr_delete(twrite, 1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, buffer */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_bc   = NULL;      /* name of classifier file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *recsep  = NULL;      /* record separator */
  CCHAR   *fldsep  = NULL;      /* field  separator */
  CCHAR   *blank   = NULL;      /* blank  character */
  int     mout     = AS_ATT;    /* table file write mode */
  double  corr     = -INFINITY; /* Laplace correction value */
  double  reg[2]   = {-INFINITY,-INFINITY}; /* regularization value */
  int     dwnull   = 0;         /* distribute weight of null values */
  int     maxllh   = 0;         /* max. likelihood est. of variance */
  TPLID   n        = 1000, r;   /* number of tuples to generate */
  ATTID   m;                    /* number of attributes */
  long    seed;                 /* seed for random number generator */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */
  seed    = (long)time(NULL);   /* get a default seed value */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] bcfile [-d|-h hdrfile] "
                     "tabfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-n#      number of tuples to generate           "
                    "(default: %"TPLID_FMT")\n", n);
    printf("-s#      seed for random number generator       "
                    "(default: time)\n");
    printf("-L#      Laplace correction                     "
                    "(default: from classifier)\n");
    printf("-t#[:#]  Tikhonov regularization                "
                    "(default: from classifier)\n");
    printf("-v/V     (do not) distrib. weight for unknowns  "
                    "(default: ignore)\n");
    printf("-m/M     (do not) use maximum likelihood est.   "
                    "(default: unbiased)\n");
    printf("-a       align fields in output table           "
                    "(default: single separator)\n");
    printf("-w       do not write field names to the output file\n");
    printf("-r#      record separator                       "
                    "(default: \"\\n\")\n");
    printf("-f#      field  separator                       "
                    "(default: \" \")\n");
    printf("-b#      blank  character                       "
                    "(default: \" \")\n");
    printf("bcfile   file containing classifier description\n");
    printf("tabfile  file to write the generated data table to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate option */
          case 'n': n       = (int)strtol(s, &s, 0); break;
          case 's': seed    =      strtol(s, &s, 0); break;
          case 'L': corr    = strtod(s, &s);         break;
          case 't': getdblvec(s, &s, 2, reg);        break;
          case 'v': dwnull  = 0;                     break;
          case 'V': dwnull |= NBC_DWNULL;            break;
          case 'm': maxllh  = 0;                     break;
          case 'M': maxllh |= NBC_MAXLLH;            break;
          case 'a': mout   |= AS_ALIGN;              break;
          case 'w': mout   &= ~AS_ATT;               break;
          case 'r': optarg  = &recsep;               break;
          case 'f': optarg  = &fldsep;               break;
          case 'b': optarg  = &blank;                break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_bc  = s;      break;
        case  1: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if (k != 2) error(E_ARGCNT);  /* and the number of arguments */
  if ((corr < 0) && (corr > -INFINITY))
    error(E_LCORR, corr);       /* check the Laplace correction */
  if ((mout & AS_ALIGN) && (mout & AS_ATT))
    mout |= AS_ALNHDR;          /* set align to header flag */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read Bayes classifier --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the classifier file */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_bc) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
  if ((scn_token(scan) == T_ID)     /* determine classifier type */
  &&  (strcmp(scn_value(scan), "fbc") == 0))
       fbc = fbc_parse(attset, scan);
  else nbc = nbc_parse(attset, scan);
  if ((!fbc && !nbc)            /* parse the Bayes classifier */
  ||   !scn_eof(scan, 1))       /* and check for end of file */
    error(E_PARSE, scn_name(scan));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  m    = as_attcnt(attset);     /* get the number of attributes */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", m);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  if ((corr >= 0) || (reg[0] > -INFINITY) || (reg[1] > -INFINITY)
  ||  dwnull || maxllh) {       /* if parameters specified */
    if (corr < 0)               /* Laplace correction */
      corr = (fbc) ? fbc_getcorr(fbc) : nbc_getcorr(nbc);
    k      = (fbc) ? fbc_getmode(fbc) : nbc_getmode(nbc);
    if (reg[0] <= -INFINITY)    /* Tikhonov regularization */
      reg[0] = (fbc) ? fbc_getreg (fbc) : 0.0;
    if (reg[1] <= -INFINITY)    /* get the classifier's parameters */
      reg[1] = (fbc) ? fbc_getlim (fbc) : 0.0;
    if (dwnull & 0x8000) k = (k & ~NBC_DWNULL) | (dwnull & ~0x8000);
    if (maxllh & 0x8000) k = (k & ~NBC_MAXLLH) | (maxllh & ~0x8000);
    if (fbc) { fbc_setmode(fbc, k); fbc_setup(fbc, 0); }
    else     { nbc_setmode(nbc, k); nbc_setup(nbc, 0); }
  }                             /* set up with new parameters */

  /* --- generate database --- */
  twrite = twr_create();        /* create a table writer and */
  if (!twrite) error(E_NOMEM);  /* configure the characters */
  twr_xchars(twrite, recsep, fldsep, blank, "");
  t = clock();                  /* start timer, open output file */
  if (twr_open(twrite, NULL, fn_out) != 0)
    error(E_FOPEN, twr_name(twrite));
  fprintf(stderr, "writing %s ... ", twr_name(twrite));
  if ((mout & AS_ATT)           /* if to write a table header */
  &&  (as_write(attset, twrite, mout) != 0))
    error(E_FWRITE, twr_name(twrite));
  mout = AS_INST | (mout & ~AS_ATT);
  rseed((unsigned)seed);        /* init. random number generator */
  for (r = 0; r < n; r++) {     /* generate random tuples */
    if (fbc) fbc_rand(fbc, drand);   /* instantiate the */
    else     nbc_rand(nbc, drand);   /* attribute set */
    if (as_write(attset, twrite, mout) != 0)
      error(E_FWRITE, twr_name(twrite));
  }                             /* write the generated tuple */
  if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
  twr_delete(twrite, 1);        /* close the output file and */
  twrite = NULL;                /* delete the table writer */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
