/*----------------------------------------------------------------------
  File    : bci.c
  Contents: naive and full Bayes classifier induction
  Author  : Christian Borgelt
  History : 1998.12.08 file created from file dti.c
            1999.02.13 input from stdin, output to stdout added
            1999.03.25 weight distribution (option -t) added
            1999.04.17 simplified using the new module 'io'
            1999.05.15 check of class count moved behind table reading
            2000.11.21 adapted to redesigned module nbayes
            2000.11.30 full Bayes classifier induction added
            2001.02.11 bug in attribute output (full Bayes) fixed
            2001.07.16 adapted to modified module scan
            2003.08.16 slight changes in error message output
            2007.02.13 adapted to modified module attset
            2007.10.10 evaluation of attribute directions added
            2010.10.12 adapted to modified module tabread
            2011.02.02 adapted to modified module scanner
            2011.07.28 adapted to modified attset and utility modules
            2011.12.15 processing without table reading improved
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.29 adapted to new function as_target()
            2014.10.24 changed from LGPL license to MIT license
            2014.12.10 Tikhonov regularization parameter added
            2015.07.31 option to control "on the fly" computation added
            2015.07.31 forward selection of attributes based on AUC
            2015.08.03 forward selection based on sufficient statistics
            2015.08.05 balancing with integer factors made possible
            2015.09.11 error message "class attribute is not binary"
            2015.09.18 date and time output for attribute selection
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <assert.h>
#include "arrays.h"
#include "params.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_DESC
#define AS_DESC
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef NBC_INDUCE
#define NBC_INDUCE
#endif
#include "nbayes.h"
#ifndef FBC_INDUCE
#define FBC_INDUCE
#endif
#include "fbayes.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "bci"
#define DESCRIPTION "naive and full Bayes classifier induction"
#define VERSION     "version 4.2 (2016.11.11)         " \
                    "(c) 1998-2016   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARSE     (-9)        /* parse error on domain file */
#define E_ATTCNT   (-10)        /* no usable attributes found */
#define E_UNKCLS   (-11)        /* unknown class attribute */
#define E_MULCLS   (-12)        /* multiple class attributes */
#define E_CLSTYPE  (-13)        /* class attribute is not nominal */
#define E_CLSBIN   (-14)        /* class attribute is not binary */
#define E_BALANCE  (-15)        /* unknown balancing mode */
#define E_SIMP     (-16)        /* unknown simplification mode */
#define E_LCORR    (-17)        /* invalid Laplace correction */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- prediction for AUC comp. --- */
  double out;                   /* computed prediction */
  double trg;                   /* actual target value */
} PRED;                         /* (prediction pair) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_ATTCNT  -10 */  "no (usable) attributes (need at least 1)",
  /* E_UNKCLS  -11 */  "unknown class attribute '%s'",
  /* E_MULCLS  -12 */  "multiple class attributes found",
  /* E_CLSTYPE -13 */  "class attribute '%s' is not nominal",
  /* E_CLSBIN  -14 */  "class attribute '%s' is not binary",
  /* E_BALANCE -15 */  "unknown balancing mode %c",
  /* E_SIMP    -16 */  "unknown simplification mode %c",
  /* E_LCORR   -17 */  "invalid Laplace correction %g",
  /*           -18 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner (for domain file) */
static TABREAD *tread  = NULL;  /* table reader */
static ATTSET  *attset = NULL;  /* attribute set */
static TABLE   *table  = NULL;  /* (training) data table */
static TABLE   *valid  = NULL;  /* validation data table */
static NBC     *nbc    = NULL;  /* naive Bayes classifier */
static FBC     *fbc    = NULL;  /* full  Bayes classifier */
static FILE    *out    = NULL;  /* classifier output file */

/*----------------------------------------------------------------------
  Forward Selection of Attributes
----------------------------------------------------------------------*/

static int cleanup (int errcode, PRED *preds, NBC *nbc, FBC *fbc)
{                               /* --- clean up after an error */
  if (nbc)   nbc_delete(nbc, 0);
  if (fbc)   fbc_delete(fbc, 0);
  if (preds) free(preds);       /* free all allocated memory */
  return errcode;               /* return the error code */
}  /* cleanup() */

/*--------------------------------------------------------------------*/

static int predcmp (const void *a, const void *b, void *data)
{                               /* --- compare two predictions */
  if (((PRED*)a)->out > ((PRED*)b)->out) return +1;
  if (((PRED*)a)->out < ((PRED*)b)->out) return -1;
  return 0;                     /* return sign of difference */
}  /* predcmp() */              /* of prediction values */

/*--------------------------------------------------------------------*/

static void printtime (void)
{                               /* --- report current date and time */
  time_t    t;                  /* current time */
  struct tm *tm;                /* current time as a time struct */
  char      buf[64];            /* for formating the current time */

  t  = time(NULL);              /* get the current time and */
  tm = localtime(&t);           /* turn it into a time struct */
  if (!tm) return;              /* check for correct conversion */
  strftime(buf, sizeof(buf)-1, "%F %T", tm);
  fprintf(stderr, "%s ", buf);  /* format and print current time */
}  /* printtime() */

/*--------------------------------------------------------------------*/

static int fwdsel (TABLE *table, TABLE *valid, ATTID trgid, int full,
                   int mode, double corr, double reg[], ATTID maxsel,
                   int verbose)
{                               /* --- forward selection of atts. */
  ATTSET  *attset;              /* underlying attribute set */
  TPLID   j, n;                 /* loop variables for tuples */
  TUPLE   *tpl;                 /* to traverse the tuples */
  ATTID   i, k, m;              /* loop variables for attributes */
  ATT     *att;                 /* to traverse the attributes */
  PRED    *preds;               /* regression predictions */
  double  p;                    /* prediction confidence/probability */
  TPLID   neg, pos, sum;        /* negative and positive counters */
  double  amin, amax, aavg;     /* area under the ROC curve (AUC) */
  double  bmin, bmax, bavg;     /* best minimum/maximum/average AUC */
  ATTID   bid;                  /* index of best attribute */
  NBC     *nbc = NULL;          /* naive Bayes classifier */
  FBC     *fbc = NULL;          /* full  Bayes classifier */
  clock_t t;                    /* for time measurements */

  assert(table                  /* check the function arguments */
  &&     valid && (trgid >= 0));
  attset = tab_attset(table);   /* get the underlying attribute set */
  m = as_attcnt(attset);        /* an get the number of attributes */
  as_setmark(attset, -1);       /* unmark all attributes */
  att = as_att(attset, trgid);  /* get and mark */
  att_setmark(att, 0);          /* the target attribute */
  assert(att_valcnt(att) <= 2); /* check for at most two classes */
  if (att_valcnt(att) < 2) return 0;
  n = tab_tplcnt(valid);        /* get number of validation tuples */
  preds = (PRED*)malloc((size_t)n *sizeof(PRED));
  if (!preds) return E_NOMEM;   /* create a prediction array */

  t = clock();                  /* start timer, print log message */
  fprintf(stderr, "selecting attributes ...\n");
  if (full) {                   /* if to use a full Bayes classifier */
     fbc = fbc_induce(table, trgid, mode);
     if (!fbc) return cleanup(E_NOMEM, preds, nbc, fbc);
     fbc_setmode(fbc, mode);    /* create and initialize */
     fbc_setcorr(fbc, corr);    /* a full Bayes classifier */
     fbc_reglim (fbc, reg[0], reg[1]); }
  else {                        /* if to use a naive Bayes classifier */
     nbc = nbc_induce(table, trgid, mode);
     if (!nbc) return cleanup(E_NOMEM, preds, nbc, fbc);
     nbc_setmode(nbc, mode);    /* create and initialize */
     nbc_setcorr(nbc, corr);    /* a naive Bayes classifier */
  }                             /* (set estimation mode and params.) */
  bmin = bmax = bavg = -1.0;    /* clear the best attribute value */
  for (k = 0; k < m-1; k++) {   /* forward selection of attributes */
    bid = -1;                   /* clear the best attribute */
    for (i = 0; i < m; i++) {   /* traverse the attributes */
      att = as_att(attset, i);  /* skip target and selected attribs. */
      if (att_getmark(att) >= 0) continue;
      att_setmark(att, 1);      /* mark the attribute as selected */

      /* --- Bayes classifier setup --- */
      if (fbc) fbc_setup(fbc, FBC_MARKED);
      else     nbc_setup(nbc, NBC_MARKED);

      /* --- evaluate model on validation data --- */
      for (j = 0; j < n; j++) { /* traverse the validation tuples */
        tpl = tab_tpl(valid, j);/* (assess the model performance) */
        preds[j].trg = (double)tpl_colval(tpl, trgid)->n;
        if (nbc) { nbc_exec(nbc, tpl, NULL, NULL);
                   preds[j].out = nbc_post(nbc, 1); }
        else     { fbc_exec(fbc, tpl, NULL, NULL);
                   preds[j].out = fbc_post(fbc, 1); }
      }                         /* collect confidence of prediction */
      obj_qsort(preds, (size_t)n, sizeof(PRED), +1, predcmp, NULL);
      neg  = pos  = sum = 0;    /* initialize the case counters */
      amin = amax = 0;          /* and the area under the ROC curve */
      p = preds[0].out;         /* get the first prediction */
      for (j = n; --j >= 0; ) { /* traverse the cases (in reverse) */
        if (preds[j].out != p){ /* if the prediction changes */
          p     = preds[j].out;  /* note the new prediction value */
          amin += (double)neg *(double)sum;
          sum  += pos;          /* add current rectangle to AUC */
          amax += (double)neg *(double)sum;
          neg   = pos = 0;      /* and reinitialize the counters */
        }                       /* (for next prediction value) */
        if (preds[j].trg > 0) pos += 1;
        else                  neg += 1;
      }                         /* update the case counters */
      amin += (double)neg *(double)sum;
      sum  += pos;              /* add last rectangle to AUC */
      amax += (double)neg *(double)sum;
      amin /= p = (double)sum *(double)(n -sum);
      amax /= p;                /* normalize by rectangle area */
      amin = 2.0 *amin -1.0;    /* compute area above diagonal */
      amax = 2.0 *amax -1.0;    /* (minimum/maximum/average) */
      aavg = 0.5 *amin +0.5 *amax;
      if (verbose) {            /* if verbose reporting requested */
        printtime();            /* print the current date and time */
        fprintf(stderr, "% .6f [% .6f,% .6f] %s\n",
                aavg, amin, amax, att_name(att));
      }                         /* print minimum and maximum AUC */
      if ((aavg > bavg) || (maxsel < -2)) {  /* update best att. */
        bavg = aavg; bmin = amin; bmax = amax; bid = i; }

      /* --- clean up the variables --- */
      att_setmark(att, -1);     /* unmark the current attribute */
      if (maxsel < -1) break;   /* abort the loop if to select */
    }                           /* the attributes in a fixed order */

    if (bid < 0) break;         /* if no attribute selected, abort */
    printtime();                /* print the current date and time */
    att = as_att(attset, bid);  /* get and report best attribute */
    fprintf(stderr, "%3"ATTID_FMT": % .6f [% .6f,% .6f] %s\n",
            k+1, bavg, bmin, bmax, att_name(att));
    att_setwgt (att, (float)bavg);   /* note AUC value in weight */
    att_setmark(att, k+1);      /* mark best attribute with step */
    if ((k >= maxsel-1) && (maxsel >= 0)) { k++; break; }
  }                             /* check for maximum number of atts. */
  if ((maxsel > -3)             /* if to do an actual selection */
  &&  (bavg   <= 0)) {          /* and result no better than random */
    as_setmark(attset, -1);     /* remove all selected attributes */
    att_setmark(as_att(attset, trgid), 0);
  }
  if (nbc) nbc_delete(nbc, 0);  /* delete the Bayes classifiers */
  if (fbc) fbc_delete(fbc, 0);  /* (whichever was created) */
  fprintf(stderr, "attribute selection [%"ATTID_FMT" attribute(s)]", k);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  free(preds);                  /* delete the prediction array */
  return 0;                     /* return 'ok' */
}  /* fwdsel() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (nbc)    nbc_delete(nbc, 0);   \
  if (fbc)    fbc_delete(fbc, 0);   \
  if (attset) as_delete(attset);    \
  if (table)  tab_delete(table, 0); \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static int read_table (ATTSET *attset, TABREAD *tread,
                       CCHAR *fn_tab, CCHAR *fn_hdr, int mode,
                       TABLE **tab)
{                               /* --- read a table file */
  int     k;                    /* return value of read function */
  ATTID   m;                    /* number of attributes */
  TPLID   n;                    /* number of data tuples */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* for time measurements */

  /* --- read table header --- */
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0) return E_FOPEN;
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) return -k;       /* read table header */
    trd_close(tread);           /* and close the file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- read table --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0) return E_FOPEN;
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  *tab = tab_create("table", attset, tpl_delete);
  if (!*tab) return E_NOMEM;    /* create a data table */
  k = tab_read(*tab, tread, mode);
  if (k < 0) return -k;         /* read the table file */
  trd_close(tread);             /* and close the file */
  m = tab_attcnt(*tab);         /* get the number of attributes */
  n = tab_tplcnt(*tab);         /* and the number of data tuples */
  w = tab_tplwgt(*tab);         /* and print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  return 0;                   /* return 'ok' */
}  /* read_table() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter, buffer */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_dom  = NULL;      /* name of domain     file */
  CCHAR   *fn_hdr  = NULL;      /* name of header     file */
  CCHAR   *fn_tab  = NULL;      /* name of table      file */
  CCHAR   *fn_val  = NULL;      /* name of validation data file */
  CCHAR   *fn_bc   = NULL;      /* name of classifier file */
  CCHAR   *recseps = NULL;      /* record     separators */
  CCHAR   *fldseps = NULL;      /* field      separators */
  CCHAR   *blanks  = NULL;      /* blank      characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  CCHAR   *comment = NULL;      /* comment    characters */
  CCHAR   *trgname = NULL;      /* name of the target/class attribute */
  int     full     = 0;         /* flag for a full Bayes classifier */
  int     mode     = AS_ATT|AS_NOXATT;     /* table file read mode */
  int     onfly    = 0;         /* flag for "on the fly" processing */
  int     balance  = 0;         /* flag for balancing class freqs. */
  int     simp     = 0;         /* flag for classifier simplification */
  double  corr     = 0;         /* Laplace correction value */
  double  reg[]    = {-1e-2,1}; /* Tikhonov regularization values */
  int     setup    = 0;         /* setup/induction mode */
  ATTID   maxsel   = -1;        /* maximum number of atts. to select */
  int     verbose  = 0;         /* flag for verbose reporting */
  int     desc     = 0;         /* description mode */
  int     maxlen   = 0;         /* maximal output line length */
  ATTID   trgid;                /* id of target/class attribute */
  ATT     *att;                 /* (target) attribute */
  ATTID   *perm;                /* permutation of attributes */
  ATTID   m, r, z;              /* number of attributes */
  TPLID   n;                    /* number of data tuples */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] domfile [-d|-h hdrfile] "
                     "tabfile bcfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-V       full Bayes classifier (covariances)    "
                    "(default: naive Bayes)\n");
    printf("-c#      name of class attribute/field          "
                    "(default: last attribute)\n");
    printf("-q#      balance class frequencies              "
                    "(default: as in input)\n");
    printf("         (l: lower, b: boost, s: shift tuple weights;\n");
    printf("          uppercase letters: use integer factors)\n");
    printf("-s#      simplify classifier (naive Bayes only) "
                    "(default: all attributes)\n"
           "         (a: by adding attributes, "
                     "r: by removing attributes)\n");
    printf("-L#      Laplace correction                     "
                    "(default: %g)\n", corr);
    printf("-t#[:#]  Tikhonov regularization parameter      "
                    "(default: %g:%g)\n", reg[0], reg[1]);
    printf("         (negative first value: use only on failure;\n");
    printf("          second value limits regularization loop)\n");
    printf("-T       distribute tuple weight for unknowns   "
                    "(default: ignore)\n");
    printf("-e       use maximum likelihood estimate        "
                    "(default: unbiased)\n");
    printf("-F       work \"on the fly\" if possible          "
                    "(default: load data)\n");
    printf("-v#      file name of validation data set       "
                    "(default: none)\n");
    printf("         (for a forward selection of attributes,\n");
    printf("         based on AUC, for only two classes)\n");
    printf("-m#      maximum number of attributes to select "
                    "(default: no limit)\n");
    printf("-a       print verbose progress messages\n");
    printf("-p       print relative frequencies (percent)   "
                    "(default: only absolute)\n");
    printf("-l#      output line length                     "
                    "(default: no limit)\n");
    printf("-r#      record     separators                  "
                    "(default: \"\\n\")\n");
    printf("-f#      field      separators                  "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank      characters                  "
                    "(default: \" \\t\\r\")\n");
    printf("-u#      null value characters                  "
                    "(default: \"?*\")\n");
    printf("-C#      comment    characters                  "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("domfile  file containing domain descriptions\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header  "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("bcfile   file to write Bayes classifier to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse the arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'V': full    = 1;                     break;
          case 'c': optarg  = &trgname;              break;
          case 'q': balance = (*s) ? *s++ : 0;       break;
          case 's': simp    = (*s) ? *s++ : 0;       break;
          case 'L': corr    =      strtod(s, &s);    break;
          case 't': getdblvec(s, &s, 2, reg);        break;
          case 'T': setup  |= NBC_DWNULL;            break;
          case 'e': setup  |= NBC_MAXLLH;            break;
          case 'F': onfly   = 1;                     break;
          case 'v': optarg  = &fn_val;               break;
          case 'm': maxsel  = (int)strtol(s, &s, 0); break;
          case 'a': verbose = 1;                     break;
          case 'p': desc   |= NBC_REL;               break;
          case 'l': maxlen  = (int)strtol(s, &s, 0); break;
          case 'r': optarg  = &recseps;              break;
          case 'f': optarg  = &fldseps;              break;
          case 'b': optarg  = &blanks;               break;
          case 'u': optarg  = &nullchs;              break;
          case 'C': optarg  = &comment;              break;
          case 'n': mode   |= AS_WEIGHT;             break;
          case 'd': mode   |= AS_DFLT;               break;
          case 'h': optarg  = &fn_hdr;               break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_bc  = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note file names */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if (k != 3) error(E_ARGCNT);  /* and the number of arguments */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" for consistency */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_dom || !*fn_dom) i++;
  if  (!fn_tab || !*fn_tab) i++;/* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if      (simp == 'a') setup |= NBC_ADD;
  else if (simp == 'r') setup |= NBC_REMOVE;
  else if (simp !=  0 )         /* check simplification mode */
    error(E_SIMP, simp);        /* (must be 'add' or 'remove') */
  if ((        balance  !=  0)  && (tolower(balance) != 'l')
  &&  (tolower(balance) != 'b') && (tolower(balance) != 's'))
    error(E_BALANCE, balance);  /* check the balancing mode */
  if (corr < 0) error(E_LCORR, corr); /* check Laplace correction */
  if (fn_val) simp = onfly = 0; /* ensure consistent parameters */
  if (!full && (balance || simp)) onfly = 0;
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse domain descriptions --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the domain file */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_dom) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if ((as_parse(attset, scan, AT_ALL, 1) != 0)
  ||  !scn_eof(scan, 1))        /* parse domain descriptions */
    error(E_PARSE, scn_name(scan));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  m    = as_attcnt(attset);     /* get the number of attributes */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", m);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  if (m <= 0) error(E_ATTCNT);  /* check for at least one attribute */

  /* --- get and check target attribute --- */
  trgid = as_target(attset, trgname, 1);
  if (trgid < 0) error(trgname ? E_UNKCLS : E_MULCLS, trgname);
  att = as_att(attset, trgid);  /* check for a nominal target att. */
  if (att_type(att) != AT_NOM) error(E_CLSTYPE, att_name(att));

  /* --- build naive Bayes classifier --- */
  if (onfly) {                  /* if to build on the fly */
    /* --- create Bayes classifier --- */
    if (full) fbc = fbc_create(attset, trgid);
    else      nbc = nbc_create(attset, trgid);
    if (!fbc && !nbc) error(E_NOMEM);

    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* configure the characters */
    trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- process data tuples --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, mode); /* read/generate table header */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
    if (!(i & AS_DFLT))         /* if not done yet, read first tuple */
      k = as_read(attset, tread, mode);
    for (w = 0, n = 0; k == 0; n++) {
      w += as_getwgt(attset);   /* sum the tuple weights */
      k = (fbc) ? fbc_add(fbc, NULL) : nbc_add(nbc, NULL);
      if (k) error(E_NOMEM);    /* process the current tuple */
      k = as_read(attset, tread, mode);
    }                           /* try to read the next tuple */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_delete(tread, 1);       /* close the input file and */
    tread = NULL;               /* delete the table reader */
    if (fbc)                    /* setup classifer and mark atts. */
         { fbc_setup(fbc, setup); m = fbc_mark(fbc); }
    else { nbc_setup(nbc, setup); m = nbc_mark(nbc); }
    fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s),", m);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a success message */

  else {                        /* if to load the data */
    /* --- read data table(s) --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* configure the characters */
    trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
    k = read_table(attset, tread, fn_tab, fn_hdr, mode, &table);
    if (k == E_NOMEM) error(E_NOMEM);
    if (k == E_FOPEN) error(E_FOPEN, trd_name(tread));
    if (k >  0)       error(k, as_errmsg(attset, NULL, 0));
    if (fn_val) {               /* if validation data is given */
      k = read_table(attset, tread, fn_val, fn_hdr, mode, &valid);
      if (k == E_NOMEM) error(E_NOMEM);
      if (k == E_FOPEN) error(E_FOPEN, trd_name(tread));
      if (k >  0)       error(k, as_errmsg(attset, NULL, 0));
    }                           /* read the valdation data */
    trd_delete(tread, 1);       /* read the table body and */
    tread = NULL;               /* delete the table reader */

    /* --- check target attribute --- */
    if (fn_val && (att_valcnt(att) > 2))
      error(E_CLSBIN, att_name(att));

    /* --- reduce and balance table --- */
    t = clock();                /* start timer, print log message */
    fprintf(stderr, "reducing%s table ... ",
                    (balance) ? " and balancing" : "");
    n = tab_reduce(table);      /* reduce table for speed up */
    if (balance) {              /* if a balancing mode is given */
      k = (tolower(balance) == 'l') ? -2
        : (tolower(balance) == 'b') ? -1 : 0;
      i = (tolower(balance) != balance);
      w = tab_balance(table, trgid, k, NULL, i);
      if (w < 0) error(E_NOMEM);/* balance the class frequencies */
    }                           /* and get the new weight sum */
    fprintf(stderr, "[%"TPLID_FMT, n);
    w = tab_getwgt(table, 0, n);/* get the total tuple weight */
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

    /* --- forward selection of attributes --- */
    if (valid) {                /* if to select attributes */
      setup &= ~(NBC_ADD|NBC_REMOVE);
      fwdsel(table, valid, trgid, full,
             setup, corr, reg, maxsel, verbose);
      /* The result of the attribute selection is represented  */
      /* by attribute markers in the underlying attribute set. */
      tab_delete(valid, 0);     /* delete the validation data */
      valid = NULL;             /* and clear the variable */
      perm  = (ATTID*)malloc((size_t)m *sizeof(ATTID));
      if (!perm) error(E_NOMEM);/* create a permutation array */
      for (r = m; --r >= 0; ) { /* traverse the attributes */
        z = att_getmark(as_att(attset, r));
        if      (z > 0) perm[z-1] = r;
        else if (z < 0) perm[--m] = r;
      }                         /* collect destinations of columns */
      perm[--m] = trgid;        /* store target attribute and */
      trgid = m;                /* get the new target index */
      setup |= (full) ? FBC_MARKED : NBC_MARKED;
      tab_colperm(table, perm); /* move selected attributes */
      free(perm);               /* to the front of the table */
    }                           /* and delete permutation array */

    /* --- build (final) classifier --- */
    t = clock();                /* start timer, print log message */
    fprintf(stderr, "building %sclassifier ... ",
                    (fn_val) ? "final " : "");
    if (full) {                 /* if full Bayes classifier */
      fbc = fbc_induce(table, trgid, setup);
      if (!fbc) error(E_NOMEM); /* induce a full Bayes classifier */
      fbc_setup(fbc, setup);    /* and set it up for execution */
      if (!fn_val) m = fbc_mark(fbc); }
    else {                      /* if naive Bayes classifier */
      nbc = nbc_induce(table, trgid, setup);
      if (!nbc) error(E_NOMEM); /* induce a naive Bayes classifier */
      nbc_setup(nbc, setup);    /* and set it up for execution */
      if (!fn_val) m = nbc_mark(nbc);
    }                           /* mark the used attributes */
    fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s)]", m);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- describe created classifier --- */
  t = clock();                  /* start timer, open output file */
  if (fn_bc && (strcmp(fn_bc, "-")) == 0) fn_bc = "";
  if (fn_bc && *fn_bc) { out = fopen(fn_bc, "w"); }
  else                 { out = stdout; fn_bc = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_bc);
  if (!out) error(E_FOPEN, fn_bc);
  k = (full || simp || fn_val)  /* print only the class and */
    ? AS_MARKED : 0;            /* the marked attributes */
  if (fn_val) k |= AS_WEIGHT;   /* print AUC values of attributes */
  if (as_desc(attset, out, k|AS_TITLE|AS_IVALS, maxlen) != 0)
    error(E_FWRITE, fn_bc);     /* describe attribute domains */
  fputc('\n', out);             /* leave one line empty */
  desc |= (nbc && simp) ? NBC_MARKED : 0;
  if (fn_val) desc |= (full) ? FBC_MARKED : NBC_MARKED;
  if ((fbc ? fbc_desc(fbc, out, desc|FBC_TITLE|FBC_INFO, maxlen)
  :          nbc_desc(nbc, out, desc|NBC_TITLE|NBC_INFO, maxlen)) != 0)
    error(E_FWRITE, fn_bc);     /* describe the Bayes classifier */
  if (out && (((out == stdout) ? fflush(out) : fclose(out)) != 0))
    error(E_FWRITE, fn_bc);     /* close the output file and */
  out = NULL;                   /* print a success message */
  fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s)]", m);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
