/*----------------------------------------------------------------------
  File    : bcx.c
  Contents: naive and full Bayes classifier execution
  Author  : Christian Borgelt
  History : 1998.12.08 file created from file dtc.c
            1999.02.13 input from stdin, output to stdout added
            1999.04.17 simplified using the new module 'io'
            2000.11.20 option -m (max. likelihood est. for var.) added
            2000.11.21 adapted to redesigned module nbayes
            2000.11.30 full Bayes classifier execution added
            2001.07.16 adapted to modified module scanner
            2002.09.18 bug concerning missing target fixed
            2003.02.02 bug in alignment in connection with -d fixed
            2003.04.23 missing AS_MARKED added for second reading
            2003.08.16 slight changes in error message output
            2005.02.22 classification threshold added (option -t)
            2006.01.17 format specification for confidence added
            2007.02.13 adapted to modified module attset
            2009.06.04 optional output of typicality for class added
            2010.10.12 adapted to modified module tabread
            2011.02.02 adapted to modified scanner module and as_parse()
            2011.07.28 adapted to modified attset and utility modules
            2011.12.15 processing without table reading improved
            2013.06.08 missing flag AS_MARKED added to table read mode
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.20 output format changed to significant digits
            2014.12.10 Tikhonov regularization parameter added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "params.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#ifndef TAB_WRITE
#define TAB_WRITE
#endif
#include "table.h"
#ifndef NBC_PARSE
#define NBC_PARSE
#endif
#include "nbayes.h"
#ifndef FBC_PARSE
#define FBC_PARSE
#endif
#include "fbayes.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

#if ATTID_MAX < VALID_MAX
#error "bcx requires ATTID_MAX >= VALID_MAX"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "bcx"
#define DESCRIPTION "naive and full Bayes classifier execution"
#define VERSION     "version 3.5 (2021.12.05)         " \
                    "(c) 1998-2021   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse error on classifier file */
#define E_LCORR     (-10)       /* invalid Laplace correction */
#define E_OUTPUT    (-11)       /* class in input or write output */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- classification result info.--- */
  ATT   *att;                   /* class attribute */
  VALID  cnt;                   /* number of classes */
  VALID  class;                 /* class (classification result) */
  CCHAR *col_cls;               /* name  of classification column */
  int    cwd_cls;               /* width of classification column */
  double prob;                  /* probability of result class */
  CCHAR *col_prob;              /* name   of probability column */
  int    cwd_prob;              /* width  of probability column */
  int    dig_prob;              /* digits of probability column */
  double typi;                  /* typicality of data for class */
  CCHAR *col_typi;              /* name   of typicality column */
  int    cwd_typi;              /* width  of typicality column */
  int    dig_typi;              /* digits of typicality column */
  int    all;                   /* whether to show all probabilities */
  int    cwd_all;               /* width  of probability columns */
  int    err;                   /* classification error indicator */
} RESULT;                       /* (classification result info.) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_LCORR   -10 */  "invalid Laplace correction %g",
  /* E_OUTPUT  -11 */  "must have class as input or write output",
  /*           -12 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner (for bayes classifier) */
static TABREAD  *tread  = NULL; /* table reader */
static TABWRITE *twrite = NULL; /* table writer */
static ATTSET   *attset = NULL; /* attribute set */
static TABLE    *table  = NULL; /* data table */
static NBC      *nbc    = NULL; /* naive Bayes classifier */
static FBC      *fbc    = NULL; /* full  Bayes classifier */
static RESULT   res     = {     /* classification result information */
  NULL, 0,                      /* class attribute and # of classes */
  0,   "bc", 0,                 /* data for classification column */
  0.0, NULL, 0, 3,              /* data for probability    column */
  0.0, NULL, 0, 6,              /* data for typicality     column */
  0, 0,                         /* flag for all probabilities */
  0 };                          /* classification error indicator */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (nbc)    nbc_delete(nbc, 0);    \
  if (fbc)    fbc_delete(fbc, 0);    \
  if (attset) as_delete(attset);     \
  if (tread)  trd_delete(tread,  1); \
  if (twrite) twr_delete(twrite, 1); \
  if (scan)   scn_delete(scan,   1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void classify (double thresh)
{                               /* --- classify the current tuple */
  VALID c;                      /* buffer for true class */

  assert(nbc || fbc);           /* check for a classifier */
  res.class = (fbc)             /* execute the Bayes classifier */
            ? fbc_exec(fbc, NULL, &res.prob, &res.typi)
            : nbc_exec(nbc, NULL, &res.prob, &res.typi);
  if (res.cnt <= 2) {           /* if this is a two class problem */
    if (res.class <= 0) {       /* check and adapt class 0 result */
      if (res.prob <   thresh) {
        res.class = 1; res.prob = 1 -res.prob; } }
    else {                      /* check and adapt class 1 result */
      if (res.prob < 1-thresh) {
        res.class = 0; res.prob = 1 -res.prob; }
    }                           /* (classify as class 0 if prob. */
  }                             /* of this class is >= threshold) */
  c = att_inst(res.att)->n;     /* compute the error indicator */
  res.err = (!isnull(c) && (res.class != c)) ? 1 : 0;
}  /* classify() */

/*--------------------------------------------------------------------*/

static void infout (ATTSET *set, TABWRITE *twrite, int mode)
{                               /* --- write additional information */
  int    n, k;                  /* character counters */
  VALID  c;                     /* loop variable for classes */
  double p;                     /* buffer for probability */

  assert(set && twrite);        /* check the function arguments */
  if (mode & AS_ATT) {          /* if to write the header */
    twr_puts(twrite, res.col_cls);  /* write class column name */
    if ((mode & AS_ALIGN)       /* if to align the column */
    && ((mode & AS_WEIGHT) || res.col_prob || res.col_typi || res.all)){
      n = (int)strlen(res.col_cls);
      k = att_valwd(res.att, 0);
      res.cwd_cls = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
      if (k > n) twr_pad(twrite, (size_t)(k-n));
    }                           /* compute width of class column */
    if (res.col_prob) {         /* if to write a class probability */
      twr_fldsep(twrite);       /* write a field separator and */
      twr_puts(twrite, res.col_prob);       /* the column name */
      if ((mode & AS_ALIGN)     /* if to align the column */
      && ((mode & AS_WEIGHT) || res.col_typi || res.all)) {
        n = (int)strlen(res.col_prob);
        k = res.dig_prob +3;    /* compute width of prob. column */
        res.cwd_prob = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
        if (k > n) twr_pad(twrite, (size_t)(k-n));
      }                         /* pad with blanks if requested */
    }
    if (res.col_typi) {         /* if to write a class typicality */
      twr_fldsep(twrite);       /* write a field separator and */
      twr_puts(twrite, res.col_typi);       /* the column name */
      if ((mode & AS_ALIGN)     /* if to align the column */
      && ((mode & AS_WEIGHT) || res.all)) {
        n = (int)strlen(res.col_typi);
        k = res.dig_typi +3;    /* compute width of typi. column */
        res.cwd_typi = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
        if (k > n) twr_pad(twrite, (size_t)(k-n));
      }                         /* pad with blanks if requested */
    }
    if (res.all) {              /* if to write all probabilities */
      if (mode & AS_ALIGN) {    /* if to align the columns */
        n = att_valwd(res.att, 0);
        k = res.dig_prob +3;    /* compute width of prob. columns */
        res.cwd_all = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
      }
      for (c = 0; c < res.cnt; c++) {
        twr_fldsep(twrite);     /* traverse the target values */
        n = twr_printf(twrite, "%s", att_valname(res.att, c));
        if (res.cwd_all > n) twr_pad(twrite, (size_t)(res.cwd_all-n));
      }                         /* print the value name and */
    } }                         /* pad with blanks if requested */
  else {                        /* if to write a normal record */
    n = twr_printf(twrite, "%s", att_valname(res.att, res.class));
    if (res.cwd_cls > n) twr_pad(twrite, (size_t)(res.cwd_cls-n));
    if (res.col_prob) {         /* if to write a class probability */
      twr_fldsep(twrite);       /* write separator and probability */
      n = twr_printf(twrite, "%.*g", res.dig_prob, res.prob);
      if (res.cwd_prob > n) twr_pad(twrite, (size_t)(res.cwd_prob-n));
    }                           /* if to align, pad with blanks */
    if (res.col_typi) {         /* if to write a class probability */
      twr_fldsep(twrite);       /* write separator and probability */
      n = twr_printf(twrite, "%.*g", res.dig_typi, res.typi);
      if (res.cwd_typi > n) twr_pad(twrite, (size_t)(res.cwd_typi-n));
    }                           /* if to align, pad with blanks */
    if (res.all) {              /* if to write all probabilities */
      for (c = 0; c < res.cnt; c++) {
        twr_fldsep(twrite);     /* traverse the target values */
        p = (fbc) ? fbc_post(fbc, c) : nbc_post(nbc, c);
        n = twr_printf(twrite, "%.*g", res.dig_prob, p);
        if (res.cwd_all > n) twr_pad(twrite, (size_t)(res.cwd_all-n));
      }                         /* print the probability and */
    }                           /* pad with blanks if requested */
  }
}  /* infout() */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, buffer */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_bc   = NULL;      /* name of classifier file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *recseps = NULL;      /* record     separators */
  CCHAR   *fldseps = NULL;      /* field      separators */
  CCHAR   *blanks  = NULL;      /* blank      characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  CCHAR   *comment = NULL;      /* comment    characters */
  double  corr     = -INFINITY; /* Laplace correction value */
  double  reg[2]   = {-INFINITY,-INFINITY}; /* regularization value */
  double  thresh   = 0.5;       /* classification threshold */
  int     dwnull   = 0;         /* distribute weight of null values */
  int     maxllh   = 0;         /* max. likelihood est. of variance */
  int     mode     = AS_ATT|AS_MARKED; /* table file read  mode */
  int     mout     = AS_ATT;           /* table file write mode */
  double  errs     = 0.0;       /* number of misclassifications */
  ATTID   m;                    /* number of attributes */
  TPLID   n, r;                 /* number of data tuples */
  double  w, u;                 /* weight of data tuples, buffer */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] bcfile [-d|-h hdrfile] "
                     "tabfile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-p#      predicition/classification field name  "
                    "(default: \"%s\")\n", res.col_cls);
    printf("-c#      confidence/probability field name      "
                    "(default: no confidence)\n");
    printf("-o#      significant digits for confidence      "
                    "(default: %d)\n", res.dig_prob);
    printf("-x       print extended confidence information  "
                    "(all probabilities)\n");
    printf("-y#      typicality for class field name        "
                    "(default: no typicality)\n");
    printf("-z#      significant digits for typicality      "
                    "(default: %d)\n", res.dig_typi);
    printf("-L#      Laplace correction                     "
                    "(default: from classifier)\n");
    printf("-t#[:#]  Tikhonov regularization parameter      "
                    "(default: %g:%g)\n", reg[0], reg[1]);
    printf("         (negative first value: use only on failure;\n");
    printf("          second value limits regularization loop)\n");
    printf("-T#      class probability threshold            "
                    "(default: %g)\n", thresh);
    printf("         (only applicable if there are "
                    "just two classes)\n");
    printf("-v/V     (do not) distrib. weight for unkowns   "
                    "(default: ignore)\n");
    printf("-m/M     (do not) use maximum likelihood estim. "
                    "(default: unbiased)\n");
    printf("-a       align fields in output table           "
                    "(default: single separator)\n");
    printf("-w       do not write field names to the output file\n");
    printf("-r#      record     separators                  "
                    "(default: \"\\n\")\n");
    printf("-f#      field      separators                  "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank      characters                  "
                    "(default: \" \\t\\r\")\n");
    printf("-u#      null value characters                  "
                    "(default: \"?*\")\n");
    printf("-C#      comment    characters                  "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("bcfile   file containing classifier description\n");
    printf("-d       use default table header "
                    "(field names = field numbers)\n");
    printf("-h       read table header (field names) from hdrfile\n");
    printf("hdrfile  file containing table header (field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write output table to (optional)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate option */
          case 'p': optarg  = &res.col_cls;               break;
          case 'c': optarg  = &res.col_prob;              break;
          case 'y': optarg  = &res.col_typi;              break;
          case 'o': res.dig_prob = (int)strtol(s, &s, 0); break;
          case 'z': res.dig_typi = (int)strtol(s, &s, 0); break;
          case 'x': res.all = 1;                          break;
          case 'L': corr    = strtod(s, &s);              break;
          case 't': getdblvec(s, &s, 2, reg);             break;
          case 'T': thresh  = strtod(s, &s);              break;
          case 'v': dwnull  = 0x8000;                     break;
          case 'V': dwnull |= 0x8000|NBC_DWNULL;          break;
          case 'm': maxllh  = 0x8000;                     break;
          case 'M': maxllh |= 0x8000|NBC_MAXLLH;          break;
          case 'n': mode   |= AS_WEIGHT;
                    mout   |= AS_WEIGHT;                  break;
          case 'a': mout   |= AS_ALIGN;                   break;
          case 'w': mout   &= ~AS_ATT;                    break;
          case 'r': optarg  = &recseps;                   break;
          case 'f': optarg  = &fldseps;                   break;
          case 'b': optarg  = &blanks;                    break;
          case 'u': optarg  = &nullchs;                   break;
          case 'C': optarg  = &comment;                   break;
          case 'd': mode   |= AS_DFLT;                    break;
          case 'h': optarg  = &fn_hdr;                    break;
          default : error(E_OPTION, *--s);                break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_bc  = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if ((k < 2) || (k > 3))       /* and the number of arguments */
    error(E_ARGCNT);            /* (output file is optional) */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_bc  || !*fn_bc)  i++;
  if  (!fn_tab || !*fn_tab) i++;/* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if ((corr < 0) && (corr > -INFINITY))
    error(E_LCORR, corr);       /* check the Laplace correction */
  if ((mout & AS_ATT) && (mout & AS_ALIGN))
    mout |= AS_ALNHDR;          /* set align to header flag */
  if (fn_out) mout |= AS_MARKED|AS_INFO1|AS_RDORD;
  else        mout  = 0;        /* set up the table write mode */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read Bayes classifier --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the classifier file */
  t = clock();                  /* start timer, open intput file */
  if (scn_open(scan, NULL, fn_bc) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
  if ((scn_token(scan) == T_ID)     /* determine classifier type */
  &&  (strcmp(scn_value(scan), "fbc") == 0))
       fbc = fbc_parse(attset, scan);
  else nbc = nbc_parse(attset, scan);
  if ((!fbc && !nbc)            /* parse the Bayes classifier */
  ||   !scn_eof(scan, 1))       /* and check for end of file */
    error(E_PARSE, scn_name(scan));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  if ((corr >= 0) || (reg[0] > -INFINITY) || (reg[1] > -INFINITY)
  ||  dwnull || maxllh) {       /* if parameters specified */
    if (corr < 0)               /* Laplace correction */
      corr = (fbc) ? fbc_getcorr(fbc) : nbc_getcorr(nbc);
    k      = (fbc) ? fbc_getmode(fbc) : nbc_getmode(nbc);
    if (reg[0] <= -INFINITY)    /* Tikhonov regularization */
      reg[0] = (fbc) ? fbc_getreg (fbc) : 0.0;
    if (reg[1] <= -INFINITY)    /* get the classifier's parameters */
      reg[1] = (fbc) ? fbc_getlim (fbc) : 0.0;
    if (dwnull & 0x8000) k = (k & ~NBC_DWNULL) | (dwnull & ~0x8000);
    if (maxllh & 0x8000) k = (k & ~NBC_MAXLLH) | (maxllh & ~0x8000);
    if (fbc) { fbc_setmode(fbc, k); fbc_setup(fbc, 0); }
    else     { nbc_setmode(nbc, k); nbc_setup(nbc, 0); }
  }                             /* set up with new parameters */

  /* --- get class attribute --- */
  res.att = (fbc) ? fbc_trgatt(fbc) : nbc_trgatt(nbc);
  res.cnt = att_valcnt(res.att);/* get the class attribute */
  as_setmark(attset, 1);        /* mark all attributes */
  att_setmark(res.att, 0);      /* except the class attribute */

  /* --- read table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- process table body --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  if (mout & AS_ALIGN) {        /* if to align the output columns */
    table = tab_create("table", attset, tpl_delete);
    if (!table) error(E_NOMEM); /* read the data table */
    k = tab_read(table, tread, mode);
    if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
    trd_delete(tread, 1);       /* delete the table reader */
    tread = NULL;               /* and clear the variable */
    m = tab_attcnt(table);      /* get the number of attributes */
    n = tab_tplcnt(table);      /* and the number of data tuples */
    w = tab_tplwgt(table);      /* and print a success message */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* configure the characters */
    twr_xchars(twrite, recseps, fldseps, blanks, nullchs);
    t = clock();                /* start timer, open output file */
    if (twr_open(twrite, NULL, fn_out) != 0)
      error(E_FOPEN, twr_name(twrite));
    fprintf(stderr, "writing %s ... ", twr_name(twrite));
    if ((mout & AS_ATT)         /* write a table header */
    &&  (as_write(attset, twrite, mout, infout) != 0))
      error(E_FWRITE, twr_name(twrite));
    mout = AS_INST | (mout & ~AS_ATT);
    m += (res.col_prob ? 2 : 1) +(res.col_typi ? 1 : 0)
      +  (res.all ? res.cnt: 0);/* add the output columns */
    for (r = 0; r < n; r++) {   /* traverse the tuples */
      tpl_toas(tab_tpl(table, r));
      classify(thresh);         /* classify the current tuple */
      u = as_getwgt(attset);    /* get the tuple weight and */
      errs += res.err *u;       /* count the misclassifications */
      if (as_write(attset, twrite, mout, infout) != 0)
        error(E_FWRITE, twr_name(twrite));
    } }                         /* write the current tuple */
  else {                        /* if to process tuples directly */
    k = as_read(attset, tread, mode); /* read/generate table header */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    if (!fn_out && (att_getmark(res.att) < 0))
      error(E_OUTPUT);          /* check for output to produce */
    if (fn_out) {               /* if to write an output file */
      twrite = twr_create();    /* create a table writer and */
      if (!twrite) error(E_NOMEM);   /* configure characters */
      twr_xchars(twrite, recseps, fldseps, blanks, nullchs);
      t = clock();              /* start timer, open output file */
      if (twr_open(twrite, NULL, fn_out) != 0)
        error(E_FOPEN, twr_name(twrite));
      if ((mout & AS_ATT)       /* write a table header */
      &&  (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      mout = AS_INST | (mout & ~AS_ATT);
    }                           /* remove the attribute flag */
    i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
    if (!(i & AS_DFLT))         /* if not done yet, read first tuple */
      k = as_read(attset, tread, mode);
    for (w = 0, n = 0; k == 0; n++) {
      classify(thresh);         /* classify the current tuple */
      w    += u = as_getwgt(attset);   /* sum the tuple weights */
      errs += res.err *u;       /* count the misclassifications */
      if (twrite                /* write the current tuple */
      && (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      k = as_read(attset, tread, mode);
    }                           /* try to read the next tuple */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_delete(tread, 1);       /* delete the table reader */
    tread = NULL;               /* and clear the variable */
    m = as_attcnt(attset);      /* get the number of attributes */
  }
  if (twrite) {                 /* if an output file was written */
    if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* close the output file and */
    twrite = NULL;              /* delete the table writer */
  }                             /* print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- print error statistics --- */
  if (att_getmark(res.att) >= 0) { /* if the target is present */
    fprintf(stderr, "%g error(s) ", errs);
    fprintf(stderr, "(%.2f%%)\n", (w > 0) ? 100.0*(errs/w) : 0);
  }                             /* print number of misclassifications */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
