/*----------------------------------------------------------------------
  File    : corr.c
  Contents: covariances/correlation coefficients computation program
  Author  : Christian Borgelt
  History : 1999.04.07 file created from file xmat.c
            1999.04.10 first version completed
            1999.04.11 options -x, -v, and -c added
            1999.04.14 option -k (expected values from known pairs)
            1999.12.01 bug in connection with option -d removed
            2000.11.10 adapted to new module mvnorm
            2002.06.07 LaTeX output option added
            2003.08.16 slight changes in error message output
            2004.04.22 bug in function dblout fixed (log(0))
            2007.02.13 adapted to modified module tabread
            2010.10.12 adapted to modified module tabread
            2011.02.02 adapted to modified scanner module and as_parse()
            2011.07.26 adapted to modified interface of st_create()
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
            2014.10.24 changed from LGPL license to MIT license
            2015.06.06 more output format options added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include "symtab.h"
#include "tabread.h"
#include "mvnorm.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "corr"
#define DESCRIPTION "compute covariance matrix/correlation coefficients"
#define VERSION     "version 3.3 (2016.11.11)         " \
                    "(c) 1999-2016   Christian Borgelt"

/* --- sizes --- */
#define BLKSIZE      32         /* block size for column vector */

/* --- output modes --- */
#define PLAIN         0         /* plain text output */
#define LOWER         1         /* print (only) lower triangle */
#define LEFT          1         /* print (only) left  triangle */
#define UPPER         2         /* print (only) upper triangle */
#define RIGHT         2         /* print (only) right triangle */
#define FULL          (LEFT|RIGHT)
#define TABS          4         /* use tabulators in plain text */
#define LATEX         8         /* produce output in LaTeX code */

/* --- error codes --- */
#define E_NONE        0         /* no error */
#define E_NOMEM     (-1)        /* not enough memory */
#define E_FOPEN     (-2)        /* file open failed */
#define E_FREAD     (-3)        /* file read failed */
#define E_FWRITE    (-4)        /* file write failed */
#define E_STDIN     (-5)        /* double assignment of stdin */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_UNKATT    (-9)        /* unknown   attribute */
#define E_DUPATT   (-10)        /* duplicate attribute */
#define E_MISATT   (-11)        /* missing   attribute */
#define E_EMPATT   (-12)        /* empty attribute name */
#define E_FLDCNT   (-13)        /* wrong number of fields */
#define E_WEIGHT   (-14)        /* invalid tuple weight */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- set of table columns --- */
  IDENT      size;              /* size of column vector */
  IDENT      cnt;               /* number of matrix columns */
  IDENT      valid;             /* number of valid matrix columns */
  size_t     max;               /* maximal length of a column name */
  const char **names;           /* table column names */
  double     *vals;             /* current values */
} COLSET;                       /* (set of table columns) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_UNKATT   -9 */  "unknown attribute '%s'",
  /* E_EMPATT  -12 */  "%s:%d(%d): empty attribute name",
  /* E_DUPATT  -10 */  "%s:%d(%d): duplicate attribute '%s'",
  /* E_FLDCNT  -13 */  "%s:%d(%d): wrong number of fields/columns",
  /* E_WEIGHT  -14 */  "%s:%d(%d): invalid tuple weight '%s'",
  /*           -15 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static TABREAD *tread  = NULL;  /* table reader */
static SYMTAB  *symtab = NULL;  /* symbol table */
static COLSET  *colset = NULL;  /* column descriptions */
static MVNORM  *mvnorm = NULL;  /* multivariate normal distribution */
static FILE    *out    = NULL;  /* output file */

/*----------------------------------------------------------------------
  Column Set Functions
----------------------------------------------------------------------*/
#define cs_create()         (COLSET*)calloc(1, sizeof(COLSET))
#define cs_colcnt(s)        ((s)->cnt)
#define cs_proc(s,m,w)      mvn_add (m, (s)->vals, w)
#define cs_procx(s,m,w)     mvn_addx(m, (s)->vals, w, 0)

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

static void cs_delete (COLSET *cset)
{                               /* --- delete a column set */
  if (cset->names)              /* if there is a names vector, */
    free((void*)cset->names);   /* delete it */
  if (cset->vals)               /* if there is a value vector, */
    free((void*)cset->vals);    /* delete it */
  free(cset);                   /* delete the column set body */
}  /* cs_delete() */

#endif
/*--------------------------------------------------------------------*/

static int cs_add (COLSET *cset, const char *name)
{                               /* --- add a column to a column set */
  IDENT n;                      /* new column vector size, buffer */
  void  *p;                     /* new names/values vector */

  assert(cset && name);         /* check the function arguments */
  n = cset->size;               /* get the column vector size and */
  if (cset->cnt >= n) {         /* if the column vector is full */
    n += (n > BLKSIZE) ? (n >> 1) : BLKSIZE;
    p = realloc((void*)cset->names, (size_t)n *sizeof(const char*));
    if (!p) return -1;          /* enlarge the names vector */
    cset->names = (const char**)p;
    p = realloc(cset->vals, (size_t)n *sizeof(double));
    if (!p) return -1;          /* enlarge the names vector */
    cset->vals  = (double*)p; cset->size = n;
  }                             /* set the new vectors and their size */
  cset->names[cset->cnt  ] = name;
  cset->vals [cset->cnt++] = MVN_NULL;
  return 0;                     /* note the table column's name */
}  /* cs_add() */               /* and return 'ok' */

/*--------------------------------------------------------------------*/

static void cs_set (COLSET *cset, IDENT colid, const char *name)
{                               /* --- set a column value */
  char   *s;                    /* end pointer for conversion */
  double *val;                  /* pointer to the value to set */

  assert(cset && (colid >= 0) && (colid < cset->cnt));
  if (!cset->names[colid])      /* if the column name has been */
    return;                     /* deleted, abort the function */
  val = cset->vals +colid;      /* get the value buffer */
  if (!name || !*name) {        /* if no value name is given, */
    *val = MVN_NULL; return; }  /* the value is null */
  *val = strtod(name, &s);      /* convert the attribute value */
  if (*s || (s == name)) {      /* if it is nominal */
    *val = MVN_NULL; cset->names[colid] = NULL; }
}  /* cs_set() */               /* invalidate the column */

/*--------------------------------------------------------------------*/

static void cs_prepare (COLSET *cset)
{                               /* --- prepare columns for output */
  IDENT      i;                 /* loop variable */
  size_t     k;                 /* length of a name */
  const char *name;             /* to traverse the column names */

  cset->valid = 0;              /* init. the number of valid columns */
  cset->max   = 9;              /* and the maximal length of a name */
  for (i = 0; i < cset->cnt; i++) {
    name = cset->names[i];      /* traverse the columns */
    if (!name) continue;        /* skip invalidated columns */
    cset->valid++;              /* count the valid columns */
    if ((k = strlen(name)) > cset->max) cset->max = k;
  }                             /* adapt the maximal name length */
}  /* cs_prepare() */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void dblout (FILE *out, double num, int len, int pad)
{                               /* --- print a floating point number */
  int  n, d;                    /* number of characters/decimals */
  char m[32], e[16];            /* mantissa and exponent */

  if (num >= 0.0) {  m[0] = ' '; }
  else { num = -num; m[0] = '-'; }
  len--;                        /* determine and store the sign */
  n = (num == 0)                /* calculate the decimal exponent */
    ? 0 : (int)floor(log10(num));
  if ((n > len) || (n <= -3)) { /* if an exponent is needed, */
    num /= pow(10, n);          /* comp. mantissa and note exponent */
    d = len -2 -(n = sprintf(e, "e%d", n)); }
  else {                        /* if no exponent is needed, */
    d = len -((n > 0) ? n+2 : 2);      /* compute the number */
    e[0] = 0; n = 0;            /* of decimal places possible */
  }                             /* and clear the exponent */
  n += sprintf(m+1, "%.*f", (d <= 0) ? 0 : d, num);
  if (pad)                      /* if to pad the output */
    while (n < len) {           /* format the mantissa, */
      fputc(' ', out); n++; }   /* fill to the requested length */
  fputs(((m[0] != ' ') || pad) ? m : m+1, out);
  fputs(e, out);                /* print mantissa and exponent */
}  /* dblout() */

/*----------------------------------------------------------------------
The above function is needed, because a format string used with the
function fprintf does not yield appropriate results.
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
  Output Functions
----------------------------------------------------------------------*/

static void expvar (COLSET *cset, MVNORM *mvn,
                    int mode, int cnt, FILE *out)
{                               /* --- print exp. values/variances */
  IDENT      i, n = 0;          /* loop variables */
  size_t     k;                 /* loop variable */
  const char *name;             /* to traverse the column names */
  double     t;                 /* expected value and variance */
  char       *s1, *s2, *s3;     /* separators */
  char       *end;              /* end of line */
  char       *fmt;              /* format for index output */
  int        pad = 1;           /* flag for padding output w/ blanks */

  assert(mvn && out);           /* check the function arguments */
  cs_prepare(cset);             /* prepare columns for output */
  if      (mode & LATEX) {      /* if to generate LaTeX code */
    fputs("\\begin{tabular}{|r|l|r|r|r|}\\hline\n", out);
    fputs("\\multicolumn{5}{|l|}", out);
    fputs("{attribute statistics\\rule{0pt}{2.4ex}} \\\\\n", out);
    fputs("\\hline\\hline\\rule{0pt}{2.4ex}%\n", out);
    fputs("no & attribute", out);  /* print a header */
    for (k = 9; k <  cset->max; k++) fputc(' ', out);
    fputs(" & exp. val. &  variance & std. dev.", out);
    fputs(" \\\\\n\\hline\\rule{0pt}{2.4ex}%\n", out);
    s1 = " & "; s2 = " &$"; s3 = "$&$"; end = "$ \\\\\n";
    fmt = "%3"IDENT_FMT; }
  else if (mode & TABS) {       /* if plain text with tabs */
    fputs("attribute statistics\n", out);
    fputs("no\tattribute\texp. val.\tvariance\tstd. dev.", out);
    if (cnt) fputs("\tcnt", out);
    fputc('\n', out);           /* print a header */
    s1 = s2 = s3 = "\t"; end = "\n"; pad = 0;
    fmt = "%"IDENT_FMT; }       /* configure the separators */
  else {                        /* if plain text with delimiters */
    fputs("attribute statistics\n", out);
    fputs(" no | attribute", out);  /* print a header */
    for (k = 9; k < cset->max; k++) fputc(' ', out);
    fputs(" | exp. val. |  variance | std. dev.", out);
    if (cnt) fputs(" |   cnt", out);
    fputs("\n----+-", out);
    for (k = 0; k < cset->max; k++) fputc('-', out);
    fputs("-+-----------+-----------+----------", out);
    if (cnt) fputs("-+------", out);
    s1 = s2 = s3 = " | "; end = "\n"; fmt = "%3"IDENT_FMT;
    fputc('\n', out);           /* configure the separators */
  }
  for (i = 0; i < cset->cnt; i++) {
    name = cset->names[i];      /* traverse the matrix columns, */
    if (!name) continue;        /* but skip invalidated columns */
    fprintf(out, fmt, ++n);     /* print the column number, */
    fputs(s1,   out);           /* the first separator */
    fputs(name, out);           /* and the column name */
    if (pad)                    /* if to pad the output */
      for (k = strlen(name); k < cset->max; k++)
        fputc(' ', out);        /* fill the output field */
    fputs(s2, out);             /* retrieve and */
    t = mvn_exp(mvn, i);        /* print the expected value */
    if (isnan(t)) fputs((pad) ? "        ?" : "?", out);
    else          dblout(out, t,       9, pad);
    t = mvn_var(mvn, i);        /* retrieve and */
    fputs(s3, out);             /* print the variance */
    if (t < 0)    fputs((pad) ? "        ?" : "?", out);
    else          dblout(out, t,       9, pad);
    fputs(s3, out);             /* print the standard deviation */
    if (t < 0)    fputs((pad) ? "        ?" : "?", out);
    else          dblout(out, sqrt(t), 9, pad);
    if (cnt) {                  /* if to print counters */
      fputs(s1, out);           /* print a separator */
      fprintf(out, (pad) ? "%5g" : "%g", mvn_cnt(mvn, i));
    }                           /* print the number of cases */
    fputs(end, out);            /* terminate the output line */
  }
  if (mode & LATEX)             /* if to generate LaTeX code */
    fputs("\\hline\n\\end{tabular}\n", out);
}  /* expvar() */

/*--------------------------------------------------------------------*/

static void covar (COLSET *cset, MVNORM *mvn, int mode, FILE *out)
{                               /* --- print covariance matrix */
  IDENT      x, y, n = 0;       /* loop variables */
  size_t     k;                 /* loop variable */
  const char *xname, *yname;    /* to traverse the column names */
  double     cov;               /* covariance */
  char       *s1, *s2, *s3, *d; /* separators, delimiter */
  char       *end;              /* end of line */
  char       *fmt;              /* format for index output */
  int        pad = 1;           /* flag for padding output w/ blanks */

  assert(cset && mvn && out);   /* check the function arguments */
  cs_prepare(cset);             /* prepare columns for output */
  if (mode & LATEX) {           /* if to generate LaTeX code */
    fprintf(out, "\\begin{tabular}"
                 "{|r|l|*{%"IDENT_FMT"}{r|}}", cset->valid);
    fputs("\\hline\n", out);    /* print a header */
    fprintf(out, "\\multicolumn{%"IDENT_FMT"}{|l|}", cset->valid +2);
    fputs("{covariance matrix\\rule{0pt}{2.4ex}} \\\\\n", out);
    fputs("\\hline\\hline\\rule{0pt}{2.4ex}%\n", out);
    fputs(" no & attribute", out);
    for (k = 9; k <  cset->max;   k++) fputc(' ', out);
    for (x = 1; x <= cset->valid; x++)
      fprintf(out, " & %8"IDENT_FMT, x);
    fputs(" \\\\\n\\hline\\rule{0pt}{2.4ex}%\n", out);
    s1 = " & "; s2 = " "; s3 = "&"; d = "$"; end = " \\\\\n";
    fmt = "%3"IDENT_FMT; }      /* configure the separators */
  else if (mode & TABS) {       /* if plain text with tabs */
    fputs("covariance matrix\n", out);
    fputs("no\tattribute", out);
    for (x = 1; x <= cset->valid; x++)
      fprintf(out, "\t%"IDENT_FMT, x);
    fputc('\n', out);           /* print a header */
    s1 = s3 = "\t"; s2 = d = ""; end = "\n"; pad = 0;
    fmt = "%"IDENT_FMT; }       /* configure the separators */
  else {                        /* if plain text with delimiters */
    fputs("covariance matrix\n", out);
    fputs(" no | attribute", out);
    for (k = 9; k <  cset->max;   k++) fputc(' ', out);
    fputs(" |", out);           /* print a header */
    for (x = 1; x <= cset->valid; x++)
      fprintf(out, " %8"IDENT_FMT, x);
    fputs("\n----+-", out);
    for (k = 0; k <  cset->max;   k++) fputc('-', out);
    fputs("-+", out);
    for (x = 0; x <  cset->valid; x++) fputs("---------", out);
    fputc('\n', out);           /* print a separating line */
    s1 = " | "; s2 = " |"; s3 = " "; d = ""; end = "\n";
    fmt = "%3"IDENT_FMT;        /* configure the separators */
  }
  for (y = 0; y < cset->cnt; y++) {
    yname = cset->names[y];     /* traverse the matrix columns, */
    if (!yname) continue;       /* but skip invalidated columns */
    fprintf(out, fmt, ++n);     /* print the column number, */
    fputs(s1,    out);          /* the first separator */
    fputs(yname, out);          /* and the column name */
    if (pad)                    /* if to pad the output */
      for (k = strlen(yname); k < cset->max; k++)
        fputc(' ', out);        /* fill the output field */
    fputs(s2, out);             /* print a separator */
    if (mode & LOWER) {         /* if to print the lower triangle */
      for (x = 0; x < y; x++) { /* traverse the columns, */
        xname = cset->names[x]; /* but skip invalidated columns */
        if (!xname) continue;   /* get the column name */
        fputs(s3, out);         /* print a separator */
        fputs(d,  out);         /* print start of field */
        cov = mvn_cov(mvn, y,x);/* get the covariance */
        if (isnan(cov)) fputs((pad) ? "      ?" : "?", out);
        else            dblout(out, cov, 8, pad);
        fputs(d,  out);         /* print the covariance */
      } }                       /* and end of field */
    else {                      /* if to skip the lower triangle */
      for (x = 0; x < y; x++) { /* skip columns before diagonal */
        if (!cset->names[x]) continue;
        fputs(s3, out);         /* print a separator */
        if (pad) fputs("        ", out);
        if (mode & LATEX) fputs("  ", out);
      }                         /* print blank entries */
    }                           /* LaTeX needs to replace '$' */
    fputs(s3, out);             /* print a separator */
    fputs(d,  out);             /* print start of field */
    cov = mvn_cov(mvn, y, y);   /* get and print the variance */
    if (isnan(cov)) fputs((pad) ? "      ?" : "?", out);
    else            dblout(out, cov, 8, pad);
    fputs(d,  out);             /* print end of field */
    if (mode & UPPER) {         /* if to print the upper triangle */
      for (x = y+1; x < cset->cnt; x++) {
        xname = cset->names[x]; /* traverse the remaining columns, */
        if (!xname) continue;   /* but skip invalidated columns */
        fputs(s3, out);         /* print a separator */
        fputs(d,  out);         /* print start of field */
        cov = mvn_cov(mvn, x, y);  /* get the covariance */
        if (isnan(cov)) fputs((pad) ? "      ?" : "?", out);
        else            dblout(out, cov, 8, pad);
        fputs(d,  out);         /* print the covariance */
      } }                       /* and end of field */
    else {                      /* skip upper/right part of matrix */
      for (x = y+1; x < cset->cnt; x++) {
        if (!cset->names[x]) continue;
        fputs(s3, out);         /* print a separator */
        if (pad) fputs("        ", out);
        if (mode & LATEX) fputs("  ", out);
      }                         /* print blank entries */
    }                           /* LaTeX needs to replace '$' */
    fputs(end, out);            /* terminate the output line */
  }
  if (mode & LATEX)             /* if to generate LaTeX code */
    fputs("\\hline\n\\end{tabular}\n", out);
}  /* covar() */

/*--------------------------------------------------------------------*/

static void correl (COLSET *cset, MVNORM *mvn, int mode, FILE *out)
{                               /* --- print correlation coefficients */
  IDENT      x, y, n = 0;       /* loop variables */
  size_t     k;                 /* loop variable */
  const char *xname, *yname;    /* to traverse the column names */
  double     r, t;              /* correlation coefficient, buffer */
  char       *s1, *s2, *s3, *d; /* separators, delimiter */
  char       *end;              /* end of line */
  char       *fmt;              /* format for index output */
  int        pad = 1;           /* flag for padding output w/ blanks */

  assert(cset && mvn && out);   /* check the function arguments */
  cs_prepare(cset);             /* prepare columns for output */
  if (mode & LATEX) {           /* if to generate LaTeX code */
    fprintf(out, "\\begin{tabular}"
                 "{|r|l|*{%"IDENT_FMT"}{r|}}", cset->valid);
    fputs("\\hline\n", out);    /* print a header */
    fprintf(out, "\\multicolumn{%"IDENT_FMT"}{|l|}", cset->valid +2);
    fputs("{correlation coefficients\\rule{0pt}{2.4ex}} \\\\\n", out);
    fputs("\\hline\\hline\\rule{0pt}{2.4ex}%\n", out);
    fputs(" no & attribute", out);
    for (k = 9; k <  cset->max;   k++) fputc(' ', out);
    for (x = 1; x <= cset->valid; x++)
      fprintf(out, " & %5"IDENT_FMT, x);
    fputs(" \\\\\n\\hline\\rule{0pt}{2.4ex}%\n", out);
    s1 = " & "; s2 = " "; s3 = "&"; d = "$"; end = " \\\\\n";
    fmt = "%3"IDENT_FMT; }      /* configure the separators */
  else if (mode & TABS) {       /* if plain text with tabs */
    fputs("correlation coefficients\n", out);
    fputs("no\tattribute", out);
    for (x = 1; x <= cset->valid; x++)
      fprintf(out, "\t%"IDENT_FMT, x);
    fputc('\n', out);           /* print a header */
    s1 = s3 = "\t"; s2 = d = ""; end = "\n"; pad = 0;
    fmt = "%"IDENT_FMT; }       /* configure the separators */
  else {                        /* if to generate text output */
    fputs("correlation coefficients\n", out);
    fputs(" no | attribute", out);
    for (k = 9; k <  cset->max;   k++) fputc(' ', out);
    fputs(" |", out);           /* print a header */
    for (x = 1; x <= cset->valid; x++)
      fprintf(out, " %5"IDENT_FMT, x);
    fputs("\n----+-", out);
    for (k = 0; k <  cset->max;   k++) fputc('-', out);
    fputs("-+", out);
    for (x = 0; x <  cset->valid; x++) fputs("------", out);
    fputc('\n', out);           /* print a separating line */
    s1 = " | "; s2 = " |"; s3 = " "; d = ""; end = "\n";
    fmt = "%3"IDENT_FMT;        /* configure the separators */
  }
  for (y = 0; y < cset->cnt; y++) {
    yname = cset->names[y];     /* traverse the matrix columns, */
    if (!yname) continue;       /* but skip invalidated columns */
    fprintf(out, fmt, ++n);     /* print the column number, */
    fputs(s1,    out);          /* the first separator */
    fputs(yname, out);          /* and the column name */
    if (pad)                    /* if to pad the output */
      for (k = strlen(yname); k < cset->max; k++)
        fputc(' ', out);        /* fill the output field */
    fputs(s2, out);             /* print a separator */
    if (mode & LOWER) {         /* if to print the lower triangle */
      for (x = 0; x < y; x++) { /* traverse the columns, */
        xname = cset->names[x]; /* but skip invalidated columns */
        if (!xname) continue;   /* get the column name */
        fputs(s3, out);         /* print a separator */
        fputs(d,  out);         /* print start of field */
        r = mvn_corr(mvn, y,x); /* get and check the correl. coeff. */
        if (isnan(r)) { fputs((pad) ? "     ?" : "?", out); continue; }
        t = fabs(r);            /* print a possible sign */
        if ((r < 0) && (t > 0.0005)) fputc('-', out);
        else if (pad)                fputc(' ', out);
        if (t >= 0.9995) fputs("1.00", out);
        else fprintf(out, ".%03.0f", t *1000);
        fputs(d,  out);         /* print the correlation coefficient */
      } }                       /* and end of field */
    else {                      /* if to skip the lower triangle */
      for (x = 0; x < y; x++) { /* skip columns before diagonal */
        if (!cset->names[x]) continue;
        fputs(s3, out);         /* print a separator */
        if (pad) fputs("     ", out);
        if (mode & LATEX) fputs("  ", out);
      }                         /* print blank entries */
    }                           /* LaTeX needs to replace '$' */
    fputs(s3, out);             /* print a separator */
    fputs(d,  out);             /* print start of field */
    if (!(mode & TABS)) fputc(' ', out);
    fputs("1.00", out);         /* diagonal is always 1.0 */
    fputs(d,  out);             /* print end of field */
    if (mode & UPPER) {         /* if to print the upper triangle */
      for (x = y+1; x < cset->cnt; x++) {
        xname = cset->names[x]; /* traverse the remaining columns, */
        if (!xname) continue;   /* but skip invalidated columns */
        fputs(s3, out);         /* print a separator */
        fputs(d,  out);         /* print start of field */
        r = mvn_corr(mvn, x,y); /* get and check the correl. coeff. */
        if (isnan(r)) { fputs((pad) ? "     ?" : "?", out); continue; }
        t = fabs(r);            /* print a possible sign */
        if ((r < 0) && (t > 0.0005)) fputc('-', out);
        else if (pad)                fputc(' ', out);
        if (t >= 0.9995) fputs("1.00", out);
        else fprintf(out, ".%03.0f", t *1000);
        fputs(d,  out);         /* print the correlation coefficient */
      } }                       /* and end of field */
    else {                      /* skip upper/right part of matrix */
      for (x = y+1; x < cset->cnt; x++) {
        if (!cset->names[x]) continue;
        fputs(s3, out);         /* print a separator */
        if (pad) fputs("     ", out);
        if (mode & LATEX) fputs("  ", out);
      }                         /* print blank entries, delimited */
    }                           /* by the second separator */
    fputs(end, out);            /* terminate the output line */
  }
  if (mode & LATEX)             /* if to generate LaTeX code */
    fputs("\\hline\n\\end{tabular}\n", out);
}  /* correl() */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (mvnorm) mvn_delete(mvnorm);   \
  if (colset) cs_delete(colset);    \
  if (symtab) st_delete(symtab);    \
  if (tread)  trd_delete(tread, 1); \
  if (out && (out != stdout) && (out != stderr)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s, *e;               /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_hdr  = NULL;      /* name of header file */
  CCHAR   *fn_tab  = NULL;      /* name of table  file */
  CCHAR   *fn_mat  = NULL;      /* name of matrix file */
  CCHAR   *recseps = NULL;      /* record     separators */
  CCHAR   *fldseps = NULL;      /* field      separators */
  CCHAR   *blanks  = NULL;      /* blank      characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  CCHAR   *comment = NULL;      /* comment    characters */
  int     hdflt    = 0;         /* flag for a default header */
  int     wgtfld   = 0;         /* flag for weight in last field */
  int     pos      = 0;         /* flag for positive values */
  int     cnt      = 0;         /* print counts for expected values */
  int     print    = 0;         /* flags for matrices to print */
  int     mode     = PLAIN|FULL;/* output mode/format */
  int     max      = 6;         /* maximal length of a value name */
  char    buf[32];              /* buffer for default name creation */
  int     *p;                   /* pointer to symbol data */
  int     d;                    /* delimiter type */
  int     m;                    /* number of attributes */
  int     n;                    /* number of data tuples */
  double  w, u;                 /* weight of data tuples, buffer */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] [-d|-h hdrfile] "
                     "tabfile matfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-x       print expected values and standard deviation\n");
    printf("-v       print covariance matrix\n");
    printf("-c       print correlation coefficients (default)\n");
    printf("-m       use maximum likelihood estimates "
                    "for the (co)variances\n");
    printf("-p       use only positive values (ignore negatives)\n");
    printf("-e       print counts with expected values\n");
    printf("-t       use tabulators to separate fields      "
                    "(default: lines)\n");
    printf("-L       print only left/lower triangle         "
                    "(default: full)\n");
    printf("-R       print only right/upper triangle\n");
    printf("-U       print only right/upper triangle\n");
    printf("-l       print output in LaTeX format           "
                    "(default: plain text)\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("-r#      record     separators                  "
                    "(default: \"\\n\")\n");
    printf("-f#      field      separators                  "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank      characters                  "
                    "(default: \" \\t\\r\")\n");
    printf("-u#      null value characters                  "
                    "(default: \"?*\")\n");
    printf("-C#      comment    characters                  "
                    "(default: \"#\")\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header  "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("matfile  file to write matrix/matrices to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'x': print |= MVN_EXPVAR;             break;
          case 'v': print |= MVN_COVAR;              break;
          case 'c': print |= MVN_CORREL;             break;
          case 'm': print |= MVN_MAXLLH;             break;
          case 'p': pos    = 1;                      break;
          case 'e': cnt    = 1;                      break;
          case 't': mode  |= TABS;                   break;
          case 'L': mode   = (mode & ~FULL) | LEFT;  break;
          case 'R': mode   = (mode & ~FULL) | RIGHT; break;
          case 'U': mode   = (mode & ~FULL) | UPPER; break;
          case 'l': mode  |= LATEX;                  break;
          case 'n': wgtfld = 1;                      break;
          case 'r': optarg = &recseps;               break;
          case 'f': optarg = &fldseps;               break;
          case 'b': optarg = &blanks;                break;
          case 'C': optarg = &comment;               break;
          case 'u': optarg = &nullchs;               break;
          case 'd': hdflt  = 1;                      break;
          case 'h': optarg = &fn_hdr;                break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_tab = s;      break;
        case  1: fn_mat = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (k != 2) error(E_ARGCNT);  /* check number of arguments */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  if (fn_hdr) hdflt = 0;        /* make header mode consistent */
  if ((!fn_tab || !*fn_tab) && (fn_hdr && !*fn_hdr))
    error(E_STDIN);             /* check assignments of stdin */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read table header --- */
  symtab = st_create(0, 0, ST_STRFN, (OBJFN*)0);
  if (!symtab) error(E_NOMEM);  /* create a symbol table */
  colset = cs_create();         /* create a column set */
  if (!colset) error(E_NOMEM);  /* for the estimation */
  tread  = trd_create();        /* create table reader and */
  if (!tread)  error(E_NOMEM);  /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, (fn_hdr) ? fn_hdr : fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  m = 0;                        /* init. the field counter */
  do {                          /* read the table header */
    d = trd_read(tread);        /* read the next field name */
    if  (d <= TRD_ERR) error(E_FREAD, trd_name(tread));
    if ((d != TRD_FLD) && wgtfld) break; /* skip tuple weight */
    s = trd_field(tread);       /* get the read field */
    if (hdflt) {                /* if to use a default header */
      sprintf(buf, "%d", m+1);  /* create an attribute name */
      p = (int*)st_insert(symtab, buf, -1, strlen(buf)+1, sizeof(int));
      if (!p) error(E_NOMEM); } /* insert name into symbol table */
    else {                      /* if to use a normal header */
      if (!*s) error(E_EMPATT, TRD_INFO(tread));
      p = (int*)st_insert(symtab, s, -1, trd_len(tread)+1, sizeof(int));
      if (!p) error(E_NOMEM);   /* insert name into symbol table */
      if (p == EXISTS) error(E_DUPATT, TRD_INFO(tread));
    }                           /* check for a duplicate name */
    *p = m++;                   /* set the attribute identifier */
    if ((k = (int)strlen(st_name(p))) > max)
      max = k;                  /* update the maximal name length */
    if (cs_add(colset, st_name(p)) != 0)
      error(E_NOMEM);           /* add a column to the column set */
    if (hdflt) cs_set(colset, m, s);   /* set the attribute value */
  } while (d == TRD_FLD);       /* while not at end of record */
  if (m <= 0)                   /* check the number of attributes */
    error(E_FLDCNT, TRD_INFO(tread));
  if (fn_hdr) {                 /* if a table header file is given, */
    trd_close(tread);           /* close the header file */
    fprintf(stderr, "[%d attribute(s)]", m);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a success message */

  /* --- create a covariance matrix --- */
  mvnorm = mvn_create(cs_colcnt(colset));
  if (!mvnorm) error(E_NOMEM);  /* create a m.v. normal distribution */

  /* --- compute covariance matrix --- */
  w = n = 0;                    /* init. the tuple counters */
  if      (fn_hdr) {            /* if a table header file is given */
    t = clock();                /* start timer, open table file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread)); }
  else if (hdflt) {             /* if to use a default header */
    w = n = 1;                  /* init. the tuple counters */
    if (wgtfld) {               /* if a tuple weight is given, */
      w = strtod(s = trd_field(tread), &e);
      if (!*e || (e == s) || (w < 0)) error(E_WEIGHT, TRD_INFO(tread));
    }                           /* convert last field to a number */
    if (pos) cs_procx(colset, mvnorm, w);
    else     cs_proc (colset, mvnorm, w);
  }                             /* aggregate for the first tuple */
  u = 1;                        /* get the default tuple weight */
  while (1) {                   /* read the table records */
    for (i = 0, d = TRD_FLD; d == TRD_FLD; i++) {
      d = trd_read(tread);      /* read the next field */
      if (d <= TRD_ERR)  error(E_FREAD,  trd_name(tread));
      if (d <= TRD_EOF)  break; /* if error or end of file, abort */
      s = trd_field(tread);     /* retrieve the field buffer */
      if (i <  m) cs_set(colset, i, s);
    }                           /* set the attribute value */
    if ((i != 0) && (i != m+wgtfld)) error(E_FLDCNT, TRD_INFO(tread));
    if (d <= TRD_EOF) break;    /* if at end of file, abort loop */
    if (wgtfld) {               /* if tuple weight in last field */
      u = strtod(s, &e);        /* convert last field to a number */
      if (*e || (e == s) || (u < 0)) error(E_WEIGHT, TRD_INFO(tread));
    }                           /* (get the tuple weight) */
    if (pos) cs_procx(colset, mvnorm, u);
    else     cs_proc (colset, mvnorm, u);
    w += u; n++;                /* aggregate for the current tuple */
  }                             /* and sum the tuple weights */
  trd_delete(tread, 1);         /* delete the table reader */
  tread = NULL;                 /* and print a success message */
  fprintf(stderr, "[%d attribute(s),", m);
  fprintf(stderr, " %d", n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- compute matrix/matrices --- */
  t = clock();                  /* start timer, print log message */
  fprintf(stderr, "computing values/matrices ... ");
  if (!print) print = MVN_CORREL;
  k = print;                    /* build the evaluation flags */
  if (k & MVN_CORREL) k |= MVN_COVAR;
  if (k & MVN_COVAR)  k |= MVN_EXPVAR;
  mvn_calc(mvnorm, k, 0, 0);    /* calculate parameters from data */
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));

  /* --- print matrix/matrices --- */
  t = clock();                  /* start timer, open output file */
  if (fn_mat && (strcmp(fn_mat, "-") == 0)) fn_mat = "";
  if (fn_mat && *fn_mat) { out = fopen(fn_mat, "w"); }
  else                   { out = stdout; fn_mat = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_mat);
  if (!out) error(E_FOPEN, fn_mat);
  if (print & MVN_EXPVAR) {     /* print the expected values */
    expvar(colset, mvnorm, mode, cnt, out);
    if (print & (MVN_COVAR|MVN_CORREL)) fputc('\n', out);
  }                             /* leave one line empty */
  if (print & MVN_COVAR) {      /* print the covariance matrix */
    covar(colset, mvnorm, mode, out);
    if (print & MVN_CORREL) fputc('\n', out);
  }                             /* leave one line empty */
  if (print & MVN_CORREL)       /* print the correlation coefficients */
    correl(colset, mvnorm, mode, out);
  if (out && (((out == stdout) ? fflush(out) : fclose(out)) != 0))
    error(E_FWRITE, fn_mat);    /* close the output file and */
  out = NULL;                   /* clear the file variable */
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
