/*----------------------------------------------------------------------
  File    : fbayes.c
  Contents: Full Bayes classifier management
  Author  : Christian Borgelt
  History : 2000.11.26 file created
            2000.11.29 first version completed
            2001.07.15 parser improved (global variables removed)
            2001.07.16 adapted to modified module scan
            2001.07.17 parser improved (conditional look ahead)
            2001.09.15 '*' instead of list of attributes made possible
            2003.04.26 function fbc_rand() added (random sampling)
            2004.08.12 adapted to new module parse
            2007.02.13 adapted to modified module attset
            2007.03.21 function fbc_exec() extended (posterior probs.)
            2007.10.19 bug in fbc_exec() fixed (posterior probs.)
            2009.06.04 parameter typi added to function fbc_exec()
            2011.02.02 adapted to modified scanner functions
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.20 output precision increased to 16 digits
            2014.10.29 bug in fbc_desc() fixed (comment line length)
            2014.12.10 Tikhonov regularization parameter added
            2015.07.31 reversed traversal order in fbc_induce()
            2015.08.02 1-in-n encoding of nominal variables added
            2015.08.03 setup with only selected attributes added
            2018.08.30 adapted to modified attmap interface
            2021.12.05 bug in function fbc_exec() fixed (tpl == null)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "fbayes.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define EPSILON     1e-12       /* to handle roundoff errors */
#define BLKSIZE     16          /* block size for arrays */

/* --- error messages --- */
/* error codes 0 to -15 defined in scanner.h */
#define E_ATTEXP    (-16)       /* attribute expected */
#define E_UNKATT    (-17)       /* unknown attribute */
#define E_DUPATT    (-18)       /* duplicate attribute */
#define E_MISATT    (-19)       /* missing attribute */
#define E_INVATT    (-20)       /* invalid attribute */
#define E_TRGATT    (-21)       /* target attribute name expected */
#define E_TRGTYPE   (-22)       /* target attribute is not nominal */
#define E_CLSCNT    (-23)       /* target has too few values/classes */
#define E_CLSEXP    (-24)       /* target value/class expected */
#define E_UNKCLS    (-25)       /* unknown class value */
#define E_DUPCLS    (-26)       /* duplicate class value */
#define E_DUPICLS   (-27)       /* duplicate class value (implicit) */
#define E_MISCLS    (-28)       /* missing class value */
#define E_ESTEXP    (-29)       /* estimation flag expected */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef FBC_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_ATTEXP  -16 */  "#attribute expected instead of '%s'",
  /* E_UNKATT  -17 */  "#unknown attribute '%s'",
  /* E_DUPATT  -18 */  "#duplicate attribute '%s'",
  /* E_MISATT  -19 */  "#missing attribute '%s'",
  /* E_INVATT  -20 */  "#invalid attribute '%s'",
  /* E_TRGATT  -21 */  "#target '%s' expected instead of '%s'",
  /* E_TRGTYPE -22 */  "#target attribute '%s' is not nominal",
  /* E_CLSCNT  -23 */  "#target attribute '%s' has too few values",
  /* E_CLSEXP  -24 */  "#target value/class expected instead of '%s'",
  /* E_UNKCLS  -25 */  "#unknown target value/class '%s'",
  /* E_DUPCLS  -26 */  "#duplicate target value/class '%s'",
  /* E_DUPICLS -27 */  "#duplicate target value/class '%s' (implicit)",
  /* E_MISCLS  -28 */  "#missing target value/class '%s'",
  /* E_ESTEXP  -29 */  "#estimation flag expected instead of '%s'",
};
#endif  /* #ifdef FBC_PARSE */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/
#ifdef FBC_INDUCE

static int clsrsz (FBC *fbc, VALID clscnt)
{                               /* --- resize class dependent arrays */
  VALID  i, n;                  /* loop variable, new array size */
  double *p;                    /* to traverse the frequency arrays */
  MVNORM **mvn;                 /* to traverse the distributions */

  assert(fbc && (clscnt >= 0)); /* check the function arguments */

  /* --- resize the class dependent arrays --- */
  n = fbc->size;                /* get the class dep. array size */
  if (clscnt >= n) {            /* if the arrays are too small */
    n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
    if (clscnt >= n) n = clscnt;
    p = (double*)realloc(fbc->frqs, (size_t)n *3 *sizeof(double));
    if (!p) return -1;          /* resize the frequencies array */
    fbc->frqs   = p;            /* and set the new array */
    fbc->priors = p +n;         /* organize the rest of the block */
    fbc->posts  = fbc->priors +n;
    memset(p, 0, (size_t)(n -fbc->size) *sizeof(double));
    mvn = (MVNORM**)realloc(fbc->mvns, (size_t)n *sizeof(MVNORM*));
    if (!mvn) return -1;        /* resize the distribution array */
    fbc->mvns = mvn;            /* and set the new array */
    fbc->size = n;              /* set the new size of the */
  }                             /* class dependent arrays */

  /* --- create new conditional distributions --- */
  mvn = fbc->mvns +clscnt;      /* traverse the normal distributions */
  for (i = clscnt -fbc->clscnt; --i >= 0; ) {
    *--mvn = mvn_create((DIMID)fbc->size);
    if (!*mvn) break;           /* create new normal distributions */
  }                             /* and set the new number of classes */
  if (i < 0) { fbc->clscnt = clscnt; return 0; }
  for (i = fbc->clscnt -i; --i >= 0; mvn++)
    mvn_delete(*mvn);           /* on error delete the newly created */
  return -1;                    /* multivariate normal distributions */
}  /* clsrsz() */

#endif
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

FBC* fbc_create (ATTSET *attset, ATTID trgid)
{                               /* --- create a full Bayes classifier */
  ATTID n;                      /* number of attributes */
  VALID i;                      /* loop variable for values */
  FBC   *fbc;                   /* created classifier */

  assert(attset                 /* check the function arguments */
  &&    (trgid >= 0) && (trgid < as_attcnt(attset))
  &&    (att_type(as_att(attset, trgid)) == AT_NOM));

  /* --- create the classifier body --- */
  n   = as_attcnt(attset);      /* get the number of attributes */
  fbc = (FBC*)malloc(sizeof(FBC) +(size_t)(n-1) *sizeof(int));
  if (!fbc) return NULL;        /* allocate the classifier body */
  fbc->attset = attset;         /* store attribute set and map */
  fbc->attmap = am_create(attset, 0, 1);
  if (!fbc->attmap) { free(fbc); return NULL; }
  am_target(fbc->attmap, trgid);
  fbc->attcnt = as_attcnt(attset);
  fbc->trgatt = as_att(attset, fbc->trgid = trgid);
  fbc->clscnt = att_valcnt(fbc->trgatt);
  fbc->size   = fbc->dimcnt = (DIMID)am_inpcnt(fbc->attmap);
  fbc->total  = fbc->corr   = fbc->reg   = fbc->lim = fbc->mode = 0;
  fbc->frqs   = fbc->priors = fbc->posts = NULL;
  fbc->mvns   = NULL;           /* clear for easier cleanup on error */
  fbc->dims   = NULL;           /* create a input vector buffer */
  fbc->vec    = (double*)malloc((size_t)fbc->size *sizeof(double));
  if (!fbc->vec)  { fbc_delete(fbc, 0); return NULL; }
  fbc->dims   = (DIMID*) malloc((size_t)fbc->size *sizeof(DIMID));
  if (!fbc->dims) { fbc_delete(fbc, 0); return NULL; }
  memset(fbc->marks, 0, (size_t)fbc->attcnt *sizeof(int));

  /* --- initialize the distributions --- */
  if (fbc->clscnt > 0) {        /* if there are classes */
    fbc->frqs = (double*)malloc((size_t)fbc->clscnt *3 *sizeof(double));
    if (!fbc->frqs) { fbc_delete(fbc, 0); return NULL; }
    fbc->priors = fbc->frqs   +fbc->clscnt;
    fbc->posts  = fbc->priors +fbc->clscnt;
    memset(fbc->frqs, 0, (size_t)fbc->clscnt *sizeof(double));
    fbc->mvns = (MVNORM**)calloc((size_t)fbc->clscnt, sizeof(MVNORM*));
    if (!fbc->mvns) { fbc_delete(fbc, 0); return NULL; }
    for (i = 0; i < fbc->clscnt; i++) {
      fbc->mvns[i] = mvn_create((DIMID)fbc->size);
      if (!fbc->mvns[i]) { fbc_delete(fbc, 0); return NULL; }
    }                           /* allocate a distribution array and */
  }                             /* create multivariate normal dists. */

  return fbc;                   /* return the created classifier */
}  /* fbc_create() */

/*--------------------------------------------------------------------*/

void fbc_delete (FBC *fbc, int delas)
{                               /* --- delete a full Bayes classifier */
  VALID i;                      /* loop variable for classes */

  assert(fbc);                  /* check the function argument */
  if (fbc->mvns) {              /* if there is a distribution array */
    for (i = fbc->clscnt; --i >= 0; )
      if (fbc->mvns[i]) mvn_delete(fbc->mvns[i]);
    free(fbc->mvns);            /* delete multivariate normal dists. */
  }                             /* and delete the frequency arrays */
  if (fbc->frqs)   free(fbc->frqs);
  if (fbc->vec)    free(fbc->vec);
  if (fbc->attmap) am_delete(fbc->attmap, 0);
  if (delas)       as_delete(fbc->attset);
  free(fbc);                    /* delete the classifier body */
}  /* fbc_delete() */

/*--------------------------------------------------------------------*/

void fbc_reglim (FBC *fbc, double reg, double lim)
{                               /* --- set regularization parameters */
  assert(fbc);                  /* check the function arguments */
  fbc->reg = reg; fbc->lim = lim;
}  /* fbc_reglim() */

/*--------------------------------------------------------------------*/

void fbc_clear (FBC *fbc)
{                               /* --- clear a full Bayes classifier */
  VALID i;                      /* loop variable */

  assert(fbc);                  /* check the function argument */
  fbc->dimcnt = fbc->size;      /* clear dimension selection */
  fbc->total  = 0;              /* clear the class frequencies */
  memset(fbc->frqs, 0, (size_t)fbc->clscnt *sizeof(double));
  for (i = fbc->clscnt; --i >= 0; )
    mvn_clear(fbc->mvns[i]);    /* clear multivariate normal dists. */
}  /* fbc_clear() */

/*--------------------------------------------------------------------*/
#ifdef FBC_INDUCE

int fbc_add (FBC *fbc, const TUPLE *tpl)
{                               /* --- add an instantiation */
  VALID  cls;                   /* value of class attribute */
  WEIGHT wgt;                   /* instantiation weight */

  assert(fbc);                  /* check the function argument */

  /* --- get class and weight --- */
  if (tpl) {                    /* if a tuple is given */
    cls = tpl_colval(tpl, fbc->trgid)->i;
    wgt = tpl_getwgt(tpl); }    /* get the class and the tuple weight */
  else {                        /* if no tuple is given */
    cls = att_inst(fbc->trgatt)->i;
    wgt = as_getwgt(fbc->attset);
  }                             /* get the class and the inst. weight */
  if (cls < 0) return 0;        /* if the class is null, abort */
  assert(wgt >= 0.0F);          /* check the tuple weight */

  /* --- update class distribution --- */
  if ((cls >= fbc->clscnt)      /* if the class is a new one, */
  &&  (clsrsz(fbc,cls+1) != 0)) /* resize the class dependent arrays */
    return -1;                  /* (frequencies and distributions) */
  fbc->frqs[cls] += wgt;        /* update the class frequency */
  fbc->total     += wgt;        /* and the total frequency */

  /* --- update conditional distributions --- */
  tpl_toas(tpl);                /* add inst. to the cond. distrib. */
  am_exec(fbc->attmap, AM_INPUTS, fbc->vec);
  mvn_add(fbc->mvns[cls], fbc->vec, wgt);
  return 0;                     /* return 'ok' */
}  /* fbc_add() */

/*--------------------------------------------------------------------*/

FBC* fbc_induce (TABLE *table, ATTID trgid, int mode)
{                               /* --- create a full Bayes classifier */
  TPLID  i, n;                  /* loop variables for tuples */
  FBC    *fbc;                  /* created full Bayes classifier */
  ATTSET *attset;               /* attribute set of the classifier */

  assert(table                  /* check the function arguments */
  &&    (trgid >= 0) && (trgid < tab_colcnt(table))
  &&    (att_type(as_att(tab_attset(table), trgid)) == AT_NOM));

  /* --- create a classifier --- */
  attset = tab_attset(table);   /* get the attribute set of the table */
  if (mode & FBC_CLONE) {       /* if the corresp. flag is set, */
    attset = as_clone(attset);  /* clone the attribute set */
    if (!attset) return NULL;   /* of the given data table, */
  }                             /* then create a classifier */
  fbc = fbc_create(attset, trgid);
  if (!fbc) { if (mode & FBC_CLONE) as_delete(attset); return NULL; }

  /* --- build the classifier --- */
  for (n = tab_tplcnt(table), i = 0; i < n; i++)
    fbc_add(fbc, tab_tpl(table, i)); /* add all tuples */

  return fbc;                    /* return the created classifier */
}  /* fbc_induce() */

/*--------------------------------------------------------------------*/

ATTID fbc_mark (FBC *fbc)
{                               /* --- mark used/selected attributes */
  ATTID m;                      /* attribute marker */
  ATTID i;                      /* loop variable for attributes */
  ATTID cnt;                    /* attribute counter */

  assert(fbc);                  /* check the function argument */
  for (cnt = i = 0; i < fbc->attcnt; i++) {
    if      (i == fbc->trgid)   m =  0;
    else if (fbc->marks[i] < 0) m = -1;
    else               { cnt++; m = +1; }
    att_setmark(as_att(fbc->attset, i), m);
  }                             /* transfer marker to attribute set */
  return cnt;                   /* return the number of marked atts. */
}  /* fbc_mark() */             /* (excluding the target attribute) */

#endif
/*--------------------------------------------------------------------*/

void fbc_setup (FBC *fbc, int mode)
{                               /* --- set up a full Bayes classifier */
  VALID  c;                     /* loop variable  for classes */
  double cnt;                   /* number of cases, sum of priors */
  ATTID  i, n;                  /* loop variables for attributes */
  ATT    *att;                  /* to traverse the attributes */
  DIMID  o, k;                  /* loop variables for dimensions */
  DIMID  *d;                    /* to collect the selected dimensions */
  int    m;                     /* calculation mode for mvn_calc() */

  assert(fbc);                  /* check the function arguments */

  /* --- estimate class probabilities --- */
  cnt = fbc->total +fbc->corr *(double)fbc->clscnt;
  if (cnt <= 0)                 /* clear the class probabilities */
    memset(fbc->priors, 0, (size_t)fbc->clscnt *sizeof(double));
  else                          /* estimate the class probabilities */
    for (c = 0; c < fbc->clscnt; c++)
      fbc->priors[c] = (fbc->frqs[c] +fbc->corr) /cnt;

  /* --- collect selected dimensions --- */
  if (!(mode & FBC_MARKED)) {   /* if to use all attributes */
    for (i = 0; i < fbc->attcnt; i++)
      fbc->marks[i] = +1;       /* set all attribute markers and */
    fbc->dimcnt = fbc->size; }  /* select all available dimensions */
  else {                        /* if to use only marked attributes */
    for (i = 0; i < fbc->attcnt; i++)
      fbc->marks[i] = -1;       /* clear all attribute markers */
    d = fbc->dims;              /* traverse attributes in map */
    for (n = am_attcnt(fbc->attmap), i = 0; i < n; i++) {
      att = am_att(fbc->attmap, i);
      if ((att == fbc->trgatt) || (att_getmark(att) < 0))
        continue;               /* traverse the marked attributes */
      fbc->marks[att_id(att)] = +1;
      o = (DIMID)am_off(fbc->attmap, i);
      for (k = (DIMID)am_cnt(fbc->attmap, i); --k >= 0; )
        *d++ = o++;             /* collect the dimension indices */
    }                           /* of the marked attributes */
    fbc->dimcnt = (DIMID)(d -fbc->dims);
  }                             /* get number of selected dimensions */
  fbc->marks[fbc->trgid] = 0;   /* set special marker for target */

  /* --- estimate conditional probabilities --- */
  m  = (fbc->mode & FBC_MAXLLH) ? MVN_MAXLLH : 0;
  m |= MVN_EXPVAR|MVN_COVAR|MVN_INVERSE|MVN_DECOM|MVN_NOZERO;
  if (fbc->dimcnt >= fbc->size) /* if to use all dimensions */
    for (c = 0; c < fbc->clscnt; c++)
      mvn_calc (fbc->mvns[c], m, fbc->reg, fbc->lim);
  else{                         /* if to use only selected dimensions */
    for (c = 0; c < fbc->clscnt; c++)
      mvn_calcx(fbc->mvns[c], m, fbc->reg, fbc->lim,
                fbc->dims, fbc->dimcnt);
  }                             /* calculate classifier parameters */
}  /* fbc_setup() */

/*--------------------------------------------------------------------*/

VALID fbc_exec (FBC *fbc, const TUPLE *tpl, double *conf, double *typi)
{                               /* --- execute a full Bayes class. */
  VALID  k, c;                  /* loop variables for classes */
  DIMID  i;                     /* loop variable for dimensions */
  double sum;                   /* sum of class probabilities */

  assert(fbc);                  /* check the function argument */
  if (tpl) tpl_toas(tpl);       /* turn tuple into a vector */
  am_exec(fbc->attmap, AM_INPUTS, fbc->vec);
  if (fbc->dimcnt < fbc->size)  /* if to use only selected dimensions */
    for (i = 0; i < fbc->dimcnt; i++)  /* collect values of selection */
      fbc->vec[i] = fbc->vec[fbc->dims[i]];
  /* The above approach of mapping everything and then projecting to */
  /* the selected dimensions is inefficient and should be improved.  */
  for (k = 0; k < fbc->clscnt; k++)
    fbc->posts[k] = fbc->priors[k] *mvn_eval(fbc->mvns[k], fbc->vec);
  sum = fbc->posts[c = 0];      /* compute posterior probabilities */
  for (k = 1; k < fbc->clscnt; k++) {
    if (fbc->posts[k] > fbc->posts[c]) c = k;
    sum += fbc->posts[k];       /* find the most probable class */
  }                             /* and sum all probabilities */
  if (typi)                     /* compute typicality index */
    *typi = fbc->posts[c] /(fbc->priors[c] *mvn_max(fbc->mvns[c]));
  sum = (sum > 0) ? 1/sum : 1;  /* compute the normalization factor */
  for (k = 0; k < fbc->clscnt; k++)
    fbc->posts[k] *= sum;       /* normalize the probabilities */
  if (conf)                     /* if requested, */
    *conf = fbc->posts[c];      /* store the confidence value */
  return c;                     /* return the classification result */
}  /* fbc_exec() */

/*--------------------------------------------------------------------*/

double* fbc_rand (FBC *fbc, double drand (void))
{                               /* --- generate a random tuple */
  ATTID  i, r, n;               /* loop variables for attributes */
  VALID  k, m;                  /* loop variable  for classes */
  ATT    *att;                  /* to traverse the attributes */
  int    t;                     /* attribute type */
  double x, sum;                /* random number, sum of probs. */
  double *p;                    /* sampled multidimensional point */

  assert(fbc                    /* check the function arguments */
  &&     drand && (fbc->dimcnt >= fbc->size));
  x = drand();                  /* generate a random number */
  for (sum = 0, k = 0; k < fbc->clscnt; k++) {
    sum += fbc->priors[k]; if (sum >= x) break; }
  if (k >= fbc->clscnt)         /* find the class that corresponds */
    k = fbc->clscnt -1;         /* to the generated random number */
  att_inst(fbc->trgatt)->n = k; /* and instantiate the class att. */
  p = mvn_rand(fbc->mvns[k], drand);   /* generate a random point */
  for (n = am_attcnt(fbc->attmap)-1, i = r = 0; i < n; r++, i++) {
    att = am_att(fbc->attmap,i);/* traverse attributes in map */
    t   = att_type(att);        /* depending on the attribute type */
    if      (t == AT_FLT) att_inst(att)->f = (DTFLT) p[r];
    else if (t == AT_INT) att_inst(att)->i = (DTINT)(p[r] +0.5);
    else {                      /* set attribute value directly */
      att_inst(att)->i = 0; x = p[r++];
      for (m = am_cnt(fbc->attmap, i), i = 0; i < m; r++, i++)
        if (p[r] > x) { x = p[r]; att_inst(att)->i = r; }
    }                           /* for a nominal attribute find the */
  }                             /* value with maximum random value */
  return p;                     /* return the generated point */
}  /* fbc_rand() */

/*--------------------------------------------------------------------*/

int fbc_desc (FBC *fbc, FILE *file, int mode, int maxlen)
{                               /* --- describe a full Bayes class. */
  ATTID i, n, m;                /* loop variables for attributes */
  VALID k;                      /* loop variable  for classes */
  int   pos, max;               /* current and maximal position */
  int   len, ind, l;            /* name length and indentation */
  ATT   *att;                   /* to traverse the attributes */
  char  buf[4*AS_MAXLEN+4];     /* output buffer */

  assert(fbc && file);          /* check the function arguments */
  max    = (maxlen > 0) ? maxlen : INT_MAX;
  maxlen = (maxlen > 0) ? maxlen : 70;

  /* --- print a header (as a comment) --- */
  if (mode & FBC_TITLE) {       /* if the title flag is set */
    fputs("/*", file); for (l = maxlen; --l >= 2; ) fputc('-', file);
    fputs("\n  full Bayes classifier\n", file);
    for (l = maxlen; --l >= 2; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */

  /* --- start description --- */
  scn_format(buf, att_name(fbc->trgatt), 0);
  fputs("fbc(", file);          /* get the class attribute name */
  fputs(buf, file);             /* and print it */
  fputs(") = {\n", file);       /* start the classifier */
  if ((fbc->corr > 0)           /* if estimation parameters */
  ||  (fbc->reg != 0)           /* (Laplace correction, */
  ||  (fbc->lim != 0)           /*  Tikhonov regularization) */
  ||   fbc->mode) {             /* differ from default values */
    fprintf(file, "  params = %.16g", fbc->corr);
    fprintf(file,          ", %.16g", fbc->reg);
    fprintf(file,          ", %.16g", fbc->lim);
    if (fbc->mode & FBC_MAXLLH) fputs(", maxllh", file);
    fputs(";\n", file);         /* print Laplace correction, */
  }                             /* regularization and estimation mode */

  /* --- print class distribution --- */
  fputs("  prob(", file);       /* print a distribution indicator */
  fputs(buf, file);             /* print the class att. name and */
  fputs(") = {\n    ", file);   /* start the class distribution */
  ind = att_valwd(fbc->trgatt, 0) +2;  /* compute the indentation */
  for (k = 0; k < fbc->clscnt; k++) {  /* traverse the classes */
    if (k > 0)                  /* if this is not the first class, */
      fputs(",\n    ", file);   /* start a new output line */
    len = (int)scn_format(buf, att_valname(fbc->trgatt, k), 0);
    fputs(buf, file);           /* get and print the class name */
    for (pos = len+2; pos < ind; pos++)
      putc(' ', file);          /* pad with blanks to equal width */
    fprintf(file, ": %.16g", fbc->frqs[k]);
    if (mode & FBC_REL)         /* print the absolute class frequency */
      fprintf(file, " (%.1f%%)", fbc->priors[k] *100);
  }                             /* print the relative class frequency */
  fputs(" };\n", file);         /* terminate the class distribution */

  /* --- print conditional distributions --- */
  if (fbc->size > 0) {          /* if conditional distributions */
    fputs("  prob(", file);     /* print a distribution indicator */
    pos = ind = 7;              /* and traverse the num. attributes */
    for (n = am_attcnt(fbc->attmap), i = m = 0; i < n; i++) {
      att = am_att(fbc->attmap, i);    /* traverse the attributes */
      if ((att == fbc->trgatt) || !fbc->marks[i])
        continue;               /* except non-selected attributes */
      if (m++ > 0) {            /* if this is not the first att., */
        fputc(',', file); pos++; }           /* print a separator */
      len = (int)scn_format(buf, att_name(att), 0);
      if ((pos+len > max-1)     /* get the condition name and */
      &&  (pos     > ind)) {    /* if the line would get too long, */
        fputc('\n', file);      /* start a new line and indent */
        for (pos = 0; pos < ind; pos++) fputc(' ', file);
      }                         /* indent to the opening parenthesis */
      fputs(buf, file); pos += len;
    }                           /* print the name of the attribute */
    fputc('|', file);           /* print condition indicator */
    scn_format(buf, att_name(fbc->trgatt), 0);
    fputs(buf, file);           /* print the class attribute name */
    fputs(") = {\n    ", file); /* and start the distribution */
    ind = att_valwd(fbc->trgatt, 0) +4;  /* compute the indentation */
    for (k = 0; k < fbc->clscnt; k++) {  /* traverse the classes */
      if (k > 0)                /* if this is not the first class, */
        fputs(",\n    ", file); /* start a new output line */
      len = (int)scn_format(buf, att_valname(fbc->trgatt, k), 0);
      fputs(buf, file);         /* get and print the class name */
      for (pos = len+4; pos < ind; pos++)
        putc(' ', file);        /* pad with blanks to equal width */
      fputs(": N(", file);      /* start a normal distribution */
      mvn_desc(fbc->mvns[k], file, -(ind+4), maxlen);
      fputc(')', file);         /* describe multivar. normal dists. */
    }                           /* and terminate the distribution */
    fputs(" };\n", file);       /* terminate the cond. distribution */
  }
  fputs("};\n", file);          /* terminate the classifier */

  /* --- print additional information (as a comment) --- */
  if (mode & FBC_INFO) {        /* if the add. info. flag is set */
    fputs("\n/*", file); for (l = maxlen; --l >= 2; ) fputc('-', file);
    fprintf(file, "\n  number of attributes: %"ATTID_FMT"+1", m);
    fprintf(file, "\n  number of tuples    : %g\n", fbc->total);
    for (l = maxlen; --l >= 2; ) fputc('-', file);
    fputs("*/\n", file);        /* print additional classifier info. */
  }                             /* (as a comment) */
  return ferror(file) ? -1 : 0; /* return the writing result */
}  /* fbc_desc() */

/*--------------------------------------------------------------------*/
#ifdef FBC_PARSE

static int parse (ATTSET *attset, SCANNER *scan, FBC **pfbc)
{                               /* --- parse a full Bayes classifier */
  ATTID i, n;                   /* loop variable for attributes */
  VALID k;                      /* loop variable for values/classes */
  int   t, c;                   /* buffer for token, counter */
  FBC   *fbc;                   /* created full Bayes classifier */
  ATT   *att;                   /* class attribute */

  /* --- read header --- */
  if ((scn_token(scan) != T_ID) /* check for 'fbc' */
  || ((strcmp(scn_value(scan), "fbc")        != 0)
  &&  (strcmp(scn_value(scan), "full_bayes") != 0)))
    SCN_ERROR(scan, E_STREXP, "fbc");
  SCN_NEXT(scan);               /* consume 'fbc' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  t = scn_token(scan);          /* check for an attribute name */
  if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_ATTEXP);
  i = as_attid(attset, scn_value(scan));
  if (i < 0)                       SCN_ERRVAL(scan, E_UNKATT);
  att = as_att(attset, i);      /* get and check the class attribute */
  if (att_type(att) != AT_NOM)     SCN_ERRVAL(scan, E_TRGTYPE);
  if (att_valcnt(att) < 1)         SCN_ERRVAL(scan, E_CLSCNT);
  *pfbc = fbc = fbc_create(attset, i); /* create a full */
  if (!fbc) SCN_ERROR(scan, E_NOMEM);  /* Bayes classifier */
  SCN_NEXT(scan);               /* consume the class attribute name */
  SCN_CHAR(scan, ')');          /* consume '(' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */

  /* --- read parameters --- */
  if ((scn_token(scan) == T_ID) /* if 'params' follows */
  &&  (strcmp(scn_value(scan), "params") == 0)) {
    SCN_NEXT(scan);             /* consume 'params' */
    SCN_CHAR(scan, '=');        /* consume '=' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    fbc->corr = strtod(scn_value(scan), NULL);
    if (fbc->corr < 0)            SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan); c = 0;      /* get and consume regularization */
    while (scn_token(scan) == ',') {
      SCN_NEXT(scan);           /* if more parameters follow */
      t = scn_token(scan);      /* get type of next parameter */
      if      (t == T_NUM) {    /* if number: Tikhonov regularization */
        if (c <= 0) {           /* first: initial regularization */
          fbc->reg = strtod(scn_value(scan), NULL);
          if (fbc->reg < 0) { SCN_ERROR(scan, E_NUMBER); } c = 1; }
        else {                  /* second: regularization limit */
          fbc->lim = strtod(scn_value(scan), NULL);
          if (fbc->lim < 0) SCN_ERROR(scan, E_NUMBER);
        } }                     /* (read regularization parameters) */
      else if ((t == T_ID)      /* if identifier: estimation flag */
      && (strcmp(scn_value(scan), "maxllh") == 0))
        fbc->mode |= FBC_MAXLLH;/* set maximum likelihood flag */
      else SCN_ERRVAL(scan, E_ESTEXP);
      SCN_NEXT(scan);           /* consume the estimator flag */
    }                           /* or the regularization parameter */
    SCN_CHAR(scan, ';');        /* consume ';' */
  }

  /* --- read class distribution --- */
  if ((scn_token(scan) != T_ID) /* check for 'prob', 'p' or 'P' */
  ||  ((strcmp(scn_value(scan), "prob") != 0)
  &&   (strcmp(scn_value(scan), "p")    != 0)
  &&   (strcmp(scn_value(scan), "P")    != 0)))
    SCN_ERROR(scan, E_STREXP, "prob");
  SCN_NEXT(scan);               /* consume   'prob', 'p' or 'P' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  t = scn_token(scan);          /* check for   class att. name */
  if (((t != T_ID) && (t != T_NUM))
  ||  (strcmp(scn_value(scan), att_name(att)) != 0))
    SCN_ERROR(scan, E_TRGATT, att_name(att), scn_value(scan));
  SCN_NEXT(scan);               /* consume the class att. name */
  SCN_CHAR(scan, ')');          /* consume ')' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  for (k = fbc->clscnt; --k >= 0; )
    fbc->frqs[k] = -1;          /* clear the class frequencies */
  while (1) {                   /* class value read loop */
    t = scn_token(scan);        /* check for a class name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_CLSEXP);
    if (t != T_NUM) t = ':';    /* if the token is no number, */
    else {                      /* the token must be a class, */
      SCN_NEXT(scan);           /* otherwise consume the token, */
      t = scn_token(scan);      /* note the next token, and */
      scn_back(scan);           /* go back to the previous one */
    }                           /* (look ahead one token) */
    if (t == ':') {             /* if a ':' follows */
      k = att_valid(att, scn_value(scan));
      if (k < 0)             SCN_ERRVAL(scan, E_UNKCLS);
      if (fbc->frqs[k] >= 0) SCN_ERRVAL(scan, E_DUPCLS);
      SCN_NEXT(scan);           /* get and consume the class name */
      SCN_CHAR(scan, ':'); }    /* consume ':' */
    else {                      /* if no ':' follows */
      if (fbc->frqs[k = (k+1) % fbc->clscnt] >= 0)
        SCN_ERROR(scan, E_DUPICLS, att_valname(att, k));
    }                           /* get and check cyclic successor id */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    fbc->frqs[k] = strtod(scn_value(scan), NULL);
    if (fbc->frqs[k] < 0)         SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and check the class frequency */
    if (scn_token(scan) == '('){/* if a relative number follows, */
      SCN_NEXT(scan);           /* consume '(' */
      if (scn_token(scan) != T_NUM)          SCN_ERROR(scan, E_NUMEXP);
      if (strtod(scn_value(scan), NULL) < 0) SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* consume the relative number */
      SCN_CHAR(scan, '%');      /* consume '%' */
      SCN_CHAR(scan, ')');      /* consume ')' */
    }
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop, */
  }                             /* otherwise consume ',' */
  SCN_CHAR(scan, '}');          /* consume '}' (end of distribution) */
  for (fbc->total = 0, k = fbc->clscnt; --k >= 0; ) {
    if (fbc->frqs[k] < 0) fbc->frqs[k] = 0;
    else    fbc->total += fbc->frqs[k];
  }                             /* sum the class frequencies */
  SCN_CHAR(scan, ';');          /* consume ';' */

  /* --- read conditional distributions --- */
  if (fbc->size <= 0) {         /* if there are cond. dimensions */
    SCN_CHAR(scan, '}');        /* consume '}' */
    SCN_CHAR(scan, ';');        /* consume ';' (end of classifier) */
    return 0;                   /* return 'ok' */
  }
  if ((scn_token(scan) != T_ID) /* check for 'prob', 'p' or 'P' */
  ||  ((strcmp(scn_value(scan), "prob") != 0)
  &&   (strcmp(scn_value(scan), "p")    != 0)
  &&   (strcmp(scn_value(scan), "P")    != 0)))
    SCN_ERROR(scan, E_STREXP, "prob");
  SCN_NEXT(scan);               /* consume   'prob', 'p' or 'P' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  if (scn_token(scan) == '*') { /* if a star follows, */
    SCN_NEXT(scan); }           /* simply consume it */
  else {                        /* if a list of attributes follows */
    n = 0;                      /* index of next attribute */
    while (1) {                 /* attribute read loop */
      if (n == fbc->trgid) n++; /* get next attribute to match */
      t = scn_token(scan);      /* check for an attribute name */
      if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_ATTEXP);
      i = as_attid(attset, scn_value(scan));
      if (i <  0)          SCN_ERRVAL(scan, E_UNKATT);
      if (i == fbc->trgid) SCN_ERRVAL(scan, E_INVATT);
      if (i <  n)          SCN_ERRVAL(scan, E_DUPATT);
      if (i >  n)          SCN_ERROR (scan, E_MISATT,
                                      as_attname(attset, n));
      SCN_NEXT(scan); n++;      /* consume the attribute name */
      if (scn_token(scan) != ',') break;
      SCN_NEXT(scan);           /* if at end of the list, abort loop, */
    }                           /* otherwise consume ',' */
    if (n == fbc->trgid) n++;   /* skip the target attribute */
    if (n <  fbc->attcnt)       /* check for a missing attribute */
      SCN_ERROR (scan, E_MISATT, as_attname(attset, n));
  }
  SCN_CHAR(scan, '|');          /* consume '|' (condition indicator) */
  t = scn_token(scan);          /* check for the class att. name */
  if (((t != T_ID) && (t != T_NUM))
  ||  (strcmp(scn_value(scan), att_name(att)) != 0))
    SCN_ERROR(scan, E_TRGATT, att_name(att), scn_value(scan));
  SCN_NEXT(scan);               /* consume the   class att. name */
  SCN_CHAR(scan, ')');          /* consume ')' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */

  for (k = fbc->clscnt; --k >= 0; )
    fbc->posts[k] = -1;         /* mark all classes as unread */
  while (1) {                   /* class value read loop */
    t = scn_token(scan);        /* check for name, number, or 'N' */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_CLSEXP);
    if (t == T_NUM) t = ':';    /* if the token is a number, */
    else {                      /* the token must be a class, */
      SCN_NEXT(scan);           /* otherwise consume the token, */
      t = scn_token(scan);      /* note the next token, and */
      scn_back(scan);           /* go back to the previous one */
    }                           /* (look ahead one token) */
    if (t == ':') {             /* if a  ':' follows */
      k = att_valid(att, scn_value(scan));
      if (k < 0)              SCN_ERRVAL(scan, E_UNKCLS);
      if (fbc->posts[k] >= 0) SCN_ERRVAL(scan, E_DUPCLS);
      SCN_NEXT(scan);           /* get and consume class value */
      SCN_CHAR(scan, ':'); }    /* consume ':' */
    else {                      /* if no ':' follows, */
      k = (k+1) % fbc->clscnt;  /* get the cyclic successor id */
      if (fbc->posts[k] >= 0)   /* check the implicit class value */
        SCN_ERROR(scan, E_DUPICLS, as_attname(attset, k));
    }                           /* (check for duplicates) */
    fbc->posts[k] = 1;          /* set the read marker */
    if ((scn_token(scan) != T_ID)  /* check for an 'N' */
    ||  (strcmp(scn_value(scan), "N") != 0))
      SCN_ERROR(scan, E_STREXP, "N");
    SCN_NEXT(scan);             /* consume 'N' */
    SCN_CHAR(scan, '(');        /* consume '(' */
    t = mvn_parse(fbc->mvns[k], scan, fbc->frqs[k]);
    if (t != 0) return t;       /* parse a multivariate normal dist. */
    SCN_CHAR(scan, ')');        /* consume ')' */
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop, */
  }                             /* otherwise consume ',' */
  for (k = 0; k < fbc->clscnt; k++) {
    if ((fbc->frqs[k] > 0) && (fbc->posts[k] < 0))
      SCN_ERROR(scan, E_MISCLS, att_valname(att, k));
  }                             /* check for a complete classifier */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' (end of distribution) */

  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' (end of classifier) */
  return 0;                     /* return 'ok' */
}  /* parse() */

/*--------------------------------------------------------------------*/

FBC* fbc_parse (ATTSET *attset, SCANNER *scan)
{                               /* --- parse a full Bayes classifier */
  FBC *fbc = NULL;              /* created full Bayes classifier */

  assert(attset && scan);       /* check the function arguments */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (parse(attset, scan, &fbc) != 0) {
    if (fbc) fbc_delete(fbc,0); /* parse a full Bayes classifier */
    return NULL;                /* if an error occurred, */
  }                             /* delete the classifier and abort */
  fbc_setup(fbc, 0);            /* set up the created classifier */
  return fbc;                   /* and then return it */
}  /* fbc_parse() */

#endif
