/*----------------------------------------------------------------------
  File    : fbayes.h
  Contents: Full Bayes classifier management
  Author  : Christian Borgelt
  History : 2000.11.10 file created
            2000.11.29 first version completed
            2001.07.16 adapted to modified module scan
            2003.04.26 function fbc_rand() added (random sampling)
            2004.08.12 adapted to new module parse
            2007.03.21 function fbc_post() added (posterior probability)
            2009.06.04 parameter typi added to function fbc_exec()
            2011.02.02 adapted to modified scanner functions
            2011.07.29 description mode FBC_INFO added
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
            2014.12.10 Tikhonov regularization parameter added
            2015.08.02 1-in-n encoding of nominal variables added
----------------------------------------------------------------------*/
#ifndef __FBAYES__
#define __FBAYES__
#ifdef FBC_PARSE
#ifndef MVN_PARSE
#define MVN_PARSE
#endif
#endif
#include "mvnorm.h"
#include "attmap.h"
#include "table.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- induction modes --- */
#define FBC_CLONE   0x0001      /* clone attribute set */
#define FBC_MAXLLH  0x0080      /* use max. likelihood est. of var. */

/* --- setup/induction modes --- */
#define FBC_ALL     0x0000      /* set up for all attributes */
#define FBC_MARKED  0x8000      /* set up only for marked attributes */

/* --- description modes --- */
#define FBC_TITLE   0x0001      /* print a title (as a comment) */
#define FBC_INFO    0x0002      /* print add. info. (as a comment) */
#define FBC_REL     0x0004      /* print relative numbers */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- full Bayes classifier --- */
  ATTSET *attset;               /* underlying attribute set */
  ATTMAP *attmap;               /* attribute map for numeric coding */
  ATTID  attcnt;                /* number of attributes */
  ATT    *trgatt;               /* target attribute */
  ATTID  trgid;                 /* identifier of target attribute */
  VALID  clscnt;                /* number of classes */
  DIMID  size;                  /* size of class dependent arrays */
  DIMID  *dims;                 /* selected dimensions */
  DIMID  dimcnt;                /* number of selected dimensions */
  int    mode;                  /* estimation mode (e.g. FBC_MAXLLH) */
  double corr;                  /* Laplace correction (for classes) */
  double reg;                   /* Tikhonov regularization parameter */
  double lim;                   /* limit for Tikhonov regularization */
  double total;                 /* total number of cases */
  double *frqs;                 /* class frequencies */
  double *priors;               /* prior     class probabilities */
  double *posts;                /* posterior class probabilities */
  double *vec;                  /* buffer for an input vector */
  MVNORM **mvns;                /* multivariate normal distributions */
  int    marks[1];              /* attribute selection markers */
} FBC;                          /* (full Bayes classifier) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern FBC*    fbc_create  (ATTSET *attset, ATTID trgid);
extern void    fbc_delete  (FBC *fbc, int delas);
extern ATTSET* fbc_attset  (FBC *fbc);
extern ATTMAP* fbc_attmap  (FBC *fbc);
extern ATT*    fbc_trgatt  (FBC *fbc);
extern ATTID   fbc_trgid   (const FBC *fbc);
extern VALID   fbc_clscnt  (const FBC *fbc);
extern ATTID   fbc_attcnt  (const FBC *fbc);

extern void    fbc_setmode (FBC *fbc, int mode);
extern double  fbc_getmode (FBC *fbc);
extern void    fbc_setcorr (FBC *fbc, double corr);
extern double  fbc_getcorr (FBC *fbc);
extern void    fbc_setreg  (FBC *fbc, double reg);
extern double  fbc_getreg  (FBC *fbc);
extern void    fbc_setlim  (FBC *fbc, double lim);
extern double  fbc_getlim  (FBC *fbc);
extern void    fbc_reglim  (FBC *fbc, double reg, double lim);

extern void    fbc_clear   (FBC *fbc);
extern double  fbc_total   (const FBC *fbc);

#ifdef FBC_INDUCE
extern int     fbc_add     (FBC *fbc, const TUPLE *tpl);
extern FBC*    fbc_induce  (TABLE *table, ATTID trgid, int mode);
extern ATTID   fbc_mark    (FBC *fbc);
#endif

extern void    fbc_setup   (FBC *fbc, int mode);
extern double  fbc_prior   (const FBC *fbc, VALID clsid);
extern MVNORM* fbc_mvnorm  (FBC *fbc, VALID clsid);
extern double  fbc_post    (const FBC *fbc, VALID clsid);
extern VALID   fbc_exec    (FBC *fbc, const TUPLE *tpl,
                            double *conf, double *typi);
extern double* fbc_rand    (FBC *fbc, double drand (void));

extern int     fbc_desc    (FBC *fbc, FILE *file, int mode, int maxlen);
#ifdef FBC_PARSE
extern FBC*    fbc_parse   (ATTSET *attset, SCANNER *scanner);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define fbc_attset(b)      ((b)->attset)
#define fbc_trgatt(b)      ((b)->trgatt)
#define fbc_trgid(b)       ((b)->trgid)
#define fbc_clscnt(b)      ((b)->clscnt)
#define fbc_attcnt(b)      ((b)->attcnt)

#define fbc_setmode(b,m)   ((b)->mode = (m))
#define fbc_getmode(b)     ((b)->mode)
#define fbc_setcorr(b,c)   ((b)->corr = (c))
#define fbc_getcorr(b)     ((b)->corr)
#define fbc_setreg(b,r)    ((b)->reg  = (r))
#define fbc_getreg(b)      ((b)->reg)
#define fbc_setlim(b,r)    ((b)->lim  = (r))
#define fbc_getlim(b)      ((b)->lim)

#define fbc_total(b)       ((b)->total)

#define fbc_prior(b,c)     ((b)->priors[c])
#define fbc_mvnorm(b,c)    ((b)->mvns[c])
#define fbc_post(b,c)      ((b)->posts[c])
#endif
