/*----------------------------------------------------------------------
  File    : mvnorm.c
  Contents: Multivariate normal distribution estimation and management
  Author  : Christian Borgelt
  History : 2000.11.10 file created from files correl.c and cluster.c
            2000.11.26 first version completed
            2001.05.24 possibilistic parameter added
            2001.05.26 handling of roundoff errors improved
            2001.07.15 some assertions added
            2004.04.15 treatment of decomposition failure improved
            2004.08.12 adapted to new module parse
            2005.09.05 bug in function decom (recomputation) fixed
            2011.02.02 adapted to modified scanner functions
            2011.02.04 parameter min added to function mvn_addx()
            2013.06.10 bug in function mvn_addx fixed (null values)
            2013.08.07 several pointer-based traversals eliminated
            2013.08.20 output precision increased to 16 digits
            2014.11.17 optional prevention of vanishing variances added
            2014.12.10 Tikhonov regularization parameter added
            2015.08.02 adapted to preprocessor definition of DIMID
            2015.08.11 bug in function mvn_calcx() fixed (covars.)
            2016.11.11 function normd() replaced by random/randn()
----------------------------------------------------------------------*/
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <float.h>
#include <math.h>
#include <assert.h>
#include "mvnorm.h"
#include "random.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI        3.14159265358979323846  /* \pi */
#endif
#define EPSILON     1e-9        /* to handle roundoff errors */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static int decom (MVNORM *mvn, DIMID size)
{                               /* --- Cholesky decomposition */
  DIMID  i, j, k;               /* loop variables */
  double *s, *r;                /* to traverse the matrix rows */
  double t, q;                  /* temporary buffers */

  assert(mvn);                  /* check the function argument */
  for (i = 0; i < size; i++) {  /* traverse the matrix rows */
    s = mvn->decom[i];          /* get the source row */
    for (t = mvn->covs[i][k = i]; --k >= 0; )
      t -= s[k] *s[k];          /* compute square of diag. element */
    if (!(t > 0)) return -1;    /* matrix is not positive definite */
    s[i] = t = sqrt(t);         /* store the diagonal element */
    if (!(t > 0)) return -1;    /* matrix is not positive definite */
    mvn->buf[i] = q = 1/t;      /* compute the corresponding factor */
    for (j = i; ++j < size; ) { /* traverse the row's other columns */
      r = mvn->decom[j];        /* get the matrix row */
      for (t = mvn->covs[j][k = i]; --k >= 0; )
        t -= s[k] *r[k];        /* compute the off-diagonal elements */
      r[i] = t *q;              /* column by column */
    }                           /* (-> lower triangular matrix) */
  }
  for (t = mvn->decom[0][0], i = size; --i > 0; )
    t *= mvn->decom[i][i];      /* multiply the diagonal elements */
  mvn->det = t*t;               /* to compute the determinant */
  if (mvn->det <= 0) return -1; /* check for a positive determinant */
  return 0;                     /* return 'ok' */
}  /* decom() */

/*--------------------------------------------------------------------*/

static void invert (MVNORM *mvn, DIMID size)
{                               /* --- invert matrix from Cholesky d. */
  DIMID  i, k, n;               /* loop variables */
  double *d;                    /* reciprocals of diagonal elements */
  double *s, *r;                /* to traverse the matrix elements */
  double t;                     /* temporary buffer */

  assert(mvn);                  /* check the function argument */
  d = mvn->buf;                 /* get reciprocals of diag. elements */
  for (n = size; --n >= 0; ) {  /* traverse the unit vectors */
    r = mvn->diff; r[n] = d[n]; /* and compute their originals */
    for (i = n; ++i < size; ) { /* traverse the matrix rows */
      s = mvn->decom[i];        /* get the source row */
      for (t = 0, k = i; --k >= n; )
        t -= r[k] *s[k];        /* do forward substitution with */
      r[i] = t *d[i];           /* the lower triangular matrix L */
    }                           /* (store result in r = mvn->diff) */
    for (i = size; --i >= n; ) {
      for (t = r[i], k = size; --k > i; )
        t -= mvn->decom[k][i] *mvn->inv[k][n];
      mvn->inv[i][n] = t*d[i];  /* do backward substitution with */
    }                           /* the upper triangular matrix L^T */
  }                             /* (using only the lower triangle) */
  mvn->norm = 1/sqrt(mvn->det *pow(2*M_PI, (double)size));
}  /* invert() */               /* compute normalization factor */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

static MVNORM* create (DIMID size)
{                               /* --- create a normal distribution */
  DIMID  i;                     /* loop variable */
  MVNORM *mvn;                  /* created normal distribution */
  double *p;                    /* to traverse the matrix rows */

  if (size <= 0) size = 1;      /* check and adapt the matrix size */
  mvn = (MVNORM*)calloc(1, sizeof(MVNORM)
                         +(size_t)(size-1) *sizeof(MVNROW*));
  if (!mvn) return NULL;        /* create the base structure */
  mvn->size  = mvn->curr = size;/* note the matrix size */
  mvn->covs  = (double**)malloc((size_t)size *4 *sizeof(double*));
  if (!mvn->covs) { free(mvn); return NULL; }
  mvn->corrs = mvn->covs  +size;   /* create arrays of pointers */
  mvn->decom = mvn->corrs +size;   /* for matrix rows */
  mvn->inv   = mvn->decom +size;
  mvn->exps  = (double*)malloc((size_t)(2*size+5) *(size_t)size
                               *sizeof(double));
  if (!mvn->exps) { mvn_delete(mvn); return NULL; }
  p = mvn->exps +size;          /* create and organize data arrays */
  for (i = size; --i >= 0; ) { mvn->covs[i]  = p; p += i+1; }
  for (i = size; --i >= 0; ) { mvn->corrs[i] = p; p += i+1; }
  for (i = size; --i >= 0; ) { mvn->decom[i] = p; p += i+1; }
  for (i = size; --i >= 0; ) { mvn->inv[i]   = p; p += i+1; }
  mvn->diff = p;                /* set buffers for difference from */
  mvn->buf  = p +size;          /* center and intermediate results */
  for (i = 0; i < size; i++) {  /* traverse the rows */
    mvn->rows[i] = (MVNROW*)malloc(sizeof(MVNROW)
                                 +(size_t)(i-1) *sizeof(MVNELEM));
    if (!mvn->rows[i]) { mvn_delete(mvn); return NULL; }
  }                             /* create the statistics rows */
  return mvn;                   /* and return it */
}  /* create() */

/*--------------------------------------------------------------------*/

MVNORM* mvn_create (DIMID size)
{                                /* --- create a normal distribution */
  MVNORM *mvn;                   /* created normal distribution */

  mvn = create(size);            /* create a normal distribution */
  if (!mvn) return NULL;
  mvn_clear(mvn);                /* clear the created normal dist. */
  return mvn;                    /* and return it */
}  /* mvn_create() */

/*--------------------------------------------------------------------*/

MVNORM* mvn_clone (const MVNORM *mvn)
{                               /* --- clone a normal distribution */
  DIMID  i;                     /* loop variable */
  MVNORM *clone;                /* created clone */

  assert(mvn);                  /* check the function argument */
  clone = create(mvn->size);    /* create a normal distribution */
  if (!clone) return NULL;      /* of the same size */
  clone->det  = mvn->det;       /* copy the determinant etc. */
  clone->norm = mvn->norm;      /* and all numeric parameters */
  memcpy(clone->exps, mvn->exps,
        (size_t)(2*mvn->size+5) *(size_t)mvn->size *sizeof(double));
  for (i = 0; i < mvn->size; i++)  /* copy all matrix rows */
    memcpy(clone->rows[i], mvn->rows[i],
           sizeof(MVNROW) +(size_t)(i-1) *sizeof(MVNELEM));
  return clone;                 /* return the created clone */
}  /* mvn_clone() */

/*--------------------------------------------------------------------*/

void mvn_delete (MVNORM *mvn)
{                               /* --- delete a normal distribution */
  DIMID i;                      /* loop variable */

  assert(mvn);                  /* check the function argument */
  for (i = 0; i < mvn->size; i++)  /* delete all statistics rows */
    if (mvn->rows[i]) free(mvn->rows[i]);
  if (mvn->exps) free(mvn->exps);  /* delete the parameter arrays, */
  if (mvn->covs) free(mvn->covs);  /* the pointer arrays, */
  free(mvn);                       /* and the base structure */
}  /* mvn_delete() */

/*--------------------------------------------------------------------*/

void mvn_clear (MVNORM *mvn)
{                               /* --- clear a normal distribution */
  DIMID i;                      /* loop variable */

  assert(mvn);                  /* check the function argument */
  mvn->curr = mvn->size;        /* remove dimension selection */
  for (i = 0; i < mvn->size; i++)
    memset(mvn->rows[i], 0,     /* clear all matrix rows */
           sizeof(MVNROW) +(size_t)(i-1) *sizeof(MVNELEM));
}  /* mvn_clear() */

/*--------------------------------------------------------------------*/

void mvn_addx (MVNORM *mvn, const double vals[], double cnt, double min)
{                               /* --- add a value vector */
  DIMID   i, k;                 /* loop variables */
  MVNROW  *r1, *r2;             /* to traverse the matrix rows */
  MVNELEM *e;                   /* to traverse the matrix elements */

  assert(mvn && vals && (cnt >= 0)); /* check the function arguments */
  for (i = 0; i < mvn->size; i++) {
    if (isnan(vals[i]) || (vals[i] <= min))
      continue;                 /* traverse the matrix rows, */
    r1 = mvn->rows[i];          /* but skip nulls and small values */
    r1->cv   = cnt    *vals[i]; /* precompute number of cases * value */
    r1->cv2  = r1->cv *vals[i]; /* and number of cases *value^2 */
    r1->cnt += cnt;             /* update the terms that are needed */
    r1->sv  += r1->cv;          /* to compute the expected value */
    r1->sv2 += r1->cv2;         /* and the variance */
    for (k = 0; k < i; k++) {   /* traverse the matrix columns */
      if (isnan(vals[k]) || (vals[k] <= min))
        continue;               /* traverse the preceding rows */
      r2 = mvn->rows[k];        /* but skip those rows, for which */
      e  = r1->elems +k;        /* the value is not large enough */
      e->cnt += cnt;
      e->sr  += r1->cv; e->sr2 += r1->cv2;
      e->sc  += r2->cv; e->sc2 += r2->cv2;
      e->src += r1->cv *vals[k];
    }                           /* update the terms needed */
  }                             /* to compute the covariance */
}  /* mvn_addx() */

/*--------------------------------------------------------------------*/

int mvn_calc (MVNORM *mvn, int mode, double reg, double lim)
{                               /* --- calc. parameters from data */
  DIMID   i, k;                 /* loop variables */
  MVNROW  *row;                 /* to traverse the matrix rows */
  MVNELEM *e;                   /* to traverse the matrix elements */
  double  *x, *v, *c;           /* to traverse the estim. parameters */
  double  n, t, y;              /* number of cases (-1), buffer */
  int     err = 0;              /* return code of decomposition */

  assert(mvn);                  /* check the function argument */
  if (mode & MVN_EXPVAR) {      /* expected values and variances */
    for (x = mvn->exps, i = 0; i < mvn->size; i++) {
      row  = mvn->rows[i];      /* traverse the statistics rows */
      x[i] = (row->cnt > 0) ? row->sv /row->cnt : 0;
      t    = row->sv2 -x[i] *row->sv;
      n    = (mode & MVN_MAXLLH) ? row->cnt : row->cnt-1;
      mvn->covs[i][i] = ((t > 0) && (n > 0)) ? t /n : 0;
      if ((mvn->covs[i][i] <= 0) && (mode & MVN_NOZERO))
        mvn->covs[i][i] = EPSILON;
    }                           /* compute the expected value */
  }                             /* and the variance */
  if (mode & MVN_COVAR) {       /* if to compute covariances */
    for (x = mvn->exps, i = 0; i < mvn->size; i++) {
      row = mvn->rows[i];       /* traverse the statistics rows */
      v   = mvn->covs[i];       /* and the covariance rows */
      if (v[i] <= 0) {          /* if the variance is zero */
        memset(v, 0, (size_t)i *sizeof(double));
        continue;               /* set all covariances to zero */
      }                         /* (to avoid inconsistencies) */
      for (e = row->elems, k = 0; k < i; k++) {
        n    = (mode & MVN_MAXLLH) ? e[k].cnt : e[k].cnt-1;
        v[k] = (n > 0)          /* compute the covariance */
             ? (e[k].src -x[k] *e[k].sr -x[i] *e[k].sc
               +e[k].cnt *x[i] *x[k]) /n : 0;
      }                         /* (this somewhat complicated formula */
    }                           /* takes missing values into account) */
  }
  if (mode & MVN_CORREL) {      /* correlation coefficients */
    for (i = 0; i < mvn->size; i++) {
      v = mvn->covs [i];        /* traverse the covariances and */
      c = mvn->corrs[i];        /* the correlation coefficients */
      c[i] = 1.0;               /* the correlation coefficient of */
      for (k = 0; k < i; k++) { /* a variable with itself is 1 */
        t    = v[i] *mvn->covs[k][k];
        t    = (t > 0) ? sqrt(t) : 0;
        c[k] = (t > 0) ? v[k] /t : 0;
      }                         /* compute the correlation coeffs. */
    }                           /* (set correl. coeff. to 'null' */
  }                             /* if the covariance is null) */
  if (mode & MVN_DECOM) {       /* do Cholesky decomposition */
    y = lim *lim;               /* compute regularization limit */
    t = reg *reg;               /* and initial regularization value */
    for (i = mvn->size; --i >= 0; )   /* save the diagonal */
      mvn->diff[i] = mvn->covs[i][i]; /* for the regularization */
    if (reg > 0) {              /* if to use Tikhonov regularization */
      for (i = mvn->size; --i >= 0; ) mvn->covs[i][i] += t;
      t *= 4;                   /* apply initial regularization */
    }                           /* and update regularization value */
    while (1) {                 /* matrix inversion loop */
      err = decom(mvn, mvn->size); /* try Cholesky decomposition */
      if (err == 0) break;      /* on success abort the loop */
      if (t >  y)   break;      /* limit regularization steps */
      if (t <= 0) t = 1e-6;     /* ensure positive regularization */
      for (i = mvn->size; --i >= 0; )      /* on failure */
        mvn->covs[i][i] = mvn->diff[i] +t; /* restore the diagonal */
      t *= 4;                   /* add regularization to diagonal */
    }                           /* and update regularization value */
    if (err != 0) {             /* if decomposition failed */
      mvn->det = pow(EPSILON, (double)mvn->size);
      for (i = mvn->size; --i >= 0; ) {
        c = mvn->decom[i]; c[i] = EPSILON;
        mvn->buf[i] = 1.0/EPSILON;
        for (k = i; ++k < mvn->size; ) mvn->decom[i][k] = 0;
      }                         /* set the decomposition of a */
    }                           /* minimal covariance matrix */
  }                             /* (min. variances, no covariances) */
  if (mode & MVN_INVERSE)       /* compute the inverse */
    invert(mvn, mvn->size);     /* from the Cholesky decomposition */
  mvn->curr = mvn->size;        /* current size is full size */
  return err;                   /* return the error flag */
}  /* mvn_calc() */

/*--------------------------------------------------------------------*/

int mvn_calcx (MVNORM *mvn, int mode, double reg, double lim,
               DIMID *sel, DIMID cnt)
{                               /* --- calc. parameters from data */
  DIMID   i, k, r, s;           /* loop variables */
  MVNROW  *row;                 /* to traverse the matrix rows */
  MVNELEM *e;                   /* to traverse the matrix elements */
  double  *x, *v, *c;           /* to traverse the estim. parameters */
  double  n, t, y;              /* number of cases (-1), buffer */
  int     err = 0;              /* return code of decomposition */

  assert(mvn);                  /* check the function argument */
  if (!sel) return mvn_calc(mvn, mode, reg, lim);
  if (mode & MVN_EXPVAR) {      /* expected values and variances */
    for (x = mvn->exps, r = 0; r < cnt; r++) {
      row  = mvn->rows[sel[r]]; /* traverse the statistics rows */
      x[r] = (row->cnt > 0) ? row->sv /row->cnt : 0;
      t    = row->sv2 -x[r] *row->sv;
      n    = (mode & MVN_MAXLLH) ? row->cnt : row->cnt-1;
      mvn->covs[r][r] = ((t > 0) && (n > 0)) ? t /n : 0;
      if ((mvn->covs[r][r] <= 0) && (mode & MVN_NOZERO))
        mvn->covs[r][r] = EPSILON;
    }                           /* compute the expected value */
  }                             /* and the variance */
  if (mode & MVN_COVAR) {       /* if to compute covariances */
    for (x = mvn->exps, r = 0; r < cnt; r++) {
      row = mvn->rows[sel[r]];  /* traverse the statistics rows */
      v   = mvn->covs[r];       /* and the covariance rows, */
      if (v[r] <= 0) {          /* if the variance is zero */
        memset(v, 0, (size_t)r *sizeof(double));
        continue;               /* set all covariances to zero */
      }                         /* (to avoid inconsistencies) */
      for (e = row->elems, s = 0; s < r; s++) {
        k    = sel[s];          /* traverse the columns */
        n    = (mode & MVN_MAXLLH) ? e[k].cnt : e[k].cnt-1;
        v[s] = (n > 0)          /* compute the covariance */
             ? (e[k].src -x[s] *e[k].sr -x[r] *e[k].sc
               +e[k].cnt *x[r] *x[s]) /n : 0;
      }                         /* (this somewhat complicated formula */
    }                           /* takes missing values into account) */
  }
  if (mode & MVN_CORREL) {      /* correlation coefficients */
    for (i = 0; i < cnt; i++) { /* traverse selected dimensions */
      v = mvn->covs [i];        /* traverse the covariances and */
      c = mvn->corrs[i];        /* the correlation coefficients */
      c[i] = 1.0;               /* the correlation coefficient of */
      for (k = 0; k < i; k++) { /* a variable with itself is 1 */
        t    = v[i] *mvn->covs[k][k];
        t    = (t > 0) ? sqrt(t) : 0;
        c[k] = (t > 0) ? v[k] /t : 0;
      }                         /* compute the correlation coeffs. */
    }                           /* (set correl. coeff. to 'null' */
  }                             /* if the covariance is null) */
  if (mode & MVN_DECOM) {       /* do Cholesky decomposition */
    y = lim *lim;               /* compute regularization limit */
    t = reg *reg;               /* and initial regularization value */
    for (i = mvn->size; --i >= 0; )   /* save the diagonal */
      mvn->diff[i] = mvn->covs[i][i]; /* for the regularization */
    if (reg > 0) {              /* if to use Tikhonov regularization */
      for (i = cnt; --i >= 0; ) mvn->covs[i][i] += t;
      t *= 4;                   /* apply initial regularization */
    }                           /* and update regularization value */
    while (1) {                 /* matrix inversion loop */
      err = decom(mvn, cnt);    /* try Cholesky decomposition */
      if (err == 0) break;      /* on success abort the loop */
      if ((reg >= 0) || (t > y)) break;
      for (i = cnt; --i >= 0; ) /* on failure restore the diagonal */
        mvn->covs[i][i] = mvn->diff[i];
      for (i = cnt; --i >= 0; ) mvn->covs[i][i] += t;
      t *= 4;                   /* add regularization to diagonal */
    }                           /* and update regularization value */
    if (err != 0) {             /* if decomposition failed */
      mvn->det = pow(EPSILON, (double)cnt);
      for (i = cnt; --i >= 0; ) {
        c = mvn->decom[i]; c[i] = EPSILON;
        mvn->buf[i] = 1.0/EPSILON;
        for (k = i; ++k < cnt; ) mvn->decom[i][k] = 0;
      }                         /* set the decomposition of a */
    }                           /* minimal covariance matrix */
  }                             /* (min. variances, no covariances) */
  if (mode & MVN_INVERSE)       /* compute the inverse */
    invert(mvn, cnt);           /* from the Cholesky decomposition */
  mvn->curr = cnt;              /* current size is reduced size */
  return err;                   /* return the error flag */
}  /* mvn_calcx() */

/*--------------------------------------------------------------------*/

double mvn_eval (MVNORM *mvn, const double vals[])
{                               /* --- evaluate a normal distribution */
  DIMID  i, k;                  /* loop variables */
  double *d, *p;                /* to traverse arrays and matrices */
  double t;                     /* temporary buffer for exponent */

  assert(mvn && vals);          /* check the function arguments */
  p = mvn->buf; d = mvn->diff;  /* compute the difference vector */
  for (i = 0; i < mvn->curr; i++)
    d[i] = vals[i] -mvn->exps[i];
  for (i = 0; i < mvn->curr; i++) {
    p[i] = 0;                   /* traverse the matrix columns */
    for (k = mvn->curr; --k >= i; )
      p[i] += d[k] *mvn->inv[k][i];
    for (k = i;         --k >= 0; )
      p[i] += d[k] *mvn->inv[i][k];
  }                             /* multiply diff. vec. with inverse */
  for (t = 0, i = 0; i < mvn->curr; i++)
    t += d[i] *p[i];            /* calc. product with d^T * \Sigma^-1 */
  return mvn->norm *exp(-0.5 *t);  /* return the prob. density */
}  /* mvn_eval() */

/*--------------------------------------------------------------------*/

double* mvn_rand (MVNORM *mvn, double drand (void))
{                               /* --- generate random sample point */
  DIMID  i, k;                  /* loop variables */
  double *b, *r;                /* to access the buffer/matrix rows */

  assert(mvn && drand);         /* check the function arguments */
  b = mvn->buf;                 /* get the internal buffer */
  for (i = 0; i < mvn->curr; i++)
    b[i] = randn(drand);        /* generate points from N(0,1)^n */
  for (i = mvn->curr; --i >= 0; ) {
    r = mvn->decom[i];          /* traverse the matrix rows */
    b[i] *= r[i];               /* multiply the random vector */
    for (k = i; --k >= 0; )     /* with the Cholesky decomposed */
      b[i] += r[k] *b[k];       /* covariance matrix (transposed) */
    b[i] += mvn->exps[i];       /* (hyperellipsoidal norm. dist.) */
  }                             /* and add the expected value vector */
  return b;                     /* return the created sample point */
}  /* mvn_rand() */

/*--------------------------------------------------------------------*/

int mvn_desc (MVNORM *mvn, FILE *file, int offs, int maxlen)
{                               /* --- describe a normal distribution */
  int    k;                     /* loop variable for offset */
  DIMID  i;                     /* loop variable for rows/columns */
  double *p;                    /* to traverse the matrix rows */

  assert(mvn && file);          /* check the function arguments */

  /* --- expected values --- */
  if (offs > 0) {               /* if the offset is positive, */
    for (k = offs; --k >= 0; )  /* indent the first line */
      fputc(' ', file);         /* to the given offset */
  }                             /* (indent all but the first line */
  offs = abs(offs);             /* if the offset is negative) */
  fputc('[', file);             /* start the expected values vector */
  for (p = mvn->exps, i = 1; i < mvn->curr; i++)
    fprintf(file, "%.16g, ", *p++); /* print the expected values and */
  fprintf(file, "%.16g],\n", *p);   /* then terminate the vector */

  /* --- (co)variances --- */
  for (i = 0; i < mvn->curr; i++) {
    if (i > 0)                  /* if this is not the first row, */
      fputs(",\n", file);       /* print a comma and start new line */
    for (k = offs; --k >= 0; )  /* indent the current line */
      fputc(' ', file);         /* to the given offset */
    fputc('[', file);           /* start a new matrix row */
    for (p = mvn->covs[i], k = 0; k < i; k++)
      fprintf(file, "%.16g, ", *p++);      /* print the covariances */
    fprintf(file, "%.16g]", *p);   /* and then terminate the vector */
  }                                /* by printing the variance */

  return ferror(file) ? -1 : 0; /* return file write status */
}  /* mvn_desc() */

/*--------------------------------------------------------------------*/
#ifdef MVN_PARSE

static int get_exps (MVNORM *mvn, SCANNER *scan)
{                               /* --- parse expected values */
  DIMID  i;                     /* loop variable */
  MVNROW *row;                  /* to traverse the matrix rows */
  double t;                     /* temporary buffer */

  assert(mvn && scan);          /* check the function arguments */
  SCN_CHAR(scan, '[');          /* consume '(' */
  for (i = 0; i < mvn->size; i++) {
    if (i > 0) SCN_CHAR(scan, ',');
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    mvn->exps[i] = t = strtod(scn_value(scan), NULL);
    SCN_NEXT(scan);             /* get and consume expected value */
    row = mvn->rows[i];         /* recompute the sum of values */
    row->sv = t *row->cnt;      /* from the expected value */
  }                             /* and the number of cases */
  SCN_CHAR(scan, ']');          /* consume ')' */
  SCN_CHAR(scan, ',');          /* consume ',' */
  return 0;                     /* return 'ok' */
}  /* get_exps() */

/*--------------------------------------------------------------------*/

static int get_covs (MVNORM *mvn, SCANNER *scan)
{                               /* --- parse (co)variances */
  DIMID   i, k;                 /* loop variables */
  MVNROW  *row;                 /* to traverse the matrix rows */
  MVNELEM *e;                   /* to traverse the matrix elements */
  double  *v;                   /* to traverse the (co)variances */

  assert(mvn && scan);          /* check the function arguments */
  for (i = 0; i < mvn->size; i++) {
    row = mvn->rows[i];         /* traverse the matrix rows */
    if (i > 0) SCN_CHAR(scan, ',');
    SCN_CHAR(scan, '[');        /* consume '[' */
    e = row->elems;             /* traverse the statistics and */
    v = mvn->covs[i];           /* the covariances of each row */
    for (k = 0; k < i; k++) {   /* check for a number */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      v[k] = strtod(scn_value(scan), NULL);
      e[k].src = v[k] *(row->cnt -1) +row->sv *mvn->exps[k];
      SCN_NEXT(scan);           /* recompute mixed sum, */
      SCN_CHAR(scan, ',');      /* consume covariance, */
    }                           /* and consume ',' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    v[i] = strtod(scn_value(scan), NULL);
    if (v[i] < 0)                 SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and consume variance */
    SCN_CHAR(scan, ']');        /* consume ']' */
    row->sv2 = v[i] *(row->cnt-1) +row->sv *mvn->exps[i];
    if (row->sv2 < 0) row->sv2 = 0; /* recompute sum of squares */
    for (k = i; --k >= 0; ) {   /* traverse the elements again */
      e[k].cnt = row->cnt;      /* set the number of cases */
      e[k].sr  = row->sv;  e[k].sc  = mvn->rows[k]->sv;
      e[k].sr2 = row->sv2; e[k].sc2 = mvn->rows[k]->sv2;
    }                           /* get the sums of values and */
  }                             /* the sums of squared values */
  return 0;                     /* return 'ok' */
}  /* get_covs() */

/*--------------------------------------------------------------------*/

static int get_poss (MVNORM *mvn, SCANNER *scan)
{                               /* --- get possibilistic parameter */
  assert(mvn && scan);          /* check the function arguments */
  if (scn_token(scan) != ',') { /* if no comma follows, */
    mvn->poss = 1; return 0; }  /* set a default value */
  SCN_NEXT(scan);               /* consume ',' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  mvn->poss = strtod(scn_value(scan), NULL);
  if (mvn->poss <= 0)           SCN_ERROR(scan, E_NUMBER);
  SCN_NEXT(scan);               /* get and consume the parameter */
  return 0;                     /* return 'ok' */
}  /* get_poss() */

/*--------------------------------------------------------------------*/

int mvn_parse (MVNORM *mvn, SCANNER *scan, double cnt)
{                               /* --- parse a normal distribution */
  DIMID i;                      /* loop variable */
  int   r;                      /* function result */

  assert(mvn && scan);          /* check the function arguments */
  if (cnt < 2) cnt = 2;         /* adapt the number of cases */
  for (i = 0; i < mvn->size; i++)
    mvn->rows[i]->cnt = cnt;    /* set number of cases */
  r = get_exps(mvn, scan);      /* read the expected values */
  if (r < 0) return r;
  r = get_covs(mvn, scan);      /* read the (co)variances */
  if (r < 0) return r;
  r = get_poss(mvn, scan);      /* read the possibilistic parameter */
  if (r < 0) return r;
  return 0;                     /* return 'ok' */
}  /* mvn_parse() */

#endif
