/*----------------------------------------------------------------------
  File    : mvnorm.h
  Contents: multivariate normal distribution estimation and management
  Author  : Christian Borgelt
  History : 2000.11.10 file created
            2000.11.26 first version completed
            2001.05.24 possibilistic parameter added
            2004.04.22 functions mvn_addx() and mvn_cnt() added
            2004.08.12 adapted to new module parse
            2009.06.04 function mvn_max() added (get maximum value)
            2011.02.04 parameter min added to function mvn_addx()
            2014.11.17 optional prevention of vanishing variances added
            2015.08.02 preprocessor definition of type DIMID added
----------------------------------------------------------------------*/
#ifndef __MVNORM__
#define __MVNORM__
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#ifdef MVN_PARSE
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#include "scanner.h"
#endif

#ifdef _MSC_VER
#ifndef INFINITY
#define INFINITY     (DBL_MAX+DBL_MAX)
#endif
#ifndef NAN
#define NAN          (INFINITY-INFINITY)
#endif
#ifndef isnan
#define isnan        _isnan
#endif
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifdef EXTDATA

#ifndef DIMID
#define DIMID       ptrdiff_t   /* dimension identifier */
#endif

#else

#ifndef DIMID
#define DIMID       int         /* dimension identifier */
#endif

#endif
/*--------------------------------------------------------------------*/

#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

/*--------------------------------------------------------------------*/

#if   DIMID==int
#ifndef DIMID_MAX
#define DIMID_MAX   INT_MAX     /* maximum dimension identifier */
#endif
#ifndef DIMID_SGN
#define DIMID_SGN   INT_MIN     /* sign of dimension identifier */
#endif
#ifndef DIMID_FMT
#define DIMID_FMT   "d"         /* printf format code for int */
#endif
#ifndef strtodim
#define strtodim(s,p)   (int)strtol(s,p,0)
#endif

#elif DIMID==long
#ifndef DIMID_MAX
#define DIMID_MAX   LONG_MAX    /* maximum dimension identifier */
#endif
#ifndef DIMID_SGN
#define DIMID_SGN   LONG_MIN    /* sign of dimension identifier */
#endif
#ifndef DIMID_FMT
#define DIMID_FMT   "ld"        /* printf format code for long */
#endif
#ifndef strtodim
#define strtodim(s,p)   strtol(s,p,0)
#endif

#elif DIMID==ptrdiff_t
#ifndef DIMID_MAX
#define DIMID_MAX   PTRDIFF_MAX /* maximum dimension identifier */
#endif
#ifndef DIMID_SGN
#define DIMID_SGN   PTRDIFF_MIN /* sign of dimension identifier */
#endif
#ifndef DIMID_FMT
#  ifdef _MSC_VER
#  define DIMID_FMT "Id"        /* printf format code for ptrdiff_t */
#  else
#  define DIMID_FMT "td"        /* printf format code for ptrdiff_t */
#  endif                        /* in case C99 is not supported */
#endif
#ifndef strtodim
#define strtodim(s,p)   (ptrdiff_t)strtoll(s,p,0)
#endif

#else
#error "DIMID must be either 'int', 'long' or 'ptrdiff_t'"
#endif

/*--------------------------------------------------------------------*/

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*--------------------------------------------------------------------*/
#define MVN_NULL     NAN        /* a null value */

/* --- evaluation flags --- */
#define MVN_EXPVAR   0x0001     /* compute exp. values and variances */
#define MVN_COVAR    0x0002     /* compute covariances */
#define MVN_CORREL   0x0004     /* compute correlation coefficients */
#define MVN_DECOM    0x0008     /* compute Cholesky decomposition */
#define MVN_INVERSE  0x0010     /* compute inverse matrix */
#define MVN_ALL      0x001f     /* compute all parameters */
#define MVN_NOZERO   0x0040     /* prevent vanishing variances */
#define MVN_MAXLLH   0x0080     /* use max. likelihood estimates */
                                /* default: use unbiased estimates */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- matrix element --- */
  double  cnt;                  /* sum of number of cases */
  double  sr;                   /* sum of row    var. values */
  double  sr2;                  /* sum of squared row    var. values */
  double  sc;                   /* sum of column var. values */
  double  sc2;                  /* sum of squared column var. values */
  double  src;                  /* sum of product of values */
} MVNELEM;                      /* (matrix element) */

typedef struct {                /* --- matrix row --- */
  double  cv;                   /* number of cases * current value */
  double  cv2;                  /* number of cases * current value^2 */
  double  cnt;                  /* sum of number of cases */
  double  sv;                   /* sum of values */
  double  sv2;                  /* sum of squared values */
  MVNELEM elems[1];             /* matrix elements */
} MVNROW;                       /* (matrix row) */

typedef struct {                /* --- multivariate normal dist. --- */
  DIMID   size;                 /* number of rows / columns */
  DIMID   curr;                 /* computed number of rows / columns */
  double  *exps;                /* expected values */
  double  **covs;               /* covariance matrix */
  double  **corrs;              /* correlation coefficients */
  double  **decom;              /* Cholesky decomposition */
  double  **inv;                /* inverse of covariance matrix */
  double  det;                  /* determinant of inverse matrix */
  double  norm;                 /* normalization factor */
  double  poss;                 /* param. for possibilistic interpr. */
  double  *diff;                /* difference vector from center */
  double  *buf;                 /* buffer for intermediate results */
  MVNROW  *rows[1];             /* matrix rows for statistics */
} MVNORM;                       /* (multivariate normal distribution) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern MVNORM* mvn_create (DIMID size);
extern MVNORM* mvn_clone  (const MVNORM *mvn);
extern void    mvn_delete (MVNORM *mvn);
extern DIMID   mvn_size   (const MVNORM *mvn);

extern void    mvn_clear  (MVNORM *mvn);
extern void    mvn_add    (MVNORM *mvn, const double vals[],
                           double cnt);
extern void    mvn_addx   (MVNORM *mvn, const double vals[],
                           double cnt, double min);
extern int     mvn_calc   (MVNORM *mvn, int mode, double reg,
                           double lim);
extern int     mvn_calcx  (MVNORM *mvn, int mode, double reg,
                           double lim, DIMID *sel, DIMID cnt);
extern double  mvn_eval   (MVNORM *mvn, const double vals[]);
extern double* mvn_rand   (MVNORM *mvn, double drand (void));
extern double  mvn_max    (const MVNORM *mvn);

extern double  mvn_cnt    (MVNORM *mvn, DIMID index);
extern double  mvn_exp    (MVNORM *mvn, DIMID index);
extern double  mvn_var    (MVNORM *mvn, DIMID index);
extern double  mvn_cov    (MVNORM *mvn, DIMID index1, DIMID index2);
extern double  mvn_corr   (MVNORM *mvn, DIMID index1, DIMID index2);
extern double  mvn_decom  (MVNORM *mvn, DIMID index1, DIMID index2);
extern double  mvn_inv    (MVNORM *mvn, DIMID index1, DIMID index2);
extern double  mvn_det    (MVNORM *mvn);
extern double  mvn_poss   (MVNORM *mvn);

extern int     mvn_desc   (MVNORM *mvn, FILE *file,
                           int offs, int maxlen);
#ifdef MVN_PARSE
extern int     mvn_parse  (MVNORM *mvn, SCANNER *scanner, double cnt);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define mvn_size(d)       ((d)->size)
#define mvn_add(m,v,c)    mvn_addx(m,v,c,-INFINITY)

#define mvn_max(d)        ((d)->norm)
#define mvn_cnt(d,i)      ((d)->rows[i]->cnt)
#define mvn_exp(d,i)      ((d)->exps[i])
#define mvn_var(d,i)      ((d)->covs[i][i])
#define mvn_cov(d,i,k)    (((i) > (k)) ? (d)->covs [i][k] \
                                       : (d)->covs [k][i])
#define mvn_corr(d,i,k)   (((i) > (k)) ? (d)->corrs[i][k] \
                                       : (d)->corrs[k][i])
#define mvn_decom(d,i,k)  (((i) > (k)) ? (d)->decom[i][k] \
                                       : (d)->decom[k][i])
#define mvn_inv(d,i,k)    (((i) > (k)) ? (d)->inv  [i][k] \
                                       : (d)->inv  [k][i])
#define mvn_det(d)        ((d)->det)
#define mvn_poss(d)       ((d)->poss)

#endif
