/*----------------------------------------------------------------------
  File    : nbayes.c
  Contents: Naive Bayes classifier management
  Author  : Christian Borgelt
  History : 1998.12.07 file created
            1998.12.08 nbc_create(), nbc_delete(), nbc_add() programmed
            1998.12.10 function nbc_desc() completed
            1998.12.11 function nbc_exec() completed
            1998.12.12 function nbc_parse() completed
            1998.12.16 all functions debugged
            1999.02.13 tuple parameters added to nbc_add and nbc_exec
            1999.01.10 execution for one att. made a separate function
            1999.03.11 function nbc_induce() added
            1999.03.25 distrib. of tuple weight for null values added
            1999.03.27 functions nbc_exp() und nbc_var() added
            1999.05.15 automatic frequency array resizing added
            2000.11.10 function nbc_exec() adapted
            2000.11.18 function nbc_setup() added, nbc_exec() adapted
            2000.11.21 redesign completed
            2001.02.11 bug in function nbc_mark (> instead of >=) fixed
            2001.07.15 parser improved (global variables removed)
            2001.07.16 adapted to modified module scan
            2001.07.17 parser improved (conditional look ahead)
            2003.04.26 function nbc_rand() added
            2004.04.15 zero variances replaced by EPSILON
            2004.08.12 adapted to new module parse
            2007.02.13 adapted to modified module attset
            2007.03.21 function nbc_exec() extended (posterior probs.)
            2009.06.04 parameter typi added to function nbc_exec()
            2009.06.05 typicality computation fixed (missing values)
            2011.02.02 adapted to modified scanner functions
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.20 output precision increased to 16 digits
            2014.10.29 bug in nbc_desc() fixed (comment line length)
            2015.07.31 reversed traversal order in fbc_induce()
            2015.08.02 some adaptations to simplify external changes
            2015.08.04 bug in function nbc_delete() fixed
            2016.11.11 function normd() replaced by random/randn()
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "nbayes.h"
#include "random.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI        3.14159265358979323846  /* \pi */
#endif
#define EPSILON     1e-12       /* to handle roundoff errors */
#define BLKSIZE     16          /* block size for arrays */

/* --- error messages --- */
/* error codes 0 to -15 defined in scanner.h */
#define E_ATTEXP    (-16)       /* attribute expected */
#define E_UNKATT    (-17)       /* unknown attribute */
#define E_DUPATT    (-18)       /* duplicate attribute */
#define E_MISATT    (-19)       /* missing attribute */
#define E_INVATT    (-20)       /* invalid attribute */
#define E_TRGATT    (-21)       /* target attribute name expected */
#define E_TRGTYPE   (-22)       /* target attribute is not nominal */
#define E_CLSCNT    (-23)       /* target has too few values/classes */
#define E_CLSEXP    (-24)       /* target value/class expected */
#define E_UNKCLS    (-25)       /* unknown target value/class */
#define E_DUPCLS    (-26)       /* duplicate target value/class */
#define E_DUPICLS   (-27)       /* duplicate class (implicit) */
#define E_MISCLS    (-28)       /* missing target value/class */
#define E_VALEXP    (-29)       /* attribute value expected */
#define E_UNKVAL    (-30)       /* unknown attribute value */
#define E_DUPVAL    (-31)       /* duplicate attribute value */
#define E_DUPIVAL   (-32)       /* duplicate attribute value (impl.) */
#define E_ESTEXP    (-33)       /* estimation flag expected */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- selectable attribute --- */
  ATTID  attid;                 /* attribute identifier */
  double errs;                  /* number of misclassifications */
} SELATT;                       /* (selectable attribute) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef NBC_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_ATTEXP  -16 */  "#attribute expected instead of '%s'",
  /* E_UNKATT  -17 */  "#unknown attribute '%s'",
  /* E_DUPATT  -18 */  "#duplicate attribute '%s'",
  /* E_MISATT  -19 */  "#missing attribute '%s'",
  /* E_INVATT  -20 */  "#invalid attribute '%s'",
  /* E_TRGATT  -21 */  "#target attribute expected instead of '%s'",
  /* E_TRGTYPE -22 */  "#target attribute '%s' is not nominal",
  /* E_CLSCNT  -23 */  "#target attribute '%s' has too few values",
  /* E_CLSEXP  -24 */  "#target value/class expected instead of '%s'",
  /* E_UNKCLS  -25 */  "#unknown target value/class '%s'",
  /* E_DUPCLS  -26 */  "#duplicate target value/class '%s'",
  /* E_DUPICLS -27 */  "#duplicate target value/class '%s' (implicit)",
  /* E_MISCLS  -28 */  "#missing target value/class '%s'",
  /* E_VALEXP  -29 */  "#attribute value expected instead of '%s'",
  /* E_UNKVAL  -30 */  "#unknown attribute value '%s'",
  /* E_DUPVAL  -31 */  "#duplicate attribute value '%s'",
  /* E_DUPIVAL -32 */  "#duplicate attribute value '%s' (implicit)",
  /* E_ESTEXP  -33 */  "#estimation flag expected instead of '%s'",
};
#endif

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/
#ifdef NBC_INDUCE

static int clsrsz (NBC *nbc, VALID clscnt)
{                               /* --- resize class dependent arrays */
  ATTID  i;                     /* loop variables for attributes */
  VALID  k, n, x;               /* loop variables for values */
  size_t s;                     /* size of value frequency arrays */
  CDIST  *dist;                 /* to traverse the condit.  distribs. */
  NORMD  *normd;                /* to traverse the normal   distribs. */
  DISCD  *discd;                /* to traverse the discrete distribs. */
  double *p;                    /* to traverse the frequency arrays */

  assert(nbc && (clscnt >= 0)); /* check the function arguments */

  /* --- resize the class dependent arrays --- */
  n = nbc->size;                /* get the class dep. array size */
  if (clscnt >= n) {            /* if the arrays are too small */
    n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
    if (n < clscnt) n = clscnt; /* compute the new array size */
    p = (double*)realloc(nbc->frqs, (size_t)n *4 *sizeof(double));
    if (!p) return -1;          /* resize the frequencies array */
    nbc->frqs   = p;            /* and set the new array */
    nbc->priors = p += n;       /* organize the rest of */
    nbc->posts  = p += n;       /* the allocated memory */
    nbc->cond   = p += n;       /* (split into arrays) */
    x = n -nbc->clscnt;         /* clear the new array fields */
    memset(nbc->frqs +nbc->clscnt, 0, (size_t)x *sizeof(double));
    for (i = 0; i < nbc->attcnt; i++) {
      dist = nbc->dists +i;     /* traverse all attributes */
      if (dist->type == 0) continue;  /* except the target */
      if (dist->type == AT_NOM){/* if the attribute is nominal */
        discd = (DISCD*)realloc(dist->discds, (size_t)n *sizeof(DISCD));
        if (!discd) return -1;  /* resize the discrete dists. array */
        dist->discds = discd;   /* and clear the new array elements */
        memset(discd +nbc->clscnt, 0, (size_t)x *sizeof(DISCD)); }
      else {                    /* if the attribute is numeric */
        normd = (NORMD*)realloc(dist->normds, (size_t)n *sizeof(NORMD));
        if (!normd) return -1;  /* resize the normal dists. array */
        dist->normds = normd;   /* and clear the new array elements */
        memset(normd +nbc->clscnt, 0, (size_t)x *sizeof(NORMD));
      }
    }  /* for (i = 0; ... */
    nbc->size = n;              /* set new size of the class arrays */
  }  /* if (clscnt >= n) ... */

  /* --- create new value frequency arrays --- */
  for (i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse all nominal attributes */
    if (dist->type != AT_NOM) continue;
    s = (size_t)dist->size *sizeof(double);
    for (k = nbc->clscnt; k < clscnt; k++) {
      discd = dist->discds +k;  /* traverse the new classes */
      discd->frqs = (double*)malloc(s+s);
      if (!discd->frqs) break;  /* allocate a value frequency array */
      discd->probs = discd->frqs +dist->size;
      memset(discd->frqs, 0,s); /* init. the value frequencies */
    }
    if (k < clscnt) break;      /* on error abort the loop */
  }                             /* finally set new number of classes */
  if (i >= nbc->attcnt) { nbc->clscnt = clscnt; return 0; }

  /* --- clean up on error --- */
  for ( ; i >= 0; i--) {        /* traverse the processed attributes */
    dist = nbc->dists +i;       /* (all of which are nominal) */
    if (dist->type != AT_NOM) continue;
    for (k = nbc->clscnt; k < clscnt; k++) {
      discd = dist->discds +k;  /* traverse the new classes */
      if (discd->frqs) { free(discd->frqs); discd->frqs = NULL; }
    }                           /* on error delete the created */
  }                             /* value frequency arrays */
  return -1;                    /* return that resizing failed */
}  /* clsrsz() */

/*--------------------------------------------------------------------*/

static int valrsz (CDIST *dist, VALID clscnt, VALID valcnt)
{                               /* --- resize the value freq. arrays */
  VALID  k, n, x;               /* loop variable, buffers */
  size_t s;                     /* size of probability arrays */
  DISCD  *discd;                /* to traverse the discrete distribs. */
  double *p;                    /* to traverse the frequency arrays */

  assert(dist                   /* check the function argument */
  &&    (dist->type == AT_NOM) && (clscnt >= 0) && (valcnt >= 0));
  n = dist->size;               /* get the value freq. array size */
  if (valcnt > n) {             /* if the arrays are too small */
    n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
    if (valcnt > n) n = valcnt; /* compute the new array size and */
    x = n -dist->cnt;           /* get the number of new elements */
    s = (size_t)n *2 *sizeof(double);  /* and the allocation size */
    for (k = 0; k < clscnt; k++) {
      discd = dist->discds +k;  /* traverse the discrete distribs. */
      p = (double*)realloc(discd->frqs, s);
      if (!p) break;            /* resize the value freq. array */
      discd->frqs  = p;         /* and the probabilities array */
      discd->probs = p +n;      /* and set the new arrays */
      memset(p +dist->cnt, 0, (size_t)x *sizeof(double));
    }                           /* clear the new array elements */
    if (k < clscnt) {           /* if an error occurred */
      s = (size_t)dist->size *2 *sizeof(double);
      while (--k >= 0) {        /* get the old value array size and */
        discd = dist->discds +k;/* traverse the processed distribs. */
        discd->frqs  = (double*)realloc(discd->frqs, s);
        discd->probs = discd->frqs +dist->size;
      }                         /* shrink all value freq. arrays */
      return -1;                /* to their old size */
    }                           /* and then abort */
    dist->size = n;             /* set the new array size */
  }
  dist->cnt = valcnt;           /* set the new number of values */
  return 0;                     /* return 'ok' */
}  /* valrsz() */

#endif
/*--------------------------------------------------------------------*/

static int exec (const NBC *nbc, ATTID attid, const INST *inst)
{                               /* --- execute for one attribute */
  VALID       i, k;             /* loop variable, value buffer */
  const CDIST *dist;            /* to traverse the conditional dists. */
  double      v, d, s;          /* temporary buffers */

  assert(nbc && inst            /* check the function arguments */
  &&    (attid >= 0) && (attid < nbc->attcnt));
  dist = nbc->dists +attid;     /* get the distribution array */
  assert(dist->type != 0);      /* and check the attribute type */
  if (dist->type == AT_NOM) {   /* --- if the attribute is nominal */
    k = inst->n;                /* get and check the attribute value */
    if ((k < 0) || (k >= dist->cnt)) return -1;
    for (i = 0; i < nbc->clscnt; i++) {
      d = dist->discds[i].probs[k];
      nbc->cond[i] = (d > 0) ? d : EPSILON;
    } }                         /* copy the class probabilities */
  else {                        /* --- if the attribute is numeric */
    if (dist->type == AT_INT) { /* if the attribute is integer-valued */
      if (isnull(inst->i)) return -1;
      v = (double)inst->i; }    /* check and get the attribute value */
    else {                      /* if the attribute is real-valued */
      if (isnan (inst->f)) return -1;
      v = (double)inst->f;      /* check and get the attribute value */
    }                           /* (convert both types to double) */
    for (i = 0; i < nbc->clscnt; i++) {
      d = v -dist->normds[i].exp;
      s =    dist->normds[i].var;
      if (s < EPSILON) s = EPSILON;
      s *= 2; d = exp(-(d*d)/s) /sqrt(M_PI*s);
      nbc->cond[i] = (d > 0) ? d : EPSILON;
    }                           /* compute the probability density */
  }                             /* at the value of the attribute */
  return 0;                     /* return 'ok' */
}  /* exec() */

/*--------------------------------------------------------------------*/

static double typical (const NBC *nbc, const TUPLE *tpl, VALID class)
{                               /* --- get typicality index */
  ATTID       i;                /* loop variable for attributes */
  const CDIST *dist;            /* to traverse the condit. distribs. */
  const INST  *inst;            /* to traverse the instances */
  double      typi;             /* maximal probability */

  typi = nbc->scale *(nbc->posts[class] /nbc->priors[class]);
  for (i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse all attributes */
    if ((dist->type == 0)       /* except the target attribute */
    ||  (dist->mark <  0))      /* and all unmarked attributes */
      continue;                 /* (not part of classifier) */
    inst = (tpl)                /* get the attribute instantiation */
         ? tpl_colval(tpl, i)   /* from the tuple or the attrib. set */
         : att_inst(as_att(nbc->attset, i));
    if      (dist->type == AT_INT) { /* if integer attribute */
      if (!isnull(inst->i)) typi /= dist->normds[class].max; }
    else if (dist->type == AT_FLT) { /* if real-valued attribute */
      if (!isnan (inst->f)) typi /= dist->normds[class].max; }
    else {                      /* if nominal attribute */
      if (!isnone(inst->n) && (inst->n < dist->cnt))
        typi /= dist->discds[class].max;
    }                           /* multiply with maximal probability */
  }                             /* (find best possible probability) */
  return typi;                  /* return the typicality index */
}  /* typical() */

/*--------------------------------------------------------------------*/
#ifdef NBC_INDUCE

static int eval (NBC *nbc, TABLE *table, int mode,
                 SELATT *selas, ATTID cnt)
{                               /* --- evaluate selectable attributes */
  TPLID  n;                     /* loop variable  for tuples */
  ATTID  i, k;                  /* loop variables for attributes */
  VALID  v;                     /* loop variable  for values */
  TUPLE  *tpl;                  /* to traverse the tuples */
  double *s, *d;                /* to traverse the probabilities */
  double max, tmp;              /* maximum of probabilities, buffer */
  VALID  old, new;              /* old and new predicted class */
  VALID  cls;                   /* actual class of a tuple */

  assert(nbc && table && selas  /* check the function arguments */
  &&    (cnt > 0) && (mode & (NBC_ADD|NBC_REMOVE)));
  for (n = tab_tplcnt(table); --n >= 0; ) {
    tpl = tab_tpl(table, n);    /* traverse the tuples in the table */
    cls = tpl_colval(tpl, nbc->trgid)->n;
    if (cls < 0) continue;      /* skip tuples with an null class */
    old = nbc_exec(nbc, tpl, NULL, NULL);
    for (i = 0; i < cnt; i++) { /* traverse the selectable attributes */
      k = selas[i].attid;       /* get the attribute identifier */
      if (exec(nbc, k, tpl_colval(tpl, k)) != 0)
        new = old;              /* evaluate the classifier and */
      else {                    /* on failure use the old class */
        s = nbc->cond;          /* if a probability distribution */
        d = nbc->posts;         /* could be determined, traverse it */
        if (mode & NBC_ADD) {   /* if to add attributes, */
          max = *d * *s;        /* multiply with cond. probability */
          for (new = 0, v = 1; v < nbc->clscnt; v++) {
            tmp = *++d * *++s;  /* compute new probability */
            if (tmp > max) { max = tmp; new = v; }
          } }                   /* find the most probable class */
        else {                  /* if to remove attributes, */
          max = *d / *s;        /* divide by cond. probability */
          for (new = 0, v = 1; v < nbc->clscnt; v++) {
            tmp = *++d / *++s;  /* compute new probability */
            if (tmp > max) { max = tmp; new = v; }
          }                     /* find the most probable class */
        }                       /* for the current tuple */
      }                         /* (det. new classification result) */
      if (new != cls) selas[i].errs += tpl_getwgt(tpl);
    }                           /* count the misclassifications */
  }                             /* of the modified classifier */
  return 0;                     /* return 'ok' */
}  /* eval() */

#endif
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

NBC* nbc_create (ATTSET *attset, ATTID trgid)
{                               /* --- create a naive Bayes class. */
  ATTID i, n;                   /* loop variables for attributes */
  VALID k;                      /* loop variable  for values/classes */
  NBC   *nbc;                   /* created classifier */
  ATT   *att;                   /* to traverse the attributes */
  CDIST *dist;                  /* to traverse the condit.  distribs. */
  DISCD *discd;                 /* to traverse the discrete distribs. */

  assert(attset && (trgid >= 0) /* check the function arguments */
  &&    (trgid < as_attcnt(attset))
  &&    (att_type(as_att(attset, trgid)) == AT_NOM));

  /* --- create the classifier body --- */
  n   = as_attcnt(attset);      /* get the number of attributes */
  nbc = (NBC*)malloc(sizeof(NBC) +(size_t)(n-1) *sizeof(CDIST));
  if (!nbc) return NULL;        /* allocate the classifier body */
  for (i = 0; i < n; i++) {     /* clear the distribution arrays */
    nbc->dists[i].discds = NULL; nbc->dists[i].normds = NULL; }
  nbc->attset = attset;         /* (for a proper clean up on error) */
  nbc->attcnt = n;              /* and initialize the other fields */
  nbc->trgatt = as_att(attset, nbc->trgid = trgid);
  nbc->size   = nbc->clscnt = att_valcnt(nbc->trgatt);
  nbc->total  = nbc->corr   = nbc->mode = 0;

  /* --- initialize the class distributions --- */
  assert(nbc->clscnt > 0);      /* check the number of classes */
  nbc->frqs   = (double*)malloc((size_t)nbc->clscnt *4 *sizeof(double));
  if (!nbc->frqs) { nbc_delete(nbc, 0); return NULL; }
  nbc->priors = nbc->frqs   +nbc->clscnt;  /* create class arrays */
  nbc->posts  = nbc->priors +nbc->clscnt;  /* and initialize them */
  nbc->cond   = nbc->posts  +nbc->clscnt;
  memset(nbc->frqs, 0, (size_t)nbc->clscnt *sizeof(double));

  /* --- initialize the conditional distributions --- */
  for (i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse all attributes */
    dist->mark = -1;            /* and unmark them */
    if (i == trgid) {           /* if this is the class attribute, */
      dist->type = 0; continue;}/* clear the type for easier recogn. */
    att = as_att(attset, i);    /* get the next attribute */
    dist->type = att_type(att); /* and its type */
    if (dist->type == AT_NOM) { /* -- if the attribute is nominal */
      dist->cnt = dist->size = att_valcnt(att);
      dist->discds = (DISCD*)calloc((size_t)nbc->size, sizeof(DISCD));
      if (!dist->discds) { nbc_delete(nbc, 0); return NULL; }
      if (dist->cnt <= 0)       /* create an array of discrete dists. */
        continue;               /* and check whether there are values */
      for (k = 0; k < nbc->clscnt; k++) {
        discd = dist->discds +k;/* traverse the conditional dists. */
        discd->frqs = (double*)malloc((size_t)dist->size
                                      *2 *sizeof(double));
        if (!discd->frqs) { nbc_delete(nbc, 0); return NULL; }
        discd->probs = discd->frqs +dist->size;
        memset(discd->frqs, 0, (size_t)dist->size *sizeof(double));
      } }                       /* init. the value frequencies */
    else {                      /* -- if the attribute is numeric */
      dist->cnt = dist->size = 0;
      dist->normds = (NORMD*)calloc((size_t)nbc->size, sizeof(NORMD));
      if (!dist->normds) { nbc_delete(nbc, 0); return NULL; }
    }                           /* clear the sums from which expected */
  }                             /* value and variance are computed */

  return nbc;                   /* return the created classifier */
}  /* nbc_create() */

/*--------------------------------------------------------------------*/

void nbc_delete (NBC *nbc, int delas)
{                               /* --- delete a naive Bayes class. */
  ATTID i;                      /* loop variable for attributes */
  VALID k;                      /* loop variable for values */
  CDIST *dist;                  /* to traverse the condit.  distribs. */

  assert(nbc);                  /* check the function argument */
  for (i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse the attributes */
    if (dist->discds) {         /* if the attribute is nominal */
      for (k = 0; k < nbc->clscnt; k++)
        if (dist->discds[k].frqs) free(dist->discds[k].frqs);
      free(dist->discds);       /* delete all frequency arrays */
    }                           /* and the distribution arrays */
    if (dist->normds)           /* if the attribute is metric, */
      free(dist->normds);       /* delete the normal distributions */
  }
  if (nbc->frqs) free(nbc->frqs);
  if (delas)     as_delete(nbc->attset);
  free(nbc);                    /* delete the classifier body */
}  /* nbc_delete() */

/*--------------------------------------------------------------------*/

void nbc_clear (NBC *nbc)
{                               /* --- clear a naive Bayes classifier */
  ATTID i;                      /* loop variable for attributes */
  VALID k;                      /* loop variable for values */
  CDIST *dist;                  /* to traverse the condit.  distribs. */

  assert(nbc);                  /* check the function argument */
  nbc->total = 0;               /* clear the frequency distribution */
  memset(nbc->frqs, 0, (size_t)nbc->clscnt *sizeof(double));
  for (i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse all attributes */
    if (dist->type == 0) continue;    /* except the target */
    if (dist->type == AT_NOM) { /* if the attribute is nominal */
      for (k = 0; k < nbc->clscnt; k++) {
        dist->discds[k].cnt = 0;/* traverse the distributions */
        memset(dist->discds[k].frqs,0,(size_t)dist->cnt*sizeof(double));
      } }                       /* clear the value frequencies */
    else {                      /* if the attribute is numeric */
      memset(dist->normds, 0, (size_t)nbc->clscnt *sizeof(NORMD));
    }                           /* clear the sums from which expected */
  }                             /* value and variance are computed */
}  /* nbc_clear() */

/*--------------------------------------------------------------------*/
#ifdef NBC_INDUCE

int nbc_add (NBC *nbc, const TUPLE *tpl)
{                               /* --- add an instantiation */
  ATTID  i;                     /* loop variable */
  VALID  cls;                   /* value of class attribute */
  WEIGHT wgt;                   /* instantiation weight */
  const  INST *inst;            /* to traverse the instances */
  CDIST  *dist;                 /* to traverse the condit. distribs. */
  NORMD  *normd;                /* to access normal   distributions */
  DISCD  *discd;                /* to access discrete distributions */
  double v;                     /* buffer (for an attribute value) */

  assert(nbc);                  /* check the function argument */

  /* --- get class and weight --- */
  if (tpl) {                    /* if a tuple is given */
    cls = tpl_colval(tpl, nbc->trgid)->n;
    wgt = tpl_getwgt(tpl); }    /* get the class and the tuple weight */
  else {                        /* if no tuple is given */
    cls = att_inst(nbc->trgatt)->n;
    wgt = as_getwgt(nbc->attset);
  }                             /* get the class and the inst. weight */
  if (cls < 0) return 0;        /* if the class is null, abort */
  assert(wgt >= 0);             /* check the tuple weight */

  /* --- update class distribution --- */
  if ((cls >= nbc->clscnt)      /* if the class is a new one, */
  &&  (clsrsz(nbc,cls+1) != 0)) /* resize the class dependent arrays */
    return -1;                  /* (frequencies and distributions) */
  nbc->frqs[cls] += wgt;        /* update the class frequency */
  nbc->total     += wgt;        /* and the total frequency */

  /* --- update conditional distributions --- */
  for (i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse all attributes */
    if (dist->type == 0) continue;    /* except the target */
    inst = (tpl)                /* get the attribute instantiation */
         ? tpl_colval(tpl, i)   /* from the tuple or the att. set */
         : att_inst(as_att(nbc->attset, i));
    if (dist->type == AT_NOM) { /* -- if the attribute is nominal */
      if (inst->n < 0)          /* if the attribute value is null, */
        continue;               /* there is nothing to do */
      if ((inst->n >= dist->cnt)
      &&  (valrsz(dist, nbc->clscnt, inst->n+1) != 0))
        return -1;              /* resize the value freq. arrays */
      discd = dist->discds +cls;   /* get the proper distribution */
      discd->frqs[inst->n] += wgt; /* and update the value frequency */
      discd->cnt += wgt; }         /* and the total frequency */
    else {                      /* -- if the attribute is numeric */
      if (dist->type == AT_INT){/* if the attribute is integer-valued */
        if (isnull(inst->i)) continue;
        v = (double)inst->i; }  /* check and get the attribute value */
      else {                    /* if the attribute is real-valued */
        if (isnan (inst->f)) continue;
        v = (double)inst->f;    /* check and get the attribute value */
      }                         /* (convert both types to double) */
      normd = dist->normds +cls;/* get the proper distribution */
      normd->cnt += wgt;        /* update the case counter */
      normd->sv  += wgt *v;     /* the sum of the values, and */
      normd->sv2 += wgt *v*v;   /* the sum of their squares */
    }                           /* (expected value and variance */
  }                             /*  are computed in nbc_setup) */
  return 0;                     /* return 'ok' */
}  /* nbc_add() */

/*--------------------------------------------------------------------*/

NBC* nbc_induce (TABLE *table, ATTID trgid, int mode)
{                               /* --- induce a naive Bayes class. */
  int    r = 0;                 /* result of function call */
  ATTID  i, k;                  /* loop variables for attributes */
  VALID  c;                     /* loop variable  for values/classes */
  TPLID  j, n;                  /* loop variables for tuples */
  ATTID  cnt;                   /* number of selectable attributes */
  VALID  cls;                   /* predicted class */
  NBC    *nbc;                  /* created classifier */
  ATTSET *attset;               /* attribute set of the classifier */
  SELATT *selas;                /* array of selectable attributes */
  TUPLE  *tpl;                  /* to traverse the tuples */
  CDIST  *dist;                 /* to traverse the cond. distribs. */
  double *p;                    /* to traverse the class probs. */
  double max;                   /* maximum of class probabilities */
  double errs;                  /* weight sum of misclassified tuples */

  assert(table                  /* check the function arguments */
  &&    (trgid >= 0) && (trgid < tab_colcnt(table))
  &&    (att_type(as_att(tab_attset(table), trgid)) == AT_NOM));

  /* --- create a classifier --- */
  attset = tab_attset(table);   /* get the attribute set of the table */
  if (mode & NBC_CLONE) {       /* if the corresp. flag is set, */
    attset = as_clone(attset);  /* clone the attribute set */
    if (!attset) return NULL;   /* of the given data table, */
  }                             /* then create a classifier */
  nbc = nbc_create(attset, trgid);
  if (!nbc) { if (mode & NBC_CLONE) as_delete(attset); return NULL; }

  /* --- build initial classifier --- */
  for (n = tab_tplcnt(table), j = 0; j < n; j++)
    nbc_add(nbc, tab_tpl(table, j));
  if (!(mode & (NBC_ADD|NBC_REMOVE)))
    return nbc;                 /* if no simp. is requested, abort */

  /* --- evaluate initial classifier --- */
  nbc_setup(nbc, mode);         /* start with a full classifier */
  if (mode & NBC_ADD) {         /* if to add attributes */
    p = nbc->frqs; errs = max = *p;
    for (c = nbc->clscnt; --c > 0; ) {
      errs += *++p;             /* traverse the class frequencies, */
      if (*p > max) max = *p;   /* sum them, and determine their */
    }                           /* maximum (find the majority class) */
    errs -= max; }              /* compute the number of prior errors */
  else {                        /* if to remove attributes */
    for (errs = 0, n = tab_tplcnt(table); --n >= 0; ) {
      tpl = tab_tpl(table,n);   /* traverse the tuples in the table */
      cls = tpl_colval(tpl, trgid)->n;
      if (cls < 0) continue;    /* skip tuples with an null class */
      assert(cls < nbc->clscnt);/* check the class value */
      if (nbc_exec(nbc, tpl, NULL, NULL) != cls)
        errs += tpl_getwgt(tpl);/* classify the current tuple */
    }                           /* and determine the number */
  }                             /* of misclassifications */
  #ifndef NDEBUG
  fprintf(stderr, "\n%8g error(s) initially\n", errs);
  #endif                        /* print a counter for debugging */

  /* --- collect selectable attributes --- */
  selas = malloc((size_t)nbc->attcnt *sizeof(SELATT));
  if (!selas) { nbc_delete(nbc, mode & NBC_CLONE); return NULL; }
  for (cnt = i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse all attributes */
    dist->mark = -1;            /* and unmark them (not selected) */
    if ( (dist->type == 0)      /* except the target attribute */
    ||  ((dist->type == AT_NOM) /* and all nominal attributes */
    &&   (dist->cnt  <= 0)))    /* that do not have any values */
      continue;
    if (mode & NBC_REMOVE)      /* if to remove attributes, */
      dist->mark = i;           /* mark all selectable attributes */
    selas[cnt++].attid = i;     /* note the selectable attributes */
  }                             /* afterwards mark the class */
  nbc->dists[nbc->trgid].mark = nbc->trgid;
  #ifndef NDEBUG
  fprintf(stderr, "%8"ATTID_FMT" selectable attribute(s)\n", cnt);
  #endif                        /* print a counter for debugging */

  /* --- select attributes --- */
  while ((cnt  > 0)             /* while there are selectable atts. */
  &&     (errs > 0)) {          /* and the classifier is not perfect */
    for (i = 0; i < cnt; i++)   /* traverse the selectable attributes */
      selas[i].errs = 0;        /* and clear the numbers of errors */
    r = eval(nbc, table, mode, selas, cnt);
    if (r < 0) break;           /* evaluate selectable attributes */
    for (i = k = 0; ++i < cnt;) /* traverse the selectable attributes */
      if (selas[i].errs < selas[k].errs)
        k = i;                  /* find least number of errors */
    if ( (selas[k].errs >  errs)/* if more tuples were misclassified */
    ||  ((selas[k].errs >= errs) && (mode & NBC_ADD)))
      break;                    /* abort the selection loop */
    errs = selas[k].errs;       /* note the new number of errors and */
    i    = selas[k].attid;      /* mark/unmark the selected attribute */
    nbc->dists[i].mark = (mode & NBC_ADD) ? i : -1;
    #ifndef NDEBUG
    fprintf(stderr, "%8g error(s) with %s\n",
                    errs, att_name(as_att(attset, i)));
    #endif                      /* print a counter for debugging */
    for (--cnt; k < cnt; k++)   /* remove the selected attribute */
      selas[k].attid = selas[k+1].attid;
  }
  free(selas);                  /* delete the selectable atts. array */

  if (r < 0) {                  /* if an error occurred, abort */
    nbc_delete(nbc, mode & NBC_CLONE); return NULL; }
  return nbc;                   /* return the created classifier */
}  /* nbc_induce() */

/*--------------------------------------------------------------------*/

ATTID nbc_mark (NBC *nbc)
{                               /* --- mark used/selected attributes */
  ATTID m;                      /* attribute marker */
  ATTID i;                      /* loop variable for attributes */
  ATTID cnt;                    /* attribute counter */

  assert(nbc);                  /* check the function argument */
  for (cnt = i = 0; i < nbc->attcnt; i++) {
    if      (i == nbc->trgid)        m =  0;
    else if (nbc->dists[i].mark < 0) m = -1;
    else                    { cnt++; m = +1; }
    att_setmark(as_att(nbc->attset, i), m);
  }                             /* transfer marker to attribute set */
  return cnt;                   /* return the number of marked atts. */
}  /* nbc_mark() */             /* (excluding the target attribute) */

#endif
/*--------------------------------------------------------------------*/

void nbc_setup (NBC *nbc, int mode)
{                               /* --- set up a naive Bayes class. */
  ATTID  i;                     /* loop variable  for attributes */
  VALID  k, n;                  /* loop variables for values/classes */
  CDIST  *dist;                 /* to traverse the condit.  distribs. */
  NORMD  *normd;                /* to traverse the normal   distribs. */
  DISCD  *discd;                /* to traverse the discrete distribs. */
  double cnt, sp;               /* number of cases, sum of priors */
  double add;                   /* Laplace corr. + distributed weight */
  double x;                     /* buffer for computations */

  assert(nbc);                  /* check the function arguments */

  /* --- estimate class probabilities --- */
  cnt = nbc->total +nbc->corr *(double)nbc->clscnt;
  if (cnt > 0)                  /* if the denominator is valid, */
    for (k = 0; k < nbc->clscnt; k++)  /* estimate class probs. */
      nbc->priors[k] = (nbc->frqs[k] +nbc->corr) /cnt;
  else                          /* otherwise clear all probabilities */
    memset(nbc->priors, 0, (size_t)nbc->clscnt *sizeof(double));

  /* --- estimate conditional probabilities --- */
  for (i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse all attributes */
    if (dist->type == 0) {      /* except the target attribute */
      dist->mark = 0; continue; }
    dist->mark = (!(mode & NBC_MARKED)
              || (att_getmark(as_att(nbc->attset, i)) >= 0)) ? 1 : -1;
    if (dist->mark < 0)         /* otherwise keep the */
      continue;                 /* selection of attributes */
    if (dist->type == AT_NOM) { /* -- if the attribute is nominal */
      if (dist->cnt <= 0) {     /* if the attribute has no values, */
        dist->mark = -1; continue; }     /* there is nothing to do */
      sp = (double)dist->cnt *nbc->corr; /* compute sum of priors */
      for (k = 0; k < nbc->clscnt; k++) {
        discd = dist->discds +k;/* traverse the distributions */
        n     = dist->cnt;      /* get the number of att. values */
        add   = (mode & NBC_DWNULL) ? nbc->frqs[k] : discd->cnt;
        cnt   = sp +add;        /* compute denominator of estimator */
        if (cnt > 0) {          /* if the estimator is valid */
          add = nbc->corr +(add -(double)discd->cnt) /(double)n;
          while (--n >= 0)      /* estimate the probabilities */
            discd->probs[n] = (discd->frqs[n] +add) /cnt; }
        else                    /* otherwise clear all probabilities */
          memset(discd->probs, 0, (size_t)n *sizeof(double));
      } }
    else {                      /* -- if the attribute is numeric */
      for (k = 0; k < nbc->clscnt; k++) {
        normd = dist->normds +k;/* traverse the distributions */
        cnt   = normd->cnt;     /* get the number of cases */
        normd->exp = (cnt > 0) ? normd->sv /cnt : 0;
        if (!(mode & NBC_MAXLLH)) cnt -= 1;
        normd->var = (cnt > 0)
                   ? (normd->sv2 -normd->exp *normd->sv) /cnt : 0;
      }                         /* estimate mean and variance */
    }                           /* of the normal distribution */
  }                             /* (unbiased or max. likelihood) */

  /* --- determine maximal probabilities --- */
  for (k = nbc->clscnt; --k >= 0; ) { /* traverse the classes */
    for (i = 0; i < nbc->attcnt; i++) {
      dist = nbc->dists +i;     /* traverse the attributes */
      if (dist->mark <= 0) continue;  /* except the target */
      if (dist->type != AT_NOM){/* -- if the attribute is numeric */
        normd = dist->normds+k; /* compute the maximum value */
        x = sqrt(2*M_PI *normd->var);
        normd->max = 1/((x > 0) ? x : sqrt(2*M_PI*EPSILON)); }
      else {                    /* -- if the attribute is nominal */
        discd = dist->discds+k; /* traverse the attribute values and */
        discd->max = 0;         /* find maximal cond. probability */
        for (n = 0; n < dist->cnt; n++)
          if (discd->probs[n] > discd->max)
            discd->max = discd->probs[n];
      }                         /* (best possible probability */
    }                           /*  per class is needed later */
  }                             /*  to compute the typicality) */

  nbc->dists[nbc->trgid].mark = 1;  /* mark the target attribute */
}  /* nbc_setup() */

/*----------------------------------------------------------------------
Point estimator for the expected value of a normal distribution:
  \hat{\mu} = \frac{1}{n} \sum_{i=1}^n x_i
(unbiased, consistent, and efficient)

Point estimator for the variance of a normal distribution:
  \hat{\sigma}^2 = \frac{1}{n-1} \sum_{i=1}^n (x_i - \hat{\mu})^2
                 = \frac{1}{n-1} (\sum_{i=1}^n x_i^2 - n\hat{\mu}^2)
(unbiased and consistent)

Maximum likelihood estimator for the variance of a normal distribution:
  \hat{\sigma}^2 = \frac{1}{n} \sum_{i=1}^n (x_i - \hat{\mu})^2
                 = \frac{1}{n} (\sum_{i=1}^n x_i^2 - n\hat{\mu}^2)
(consistent)

Source: R.L. Larsen and M.L. Marx.
        An Introduction to Mathematical Statistics and Its Applications.
        Prentice Hall, Englewood Cliffs, NJ, USA 1986, pp. 312 & 314
----------------------------------------------------------------------*/

VALID nbc_exec (NBC *nbc, const TUPLE *tpl, double *conf, double *typi)
{                               /* --- execute a naive Bayes class. */
  ATTID      i;                 /* loop variable  for attributes */
  VALID      k, c;              /* loop variables for values/classes */
  CDIST      *dist;             /* to traverse the cond. distribs. */
  const INST *inst;             /* to traverse the instances */
  double     sum;               /* sum of class probabilities */

  assert(nbc);                  /* check the function argument */

  /* --- initialize --- */
  memcpy(nbc->posts, nbc->priors, (size_t)nbc->clscnt *sizeof(double));
  nbc->scale = 1.0;             /* init. the posterior distribution */

  /* --- process attribute values --- */
  for (i = 0; i < nbc->attcnt; i++) {
    dist = nbc->dists +i;       /* traverse all attributes */
    if ((dist->type == 0)       /* except the target attribute */
    ||  (dist->mark <  0))      /* and all unmarked attributes */
      continue;                 /* (not in the classifier) */
    inst = (tpl)                /* get the attribute instantiation */
         ? tpl_colval(tpl, i)   /* from the tuple or the att. set */
         : att_inst(as_att(nbc->attset, i));
    if (exec(nbc, i, inst) < 0) /* execute the classifier */
      continue;                 /* for the current attribute */
    sum = 0;                    /* update probs. with conditionals */
    for (k = 0; k < nbc->clscnt; k++)
      sum += nbc->posts[k] *= nbc->cond[k];
    if ((sum > 1e-24) && (sum < 1e24))
      continue;                 /* if the sum is ok, continue */
    if (sum <= 0) break;        /* if the sum is fubar, abort */
    nbc->scale *= sum;          /* collect the scaling factors */
    sum = 1.0/sum;              /* and compute the actual factor */
    for (k = 0; k < nbc->clscnt; k++)
      nbc->posts[k] *= sum;     /* renormalize in order to avoid */
  }                             /* a numerical over- or underflow */
  sum = nbc->posts[c = 0];      /* default: class 0 */
  for (k = 1; k < nbc->clscnt; k++) {
    if (nbc->posts[k] > nbc->posts[c]) c = k;
    sum += nbc->posts[k];       /* find the most probable class */
  }                             /* and sum all probabilities */
  if (typi)                     /* compute the typicality index */
    *typi = typical(nbc, tpl, c);
  sum = (sum > 0) ? 1/sum : 1;  /* compute the normalization factor */
  for (k = nbc->clscnt; --k >= 0; )
    nbc->posts[k] *= sum;       /* normalize the probabilities */
  if (conf)                     /* if requested, */
    *conf = nbc->posts[c];      /* get the confidence value */
  return c;                     /* return the classification result */
}  /* nbc_exec() */

/*--------------------------------------------------------------------*/

void nbc_rand (NBC *nbc, double drand (void))
{                               /* --- generate a random tuple */
  ATTID  n;                     /* loop variables for attributes */
  VALID  i, k;                  /* loop variable  for values/classes */
  double t, sum;                /* random number, sum of probs. */
  double *p;                    /* to access the probabilities */
  CDIST  *dist;                 /* to traverse the cond. distribs. */
  INST   *inst;                 /* to traverse the instances */
  NORMD  *nd;                   /* normal distribution */

  p = nbc->priors;              /* get the prior probabilities */
  t = drand();                  /* generate a random number */
  for (sum = 0, i = 0; i < nbc->clscnt; i++) {
    sum += p[i]; if (sum >= t) break; }
  if (i >= nbc->clscnt)         /* find the class that corresponds */
    i = nbc->clscnt -1;         /* to the generated random number */
  att_inst(nbc->trgatt)->n = i; /* and instantiate the target att. */
  for (n = 0; n < nbc->attcnt; n++) {
    dist = nbc->dists +n;       /* traverse all attributes */
    if ((dist->type == 0)       /* except the target attribute */
    ||  (dist->mark <  0))      /* and all unmarked attributes */
      continue;                 /* (not in the classifier) */
    inst = att_inst(as_att(nbc->attset, n));
    if (dist->type == AT_NOM) { /* --- if the attribute is nominal */
      p = dist->discds[i].probs;/* get the conditional distribution */
      t = drand();              /* generate a random number */
      for (sum = 0, k = 0; k < dist->cnt; k++) {
        sum += p[k]; if (sum >= t) break; }
      if (k >= dist->cnt)       /* find the value that corresponds */
        k = dist->cnt -1;       /* to the generated random number */
      inst->n = k; }            /* and set the attribute instance */
    else {                      /* --- if the attribute is numeric */
      nd = dist->normds +i;     /* get the conditional distribution */
      t  = sqrt(nd->var) *randn(drand) +nd->exp;
      if (dist->type == AT_FLT) inst->f = (DTFLT)t;
      else                      inst->i = (DTINT)(t +0.5);
    }                           /* sample from the normal distrib. */
  }                             /* and transform the result */
}  /* nbc_rand() */

/*--------------------------------------------------------------------*/

int nbc_desc (NBC *nbc, FILE *file, int mode, int maxlen)
{                               /* --- describe a naive Bayes class. */
  ATTID       n, cnt;           /* loop variable for attributes */
  VALID       i, k;             /* loop variables for values */
  int         pos, max;         /* current and maximal position */
  int         len, ind, l;      /* name length and indentation */
  ATT         *att;             /* to traverse the attributes */
  const CDIST *dist;            /* to traverse the condit.  distribs. */
  const NORMD *normd;           /* to traverse the normal   distribs. */
  const DISCD *discd;           /* to traverse the discrete distribs. */
  char  buf[4*AS_MAXLEN+4];     /* output buffer for names */
  char  num[80];                /* output buffer for numbers */

  assert(nbc && file);          /* check the function arguments */
  max    = (maxlen > 0) ? maxlen : INT_MAX;
  maxlen = (maxlen > 0) ? maxlen : 70;

  /* --- print a header (as a comment) --- */
  if (mode & NBC_TITLE) {       /* if the title flag is set */
    fputs("/*", file); for (l = maxlen; --l >= 2; ) fputc('-', file);
    fputs("\n  naive Bayes classifier\n", file);
    for (l = maxlen; --l >= 2; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */

  /* --- start description --- */
  scn_format(buf, att_name(nbc->trgatt), 0);
  fputs("nbc(", file);          /* get the target attribute name */
  fputs(buf, file);             /* and print it */
  fputs(") = {\n", file);       /* start the classifier */
  if ((nbc->corr > 0)           /* if estimation parameters */
  ||   nbc->mode) {             /* differ from default values */
    fprintf(file, "  params = %.16g", nbc->corr);
    if (nbc->mode & NBC_DWNULL) fputs(", dwnull", file);
    if (nbc->mode & NBC_MAXLLH) fputs(", maxllh", file);
    fputs(";\n", file);         /* print Laplace correction */
  }                             /* and estimation mode */

  /* --- print the class distribution --- */
  fputs("  prob(", file);       /* print a distribution indicator */
  fputs(buf, file);             /* print the target att. name and */
  fputs(") = {\n    ", file);   /* start the class distribution */
  ind = att_valwd(nbc->trgatt, 0) +2;  /* compute the indentation */
  for (i = 0; i < nbc->clscnt; i++) {  /* traverse the classes */
    if (i > 0)                  /* if this is not the first class, */
      fputs(",\n    ", file);   /* start a new output line */
    len = (int)scn_format(buf, att_valname(nbc->trgatt, i), 0);
    fputs(buf, file);           /* get and print the target name */
    for (pos = len+2; pos < ind; pos++)
      fputc(' ', file);         /* pad with blanks to equal width */
    fprintf(file, ": %.16g", nbc->frqs[i]);
    if (mode & NBC_REL)         /* print the absolute class frequency */
      fprintf(file, " (%.1f%%)", nbc->priors[i] *100);
  }                             /* print the relative class frequency */
  fputs(" };\n", file);         /* terminate the class distribution */

  /* --- print the (conditional) distributions --- */
  for (cnt = n = 0; n < nbc->attcnt; n++) {
    dist = nbc->dists +n;       /* traverse all attributes, */
    if ((dist->type == 0)       /* but skip the target attribute */
    ||  ((mode & NBC_MARKED)    /* and in marked mode also */
    &&   (dist->mark < 0)))     /* all unmarked attributes */
      continue;                 /* (only selected attributes) */
    cnt += 1;                   /* count the selected attribute */
    fputs("  prob(", file);     /* print a distribution indicator */
    att = as_att(nbc->attset,n);/* and get the next attribute */
    scn_format(buf, att_name(att), 0);
    fputs(buf, file);           /* print the attribute name */
    fputc('|', file);           /* and the condition separator */
    scn_format(buf, att_name(nbc->trgatt), 0);
    fputs(buf, file);           /* print the target attribute name */
    fputs(") = {\n    ", file); /* and start the cond. distribution */
    if (dist->discds) {         /* if the attribute is nominal */
      for (i = 0; i < nbc->clscnt; i++) {
        discd = dist->discds +i;/* traverse the discrete distribs. */
        if (i > 0)              /* if this is not the first class, */
          fputs(",\n    ", file);       /* start a new output line */
        len = (int)scn_format(buf, att_valname(nbc->trgatt, i), 0);
        fputs(buf, file);       /* get and print the class name */
        for (pos = len+2; pos < ind; pos++)
          fputc(' ', file);     /* pad with blanks to equal width */
        fputs(":{", file);      /* start the value distribution and */
        pos += 3;               /* traverse the attribute values */
        for (k = 0; k < dist->cnt; k++) {
          if (k > 0) {          /* if this is not the first value, */
            fputc(',', file); pos++; }        /* print a separator */
          len  = (int)scn_format(buf, att_valname(att, k), 0);
          len += l = sprintf(num, ": %.16g", discd->frqs[k]);
          if (mode & NBC_REL)   /* format value frequency */
            len += sprintf(num +l, " (%.1f%%)", discd->probs[k]*100);
          if ((pos     > ind)   /* if the line would get too long */
          &&  (pos+len > max-4)) {
            fputc('\n', file);  /* start a new line and indent */
            for (pos = 0; pos < ind; pos++) fputc(' ', file); }
          else {                /* if there is enough space left, */
            fputc(' ', file); pos++; }  /* only print a separator */
          fputs(buf, file); fputs(num, file);
          pos += len;           /* print value and its frequency */
        }                       /* and update the output position */
        fputs(" }", file);      /* terminate the value distribution */
      } }
    else {                      /* if the attribute is numeric, */
      for (i = 0; i < nbc->clscnt; i++) {
        normd = dist->normds +i;/* traverse the normal distributions */
        if (i > 0)              /* if this is not the first class, */
          fputs(",\n    ", file);       /* start a new output line */
        len = (int)scn_format(buf, att_valname(nbc->trgatt, i), 0);
        fputs(buf, file);       /* get and print the class name */
        for (pos = len+2; pos < ind; pos++)
          fputc(' ', file);     /* pad with blanks to equal width */
        fprintf(file, ": N(%.16g, %.16g) [%.16g]",
                normd->exp, normd->var, normd->cnt);
      }                         /* print the normal distribution */
      fputc(' ', file);         /* with expected value and variance */
    }  /* if (dist->discds) .. else .. */
    fputs("};\n", file);        /* terminate the distributions */
  }  /* for (n = 0; .. */
  fputs("};\n", file);          /* terminate the classifier */

  /* --- print additional information (as a comment) --- */
  if (mode & NBC_INFO) {        /* if the add. info. flag is set */
    fputs("\n/*", file); for (l = maxlen; --l >= 2; ) fputc('-', file);
    fprintf(file, "\n  number of attributes: %"ATTID_FMT"+1", cnt);
    fprintf(file, "\n  number of tuples    : %g\n", nbc->total);
    for (l = maxlen; --l >= 2; ) fputc('-', file);
    fputs("*/\n", file);        /* print additional classifier info. */
  }                             /* (as a comment) */
  return ferror(file) ? -1 : 0; /* return the writing result */
}  /* nbc_desc() */

/*--------------------------------------------------------------------*/
#ifdef NBC_PARSE

static int freqs (SCANNER *scan, int cls,
                  ATT *att, double *frqs, double *sum)
{                               /* --- read value frequencies */
  VALID i, n;                   /* loop variable, number of values */
  int   t;                      /* buffer for token */

  assert(scan && att && frqs && sum); /* check the function arguments */
  SCN_CHAR(scan, '{');          /* consume '{' (start of distrib.) */
  for (i = n = att_valcnt(att); --i >= 0; )
    frqs[i] = -1;               /* clear the value frequencies */
  while (1) {                   /* attribute value read loop */
    t = scn_token(scan);        /* check for a value name */
    if ((t != T_ID) && (t != T_NUM))
      SCN_ERRVAL(scan, (cls) ? E_CLSEXP : E_VALEXP);
    if (t != T_NUM) t = ':';    /* if the token is no number, */
    else {                      /* the token must be a value, */
      SCN_NEXT(scan);           /* otherwise consume the token, */
      t = scn_token(scan);      /* note the next token, and */
      scn_back(scan);           /* go back to the previous one */
    }                           /* (look ahead one token) */
    if (t == ':') {             /* if a  ':' follows */
      i = att_valid(att, scn_value(scan));
      if (i < 0)        SCN_ERRVAL(scan, (cls) ? E_UNKCLS : E_UNKVAL);
      if (frqs[i] >= 0) SCN_ERRVAL(scan, (cls) ? E_DUPCLS : E_DUPVAL);
      SCN_NEXT(scan);           /* get and consume the value */
      SCN_CHAR(scan, ':'); }    /* consume ':' */
    else {                      /* if no ':' follows */
      if (frqs[i = (i+1) % n] >= 0) {
        t = (cls) ? E_DUPICLS : E_DUPIVAL;
        SCN_ERROR(scan, t, att_valname(att, i));
      }                         /* get the cyclic successor id */
    }                           /* and check for a duplicate */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    frqs[i] = strtod(scn_value(scan), NULL);
    if (frqs[i] < 0)              SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and check the value frequency */
    if (scn_token(scan) == '('){/* if a relative number follows, */
      SCN_NEXT(scan);           /* consume '(' */
      if (scn_token(scan) != T_NUM)          SCN_ERROR(scan, E_NUMEXP);
      if (strtod(scn_value(scan), NULL) < 0) SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* consume the relative number */
      SCN_CHAR(scan, '%');      /* consume '%' */
      SCN_CHAR(scan, ')');      /* consume ')' */
    }
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop, */
  }                             /* otherwise consume ',' */
  SCN_CHAR(scan, '}');          /* consume '}' (end of distribution) */
  for (*sum = 0, i = n; --i >= 0; ) {
    if (frqs[i] < 0) frqs[i] = 0; /* clear the unset frequencies */
    else     *sum += frqs[i];   /* and sum all other frequencies */
  }                             /* to obtain the total frequency */
  return 0;                     /* return 'ok' */
}  /* freqs() */

/*--------------------------------------------------------------------*/

static int discds (NBC *nbc, SCANNER *scan, ATT *att, DISCD *discds)
{                               /* --- read discrete distributions */
  VALID i;                      /* class identifier, buffer */
  int   t;                      /* buffer for token */

  assert(nbc && scan && att && discds); /* check function arguments */
  for (i = nbc->clscnt; --i >= 0; )
    discds[i].cnt = -1;         /* unmark all distributions */
  while (1) {                   /* distribution read loop */
    t = scn_token(scan);        /* check for a class name */
    if (t != '{') {             /* if a class name is given */
      if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_CLSEXP);
      i = att_valid(nbc->trgatt, scn_value(scan));
      if (i < 0)                       SCN_ERRVAL(scan, E_UNKCLS);
      if (discds[i].cnt >= 0)          SCN_ERRVAL(scan, E_DUPCLS);
      SCN_NEXT(scan);           /* get and consume class value */
      SCN_CHAR(scan, ':'); }    /* consume ':' */
    else {                      /* if no class name is given */
      if (discds[i = (i+1) % nbc->clscnt].cnt >= 0)
        SCN_ERROR(scan, E_DUPICLS, att_valname(nbc->trgatt, i));
    }                           /* get and check cyclic successor id */
    discds[i].cnt = 0;          /* clear the counter as a flag */
    t = freqs(scan, 0, att, discds[i].frqs, &discds[i].cnt);
    if (t) return t;            /* read distribution */
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop */
  }                             /* otherwise consume ',' */
  for (i = 0; i < nbc->clscnt; i++) {
    if ((nbc->frqs[i] > 0) && (discds[i].cnt < 0))
      SCN_ERROR(scan, E_MISCLS, att_valname(nbc->trgatt, i));
  }                             /* check for a complete classifier */
  return 0;                     /* return 'ok' */
}  /* discds() */

/*--------------------------------------------------------------------*/

static int normds (NBC *nbc, SCANNER *scan, ATT *att, NORMD *normds)
{                               /* --- read normal distributions */
  VALID  i;                     /* loop variable, class index */
  double w;                     /* weight for estimation */
  int    t;                     /* buffer for token */

  assert(nbc && scan && att && normds);  /* check function arguments */
  for (i = nbc->clscnt; --i >= 0; )
    normds[i].cnt = -1;         /* unmark all distributions */
  while (1) {                   /* distribution read loop */
    t = scn_token(scan);        /* check for a class name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_CLSEXP);
    if (t == T_NUM) t = ':';    /* if the token is a number, */
    else {                      /* the token must be a class */
      SCN_NEXT(scan);           /* otherwise consume the token, */
      t = scn_token(scan);      /* note the next token, and */
      scn_back(scan);           /* go back to the previous one */
    }                           /* (look ahead one token) */
    if (t == ':') {             /* if a ':' follows */
      i = att_valid(nbc->trgatt, scn_value(scan));
      if (i < 0)              SCN_ERRVAL(scan, E_UNKCLS);
      if (normds[i].cnt >= 0) SCN_ERRVAL(scan, E_DUPCLS);
      SCN_NEXT(scan);           /* get and consume the value */
      SCN_CHAR(scan, ':'); }    /* consume ':' */
    else {                      /* if no ':' follows */
      if (normds[i = (i+1) % nbc->clscnt].cnt >= 0)
        SCN_ERROR(scan, E_DUPICLS, att_valname(att, i));
    }                           /* get and check cyclic successor id */
    normds[i].cnt = 0;          /* clear the counter as a flag */
    if ((scn_token(scan) != T_ID)  /* check for an 'N' */
    ||  (strcmp(scn_value(scan), "N") != 0))
      SCN_ERROR(scan, E_STREXP, "N");
    SCN_NEXT(scan);             /* consume 'N' */
    SCN_CHAR(scan, '(');        /* consume '(' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    normds[i].exp = strtod(scn_value(scan), NULL);
    SCN_NEXT(scan);             /* get and consume the exp. value */
    SCN_CHAR(scan, ',');        /* consume ',' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    normds[i].var = strtod(scn_value(scan), NULL);
    if (normds[i].var < 0)        SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and consume the variance */
    SCN_CHAR(scan, ')');        /* consume ')' */
    if (scn_token(scan) == '['){/* if number of cases follows, */
      SCN_NEXT(scan);           /* consume '[' and */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      normds[i].cnt = strtod(scn_value(scan), NULL);
      if (normds[i].cnt < 0)        SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* consume the number of cases */
      SCN_CHAR(scan, ']'); }    /* consume ']' */
    else {                      /* if no number of cases follows */
      normds[i].cnt = nbc->frqs[i];
      if (normds[i].cnt <= 1) normds[i].cnt = 2;
    }                           /* get the class frequencies */
    normds[i].sv  = normds[i].exp *(w = normds[i].cnt);
    if (!(nbc->mode & NBC_MAXLLH)) w -= 1;
    normds[i].sv2 = normds[i].var *w +normds[i].exp *normds[i].sv;
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort loop, */
  }                             /* otherwise consume ',' */
  for (i = 0; i < nbc->clscnt; i++) {
    if ((nbc->frqs[i] > 0) && (normds[i].cnt < 0))
      SCN_ERROR(scan, E_MISCLS, att_valname(att, i));
  }                             /* check for a complete classifier */
  return 0;                     /* return 'ok' */
}  /* normds() */

/*--------------------------------------------------------------------*/

static int conds (NBC *nbc, SCANNER *scan)
{                               /* --- read cond. distributions */
  ATTID i;                      /* attribute identifier */
  int   t;                      /* buffer for token */
  CDIST *dist;                  /* to traverse the cond. distribs. */

  assert(nbc && scan);          /* check the function arguments */
  while ((scn_token(scan) == T_ID) /* while another dist. follows */
  &&     ((strcmp(scn_value(scan), "prob") == 0)
  ||      (strcmp(scn_value(scan), "p")    == 0)
  ||      (strcmp(scn_value(scan), "P")    == 0))) {
    SCN_NEXT(scan);             /* consume 'prob', 'p' or 'P' */
    SCN_CHAR(scan, '(');        /* consume '(' */
    t = scn_token(scan);        /* check for an attribute name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_ATTEXP);
    i = as_attid(nbc->attset, scn_value(scan));
    if (i < 0)           SCN_ERRVAL(scan, E_UNKATT);
    dist = nbc->dists +i;       /* get and check the attribute */
    if (dist->type == 0) SCN_ERRVAL(scan, E_INVATT);
    if (dist->mark >= 0) SCN_ERRVAL(scan, E_DUPATT);
    dist->mark = 1;             /* set the read flag */
    SCN_NEXT(scan);             /* consume the attribute name */
    SCN_CHAR(scan, '|');        /* consume '|' (condition indicator) */
    t = scn_token(scan);        /* check for the target att. name */
    if (((t != T_ID) && (t != T_NUM))
    ||  (strcmp(scn_value(scan), att_name(nbc->trgatt)) != 0))
      SCN_ERRVAL(scan, E_TRGATT);
    SCN_NEXT(scan);             /* consume the target att. name */
    SCN_CHAR(scan, ')');        /* consume ')' */
    SCN_CHAR(scan, '=');        /* consume '=' */
    SCN_CHAR(scan, '{');        /* consume '{' */
    if (scn_token(scan) != '}'){/* if cond. distributions follows */
      t = (dist->type == AT_NOM)
        ? discds(nbc, scan, as_att(nbc->attset, i), dist->discds)
        : normds(nbc, scan, as_att(nbc->attset, i), dist->normds);
      if (t) return t;          /* read conditional distributions */
    }                           /* and check for an error */
    SCN_CHAR(scan, '}');        /* consume '}' */
    SCN_CHAR(scan, ';');        /* consume ';' */
  }  /* while ((scn_token == T_ID) .. */
  return 0;                     /* return 'ok' */
}  /* conds() */

/*--------------------------------------------------------------------*/

static int parse (ATTSET *attset, SCANNER *scan, NBC **pnbc)
{                               /* --- parse a naive Bayes classifier */
  int   t, err = 0;             /* buffer for token, error flag */
  ATTID i;                      /* loop variable for attributes */
  ATT   *att;                   /* target attribute */
  NBC   *nbc;                   /* created naive Bayes classifier */

  /* --- read header --- */
  if ((scn_token(scan) != T_ID) /* check for 'nbc' */
  || ((strcmp(scn_value(scan), "nbc")         != 0)
  &&  (strcmp(scn_value(scan), "naive_bayes") != 0)))
    SCN_ERROR(scan, E_STREXP, "nbc");
  SCN_NEXT(scan);               /* consume 'nbc' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  t = scn_token(scan);          /* check for an attribute name */
  if ((t != T_ID) && (t != T_NUM)) SCN_ERRVAL(scan, E_ATTEXP);
  i = as_attid(attset, scn_value(scan));
  if (i < 0)                       SCN_ERRVAL(scan, E_UNKATT);
  att = as_att(attset, i);      /* get and check the target attribute */
  if (att_type(att) != AT_NOM)     SCN_ERRVAL(scan, E_TRGTYPE);
  if (att_valcnt(att) < 1)         SCN_ERRVAL(scan, E_CLSCNT);
  *pnbc = nbc = nbc_create(attset, i);  /* create a naive */
  if (!nbc) SCN_ERROR(scan, E_NOMEM);   /* Bayes classifier */
  SCN_NEXT(scan);               /* consume the target attribute name */
  SCN_CHAR(scan, ')');          /* consume '(' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */

  /* --- read parameters --- */
  if ((scn_token(scan) == T_ID) /* if 'params' follows */
  &&  (strcmp(scn_value(scan), "params") == 0)) {
    SCN_NEXT(scan);             /* consume 'params' */
    SCN_CHAR(scan, '=');        /* consume '=' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    nbc->corr = strtod(scn_value(scan), NULL);
    if (nbc->corr < 0)            SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* get and consume Laplace correction */
    while (scn_token(scan) == ',') {
      SCN_NEXT(scan);           /* read list of estimation flags */
      if (scn_token(scan) != T_ID) SCN_ERRVAL(scan, E_ESTEXP);
      if      (strcmp(scn_value(scan), "dwnull") == 0)
        nbc->mode |= NBC_DWNULL;/* distribute weight for nulls */
      else if (strcmp(scn_value(scan), "maxllh") == 0)
        nbc->mode |= NBC_MAXLLH;/* use max. likelihood estimate */
      else SCN_ERRVAL(scan, E_ESTEXP);
      SCN_NEXT(scan);           /* consume the estimator flag */
    }
    SCN_CHAR(scan, ';');        /* consume ';' */
  }

  /* --- read class distribution --- */
  if ((scn_token(scan) != T_ID) /* check for 'prob', 'p' or 'P' */
  ||  ((strcmp(scn_value(scan), "prob") != 0)
  &&   (strcmp(scn_value(scan), "p")    != 0)
  &&   (strcmp(scn_value(scan), "P")    != 0)))
    SCN_ERROR(scan, E_STREXP, "prob");
  SCN_NEXT(scan);               /* consume   'prob', 'p' or 'P' */
  SCN_CHAR(scan, '(');          /* consume '(' */
  t = scn_token(scan);          /* check for   target att. name */
  if (((t != T_ID) && (t != T_NUM))
  ||  (strcmp(scn_value(scan), att_name(att)) != 0))
    SCN_ERROR(scan, E_TRGATT, att_name(att), scn_value(scan));
  SCN_NEXT(scan);               /* consume the target att. name */
  SCN_CHAR(scan, ')');          /* consume ')' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  t = freqs(scan, 1, att, nbc->frqs, &nbc->total);
  if (t) return t;              /* read the class distribution */
  SCN_CHAR(scan, ';');          /* consume ';' */
  nbc->dists[i].mark = 0;       /* mark the target attribute */

  /* --- read conditional distributions --- */
  do {                          /* read the cond. distributions, */
    t = conds(nbc, scan);      /* trying to recover on errors */
    if (t) { err = t; scn_recover(scan, ';', 0, 0, 0); }
  } while (t);                  /* while not all dists. read */
  if (err) return err;          /* if an error occurred, abort */
  for (i = 0; i < nbc->attcnt; i++) {
    if ((nbc->frqs[i] > 0) && (nbc->dists[i].mark < 0))
      SCN_ERROR(scan, E_MISATT, as_attname(attset, i));
  }                             /* check for a complete classifier */

  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* parse() */

/*--------------------------------------------------------------------*/

NBC* nbc_parse (ATTSET *attset, SCANNER *scan)
{                               /* --- parse a naive Bayes class. */
  NBC *nbc = NULL;              /* created naive Bayes classifier */

  assert(attset && scan);       /* check the function arguments */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (parse(attset, scan, &nbc) != 0) {
    if (nbc) nbc_delete(nbc,0); /* parse a naive Bayes classifier */
    return NULL;                /* if an error occurred, */
  }                             /* delete the classifier and abort */
  nbc_setup(nbc, 0);            /* set up the created classifier */
  return nbc;                   /* and then return it */
}  /* nbc_parse() */

#endif
