/*----------------------------------------------------------------------
  File    : nbayes.h
  Contents: Naive Bayes classifier management
  Author  : Christian Borgelt
  History : 1998.12.07 file created
            1998.12.16 definition of type DVEC changed
            1999.02.13 tuple parameters added to nbc_add and nbc_exec
            1999.03.10 definition of NBC_MARKED added
            1999.03.25 definition of NBC_DISTUV added
            1999.03.27 some enquiry functions added
            1999.04.05 Laplace correction parameter added
            1999.04.23 parameter 'mode' added to function nbc_parse
            1999.05.15 function nbc_mark added
            2000.11.18 function nbc_setup added, nbc_exec() adapted
            2000.11.21 functions nbc_corr() and nbc_mode() added
            2001.07.16 adapted to modified module scan
            2003.04.26 function nbc_rand() added
            2004.08.12 adapted to new module parse
            2007.03.21 function nbc_post() added (posterior prob.)
            2009.06.04 parameter typi added to function nbc_exec()
            2011.02.02 adapted to modified scanner functions
            2011.07.29 description mode NBC_INFO added
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
            2015.08.02 redesigned for external attribute selection
----------------------------------------------------------------------*/
#ifndef __NBAYES__
#define __NBAYES__
#ifdef NBC_PARSE
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#include "scanner.h"
#endif
#include "table.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- induction modes --- */
#define NBC_CLONE   0x0001      /* clone attribute set */
#define NBC_ADD     0x0002      /* greedily add attributes */
#define NBC_REMOVE  0x0004      /* greedily remove attributes */
#define NBC_DWNULL  0x0010      /* distribute weight for null values */
#define NBC_MAXLLH  0x0080      /* max. likelihood estim. of variance */

/* --- setup/induction modes --- */
#define NBC_ALL     0x0000      /* set up for all attributes */
#define NBC_MARKED  0x8000      /* set up only for marked attributes */

/* --- description modes --- */
#define NBC_TITLE   0x0001      /* print a title (as a comment) */
#define NBC_INFO    0x0002      /* print add. info. (as a comment) */
#define NBC_REL     0x0004      /* print relative numbers */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- discrete distribution --- */
  double cnt;                   /* number of cases (total frequency) */
  double max;                   /* maximal conditional probability */
  double *frqs;                 /* value frequency   array */
  double *probs;                /* value probability array */
} DISCD;                        /* (discrete distribution) */

typedef struct {                /* --- normal distribution --- */
  double cnt;                   /* number of cases (total frequency) */
  double max;                   /* maximum value (height of mode) */
  double sv;                    /* sum of values */
  double sv2;                   /* sum of squared values */
  double exp;                   /* expected value */
  double var;                   /* variance */
} NORMD;                        /* (normal distribution) */

typedef struct {                /* --- conditional distributions --- */
  ATTID  mark;                  /* whether read or to be processed */
  int    type;                  /* attribute type (0: class) */
  VALID  cnt;                   /* number of attribute values */
  VALID  size;                  /* size of value frequency arrays */
  DISCD  *discds;               /* array of discrete distributions */
  NORMD  *normds;               /* array of normal distributions */
} CDIST;                        /* (conditional distribution) */

typedef struct {                /* --- naive Bayes classifier --- */
  ATTSET *attset;               /* underlying attribute set */
  ATTID  attcnt;                /* number of attributes */
  ATT    *trgatt;               /* target attribute */
  ATTID  trgid;                 /* identifier of target attribute */
  VALID  clscnt;                /* number of classes */
  VALID  size;                  /* size of class dependent arrays */
  int    mode;                  /* estimation mode (e.g. NBC_MAXLLH) */
  double corr;                  /* Laplace correction */
  double total;                 /* total number of cases */
  double scale;                 /* scaling factor */
  double *frqs;                 /* class frequencies */
  double *priors;               /* prior     class probabilities */
  double *posts;                /* posterior class probabilities */
  double *cond;                 /* buffer for conditional probs. */
  CDIST  dists[1];              /* conditional distributions */
} NBC;                          /* (naive Bayes classifier) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern NBC*    nbc_create  (ATTSET *attset, ATTID trgid);
extern void    nbc_delete  (NBC *nbc, int delas);
extern ATTSET* nbc_attset  (NBC *nbc);
extern ATT*    nbc_trgatt  (NBC *nbc);
extern ATTID   nbc_trgid   (const NBC *nbc);
extern VALID   nbc_clscnt  (const NBC *nbc);
extern ATTID   nbc_attcnt  (const NBC *nbc);
extern VALID   nbc_valcnt  (const NBC *nbc, ATTID attid);

extern void    nbc_setmode (NBC *nbc, int mode);
extern double  nbc_getmode (NBC *nbc);
extern void    nbc_setcorr (NBC *nbc, double corr);
extern double  nbc_getcorr (NBC *nbc);

extern void    nbc_clear   (NBC *nbc);
extern double  nbc_total   (const NBC *nbc);

#ifdef NBC_INDUCE
extern int     nbc_add     (NBC *nbc, const TUPLE *tpl);
extern NBC*    nbc_induce  (TABLE *table, ATTID trgid, int mode);
extern ATTID   nbc_mark    (NBC *nbc);
#endif

extern void    nbc_setup   (NBC *nbc, int mode);
extern double  nbc_prior   (const NBC *nbc, VALID clsid);
extern double  nbc_prob    (const NBC *nbc, VALID clsid, ATTID attid,
                            VALID valid);
extern double  nbc_exp     (const NBC *nbc, VALID clsid, ATTID attid);
extern double  nbc_var     (const NBC *nbc, VALID clsid, ATTID attid);
extern double  nbc_post    (const NBC *nbc, VALID clsid);
extern VALID   nbc_exec    (NBC *nbc, const TUPLE *tpl,
                            double *conf, double *typi);
extern void    nbc_rand    (NBC *nbc, double drand (void));

extern int     nbc_desc    (NBC *nbc, FILE *file, int mode, int maxlen);
#ifdef NBC_PARSE
extern NBC*    nbc_parse   (ATTSET *attset, SCANNER *scanner);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define nbc_attset(b)       ((b)->attset)
#define nbc_trgatt(b)       ((b)->trgatt)
#define nbc_trgid(b)        ((b)->trgid)
#define nbc_clscnt(b)       ((b)->clscnt)
#define nbc_attcnt(b)       ((b)->attcnt)
#define nbc_valcnt(b,a)     ((b)->dvecs[a].valcnt)
#define nbc_total(b)        ((b)->total)

#define nbc_setmode(b,m)    ((b)->mode = (m))
#define nbc_getmode(b)      ((b)->mode)
#define nbc_setcorr(b,c)    ((b)->corr = (c))
#define nbc_getcorr(b)      ((b)->corr)

#define nbc_total(b)        ((b)->total)

#define nbc_prior(b,c)      ((b)->priors[c])
#define nbc_prob(b,c,a,v)   ((b)->dists[a].discds[c].probs[v])
#define nbc_exp(b,c,a)      ((b)->dists[a].normds[c].exp)
#define nbc_var(b,c,a)      ((b)->dists[a].normds[c].var)
#define nbc_post(b,c)       ((b)->posts[c])

#endif
